/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "ImageBase.hh"

#ifdef WIN32
#  define _USE_MATH_DEFINES
#endif //WIN32

#include <math.h>
#include <string>
#include <sstream>

using std::string;
using std::stringstream;

using namespace BIAS;
using namespace std;


ImageBase::ImageBase() : Debug() 
{
#ifdef BIAS_DEBUG
  BIASGDOUT(D_IMAGE_TRACE, "called ImageBase::ImageBase()");
#endif
  ImageData_ = NULL;
  ImageDataArray_ = NULL;
  ChannelCount_ = 0;
  Width_ = Height_ = Depth_ = BitDepth_ = 0;
  PixelCount_ = 0;
  WidthStep_ = 0;
  ColorModel_ = CM_invalid;
  StorageType_=ST_invalid;

  ROIUpperLeftX_  = 0;
  ROIUpperLeftY_  = 0;
  ROILowerRightX_ = 0;
  ROILowerRightY_ = 0;
}

ImageBase::ImageBase(const ImageBase& im) 
{ 
#ifdef BIAS_DEBUG
  BIASGDOUT(D_IMAGE_TRACE, "called ImageBase::ImageBase(const ImageBase& im)");
#endif
  ImageData_ = NULL;
  ImageDataArray_ = NULL;
  Width_ = 0;
  Height_ = 0;
  ChannelCount_ = 0;
  
  ROIUpperLeftX_  = 0;
  ROIUpperLeftY_  = 0;
  ROILowerRightX_ = im.GetWidth();
  ROILowerRightY_ = im.GetHeight();
  (*this) = im;
}


ImageBase::ImageBase(unsigned int width, unsigned int height, 
                     unsigned int channels, enum EStorageType storageType,
                     bool interleaved) 
{
#ifdef BIAS_DEBUG
  BIASGDOUT(D_IMAGE_TRACE, "called ImageBase::ImageBase(width, height,"
            <<" channels, storageType, interleaved ");
#endif

  ImageData_      = NULL;
  ImageDataArray_ = NULL;
  
  ROIUpperLeftX_  = 0;
  ROIUpperLeftY_  = 0;
  ROILowerRightX_ = width;
  ROILowerRightY_ = height;
  
  ColorModel_ = CM_invalid;
  InterleavedDataOrder_ = interleaved;
  
  Init(width,height,channels,storageType);
}

ImageBase::~ImageBase()
{
  if (ImageDataArray_ != NULL) {
    delete[] (char**)ImageDataArray_; ImageDataArray_ = NULL;
  }
  if (ImageData_!=NULL){
    delete[] (char*)ImageData_; ImageData_ = NULL;
  }
}


void ImageBase::Init(unsigned int Width, unsigned int Height, 
                     unsigned int channels, enum EStorageType storageType,
                     bool interleaved)
{
  bool ok=true; // status
  
  if ((ImageData_!=NULL) || (ImageDataArray_ !=NULL)){
    BIASERR("Image<StorageType>::Init should only be called once"<<
           "\nuse Release if you want to reinit");
    BIASASSERT(ImageData_!=NULL);
    BIASASSERT(ImageDataArray_!=NULL);
  }
#ifdef DISABLE_MODULO_4_SIZE
  if (((sizeof(StorageType)*Width) % 4) != 0){
    BIASERR("Because IplImage always uses aligned image data\n"
           << "  Image only supports images with \n"
           << "  (width * sizeof(StorageType)) modulo 4 = 0.\n"
           << "  Please choose a width which is divisible by four\n"
           << "  if multiplied with sizeof(StorageType) or use either\n"
           << "  the CopyIn function (internal only) or\n"
           << "  the ImportImage function.\n");
    return; // error
  };
  
#if USE_OPENCV == true
  BIASERR("you cannot use -DDISABLE_MODULO_4_SIZE and -DUSE_OPENCV together");
#endif
#endif

  Width_ = Width;
  Height_ = Height;
  ChannelCount_ = channels;
  PixelCount_ = Width_ * Height_;
  StorageType_ = storageType;
  ROIUpperLeftX_  = 0;
  ROIUpperLeftY_  = 0;
  ROILowerRightX_ = Width;
  ROILowerRightY_ = Height;

  switch (storageType) {
  case ST_unsignedchar:
    Depth_=sizeof(unsigned char);
    BitDepth_=8*sizeof(unsigned char);
    break;
  case  ST_char:
    Depth_=sizeof(char);
    BitDepth_=8*sizeof(char);
    break;
  case ST_unsignedshortint:
    Depth_=sizeof(unsigned short int);
    BitDepth_=8*sizeof(unsigned short int);
    break;
  case ST_shortint:
    Depth_=sizeof(short int);
    BitDepth_=8*sizeof(short int);
    break;
  case ST_unsignedint:
    Depth_=sizeof(unsigned int);
    BitDepth_=8*sizeof(unsigned int);
    break;
  case ST_int:
    Depth_=sizeof(int);
    BitDepth_=8*sizeof(int);
    break;
  case ST_float:
    Depth_=sizeof(float);
    BitDepth_=8*sizeof(float);
    break;
  case ST_double:
    Depth_=sizeof(double);
    BitDepth_=8*sizeof(double);
    break;    
  default:
    BIASERR("Init: invalid storage type '"<<storageType<<"'");
    ok=false;
    break;
  };
  
  InterleavedDataOrder_ = interleaved;
  ImageData_ = (void*) new char[Width * Height * channels * Depth_];
  
  //cerr << "image data last: ("<<Width<<"x"<<Height<<") "
  //     <<(void *)(&((char *)(ImageData_))[Width * Height * channels * Depth_-1])<<endl;
  switch (channels){
  case 3:
    SetColorModel(CM_RGB);
    break;
  case 2:
    SetColorModel(CM_YUYV422);
    break;
  case 1:
    SetColorModel(CM_Grey);
    break;
  case 4:
    SetColorModel(CM_RGBA);
    break;
  default:
    SetColorModel(CM_invalid);
    break;
  };

  if (InterleavedDataOrder_)
    WidthStep_ = Width_ * Depth_ * ChannelCount_;
  else 
    WidthStep_ = Width_ * Depth_;

  BIASASSERT( ok==true ); // all cases were ok
  UpdateImageDataArrayFromImageData_();
  
}


int ImageBase::UpdateImageDataArrayFromImageData_()
{
  const unsigned height=(IsPlanar())?
    (GetHeight()*GetChannelCount()):(GetHeight());
  if (ImageDataArray_ != NULL) delete[] ImageDataArray_;
  ImageDataArray_ = (void**) new char*[height];
  ImageDataArray_[0] = GetImageData();
#ifdef BIAS_DEBUG
  unsigned int tmpstep;
  if (InterleavedDataOrder_)
    tmpstep = Width_ * Depth_ * ChannelCount_;
  else 
    tmpstep = Width_ * Depth_;
  if ( tmpstep!=WidthStep_) {
    BIASERR("Congratulations! You found a major bug in BIAS."
            <<" Inconsistency in image dimension data.");
    abort();
  }
#endif
  for (register unsigned int i=1; i < height; i++) {
    ImageDataArray_[i] = (void*)((char*)(ImageDataArray_[i-1]) + 
                                 WidthStep_);
  };
  return 0;
}

void ImageBase::Release() {
#ifdef BIAS_DEBUG
  if (IsEmpty()) {
    BIASERR("ImageBase::Release called with empty image !");
  }
#endif
  // check if we really need to delete the mem
  if (ImageDataArray_!=NULL) {
    delete[] (char**)ImageDataArray_; 
    ImageDataArray_ = NULL;
  }
  if (ImageData_!=NULL) {
    delete[] (char*)ImageData_; 
    ImageData_=NULL;
  }
  ChannelCount_ = 0;
  Width_ = Height_ = Depth_ = BitDepth_ = 0;
  PixelCount_ = 0;
  WidthStep_  = 0;
  ColorModel_ = CM_invalid;
  StorageType_= ST_invalid;
    
  ROIUpperLeftX_  = 0;
  ROIUpperLeftY_  = 0;
  ROILowerRightX_ = 0;
  ROILowerRightY_ = 0;

  _MetaData.clear();
}

int ImageBase::StealImage(ImageBase &source) {
  if (source.IsEmpty()) {
    BIASERR("tried to steal empty image");
    return -1;  
  }
  if (!IsEmpty())
    Release();
  ImageData_ = source.GetImageData();
  ImageDataArray_ = source.GetImageDataArray();
  ChannelCount_ = source.GetChannelCount();
  ColorModel_ = source.GetColorModel();
  StorageType_ = source.GetStorageType();
  InterleavedDataOrder_=!source.IsPlanar();

  Width_ = source.GetWidth();
  Height_ = source.GetHeight();
  Depth_ = source.GetDepth();
  BitDepth_ = source.GetBitDepth();
  PixelCount_ = source.GetPixelCount();
  WidthStep_ = source.GetWidthStep();

  source.GetROI(ROIUpperLeftX_, ROIUpperLeftY_, 
                ROILowerRightX_, ROILowerRightY_);

  source.ImageData_=NULL;
  source.ImageDataArray_=NULL;
  copy(_MetaData.begin(), _MetaData.end(), source._MetaData.begin());
  source._MetaData.clear();
  
  return 0;
}

int ImageBase::GetChannel(ImageBase &source, unsigned int channel)
{
  if (IsEmpty())
    Init(source.GetWidth(), source.GetHeight(), 1,source.GetStorageType());

#ifdef BIAS_DEBUG
  if ( (source.GetHeight() != GetHeight()) || 
       ( (source.GetWidth() != GetWidth()) &&
         (source.GetWidth() != GetWidth()/2) ) )
    BIASERR("ImageConvert<StorageType>::GetChannel() invalid size of image");
  if ( GetChannelCount() != 1 )
    BIASERR("ImageConvert<StorageType>::GetChannel() should be called with one"
            <<" channel image");
  if ( ! source.IsPlanar() )
    BIASERR("ImageConvert<StorageType>::GetChannel() should be called from "<<
            "planar image");
  if ( source.GetChannelCount() == 1 )
    BIASERR("ImageConvert<StorageType>::GetChannel() only makes sense with"
            <<" color images");
#endif

  if (source.GetColorModel()==CM_YUYV422){
    switch (channel){
    case 0:
      memcpy(GetImageData(), source.GetImageData(), 
             source.GetPixelCount() * source.GetDepth());
      break;
    case 1:
      memcpy(GetImageData(), (char*)(source.GetImageData()) + 
             source.GetPixelCount()*source.GetDepth(), 
             source.GetPixelCount() * source.GetDepth() / 2);
      break;
    case 2:
      memcpy(GetImageData(), 
             (char*)(source.GetImageData()) + source.GetDepth()* 
             (source.GetPixelCount() + source.GetPixelCount()/2),
             source.GetPixelCount() * source.GetDepth() / 2);
      break;
    default:
      BIASERR("ImageConvert<StorageType>::GetChannel() invalid Channel");
      return -1;
      break;
    }
  } else 
    if ( (source.GetColorModel()==CM_RGB) || 
         (source.GetColorModel()==CM_BGR) ||
         (source.GetColorModel()==CM_HSV) ) {
      if (source.IsPlanar()){
        switch (channel){
        case 0:
          memcpy(GetImageData(), source.GetImageData(), 
                 source.GetPixelCount() * source.GetDepth());
          break;
        case 1:
          memcpy(GetImageData(), 
                 (char*)(source.GetImageData()) +
                 source.GetPixelCount()*source.GetDepth(), 
                 source.GetPixelCount() * source.GetDepth());
          break;
        case 2:
          memcpy(GetImageData(), 
                 (char*)(source.GetImageData()) + 2 * source.GetPixelCount()* 
                 source.GetDepth(), 
                 source.GetPixelCount() * source.GetDepth());
          break;
        default:
          BIASERR("ImageConvert<StorageType>::GetChannel() invalid Channel");
          return -1;
          break;
        }
      } else { // interleaved && (CM_RGB || CM_BGR || CM_HSV)
#ifdef BIAS_DEBUG
        if ((channel<0) || (channel>2)){
          BIASERR("ImageConvert<StorageType>::GetChannel() invalid Channel");
          return -1;
        }
#endif
        char *src=(char *)source.GetImageData();
        char *dst=(char *)GetImageData();
        char *end=dst+GetPixelCount()*GetDepth();
        register int depth=GetDepth(), depth2=depth<<1;
        switch (channel){
        case 0:
          while (dst<end){
            for (int i=0; i<depth; i++) *dst++=*src++;
            src+=depth2;
          }
          break;
        case 1:
          while (dst<end){
            src+=depth;
            for (int i=0; i<depth; i++) *dst++=*src++;
            src+=depth;
          }
          break;
        case 2:
          while (dst<end){
            src+=depth2;
            for (int i=0; i<depth; i++) *dst++=*src++;
          }
          break;
        default: break;
        }
      }
    } else {
      BIASERR("ImageConvert<StorageType>::GetChannel() not yet implemented for"
              <<" this color model "<<source.GetColorModel());
      return -1;   
    }
  return 0;
}

int ImageBase::SetROI(unsigned int UpperLeftX, 
                      unsigned int UpperLeftY, 
                      unsigned int LowerRightX, 
                      unsigned int LowerRightY) {

#ifdef BIAS_DEBUG
  if (UpperLeftX < 0 || UpperLeftX > GetWidth()) {
    BIASERR( "UpperLeftX "<<UpperLeftX<<" is out of image range "
             <<GetWidth());
    BIASASSERT(false);
    return -1;
  }
  if (UpperLeftY < 0 || UpperLeftY > GetHeight()) {
    BIASERR("UpperLeftY "<<UpperLeftY<<" is out of image range "
            <<GetHeight());
    BIASASSERT(false);
    return -2;
  }
  if (LowerRightX < 0 || LowerRightX > GetWidth() ) {
    BIASERR("LowerRightX "<<LowerRightX<<" is out of image range "
            <<GetWidth());
    BIASASSERT(false);
    return -3;
  }
  if (LowerRightY < 0 || LowerRightY > GetHeight() ) {
    BIASERR("LowerRightY "<<LowerRightY<<" is out of image range "
            <<GetHeight());
    BIASASSERT(false);
    return -4;
  }

  if (LowerRightX < UpperLeftX) {
    unsigned int temp = UpperLeftX;
    UpperLeftX = LowerRightX;
    LowerRightX = temp;
    BIASERR("WARNING: You swapped the x coordinates of the upper left and the"
            <<" lower right corner. It's been re-swapped!");
    BIASASSERT(false);
  }

  if (LowerRightY < UpperLeftY) {
    unsigned int temp = UpperLeftY;
    UpperLeftY = LowerRightY;
    LowerRightY = temp;
    BIASERR("WARNING: You swapped the y coordinates of the upper left and the"
            <<" lower right corner. It's been re-swapped!");
    BIASASSERT(false);
  }
#endif  

  ROIUpperLeftX_  = UpperLeftX;
  ROILowerRightX_ = LowerRightX;
  ROIUpperLeftY_  = UpperLeftY;
  ROILowerRightY_ = LowerRightY;
  return 0;
}


void ImageBase::PrintROI(ostream& os) const {
	os << "Region Of Interest (ROI):" << endl;
	os << "UpperLeftX    : " << ROIUpperLeftX_  
	   << "  /  UpperLeftY   : " << ROIUpperLeftY_ 
	   << endl;
	os << "   LowerRightX: " << ROILowerRightX_ 
	   << "  /    LowerRightY: " << ROILowerRightY_ << endl;	
}

int ImageBase::GetCopyOfROI2(ImageBase &copy) const
{
#ifdef BIAS_DEBUG
  if (IsPlanar() && GetChannelCount()>1) {
    BIASERR("only implemented for interleaved images...");
    abort();
  }
#endif
  char **dst, **src;
  register unsigned int y;

  if (!copy.IsEmpty()) copy.Release();
  copy.Init(ROILowerRightX_ - ROIUpperLeftX_ + 1, 
	    ROILowerRightY_ - ROIUpperLeftY_ + 1, GetChannelCount(), 
	    GetStorageType());
  copy.SetColorModel(GetColorModel());

  // following line must be changed if methods is extendend 
  src = (char**)GetImageDataArray();
  dst = (char**)copy.GetImageDataArray();

  unsigned int length = (ROILowerRightX_-ROIUpperLeftX_ + 1) 
    * GetChannelCount() * GetDepth();
  for (y = ROIUpperLeftY_; y <= ROILowerRightY_; y++){
    memcpy(&(dst[y-ROIUpperLeftY_][0]), 
           &(src[y][ROIUpperLeftX_*GetChannelCount()*GetDepth()]), length);
  }
  return 0;
  
}

int ImageBase::GetCopyOfROI(ImageBase &copy) const
{
#ifdef BIAS_DEBUG
  if (IsPlanar() && GetChannelCount()>1) {
    BIASERR("only implemented for interleaved images...");
    abort();
  }
#endif
  char **dst, **src;
  register unsigned int y;

  if (!copy.IsEmpty()) copy.Release();
  copy.Init(ROILowerRightX_ - ROIUpperLeftX_, 
            ROILowerRightY_ - ROIUpperLeftY_, GetChannelCount(), 
            GetStorageType());
  copy.SetColorModel(GetColorModel());

  // following line must be changed if methods is extendend 
  src = (char**)GetImageDataArray();
  dst = (char**)copy.GetImageDataArray();

  const unsigned int length = 
    (ROILowerRightX_-ROIUpperLeftX_) * GetChannelCount() * GetDepth();
  
  for (y = ROIUpperLeftY_; y < ROILowerRightY_; y++){
    memcpy(&(dst[y-ROIUpperLeftY_][0]), 
           &(src[y][ROIUpperLeftX_*GetChannelCount()*GetDepth()]), length);
  }
  return 0;
}


int ImageBase::Paste2ROI(const ImageBase& Image) 
{
#ifdef BIAS_DEBUG 
  if ( (Image.GetWidth() != (ROILowerRightX_ - ROIUpperLeftX_)) ||
       (Image.GetHeight() != (ROILowerRightY_ - ROIUpperLeftY_)) ){
    BIASERR("ImageBase:Paste2ROI():  ROI size in destination is not equal "<<
            "to image size in source, aborting! ROI:WxH  "
            <<(ROILowerRightX_ - ROIUpperLeftX_)<<"x"
            <<(ROILowerRightY_ - ROIUpperLeftY_)<<"  Im:WxH "
            <<Image.GetWidth()<<"x"<<Image.GetHeight());
    return(-1);}
  if (Image.GetColorModel() != GetColorModel()){
    BIASERR("ImageBase::Paste2ROI():  ColorModels do not match, aborting!");
    return(-1);}
#endif
  
  char **dst, **src;
  register unsigned int y;

  dst = (char**)GetImageDataArray();
  src = (char**)Image.GetImageDataArray();
  unsigned int length= (ROILowerRightX_-ROIUpperLeftX_)*GetChannelCount()*
    GetDepth();
  for (y = ROIUpperLeftY_; y < ROILowerRightY_; y++){
    memcpy(&(dst[y][ROIUpperLeftX_*GetChannelCount()*GetDepth()]), 
           &(src[y-ROIUpperLeftY_][0]),   length);
  }
  return 0;
}


int ImageBase::Cut2ROI()
{
  
  const unsigned int bytesInRow= (ROILowerRightX_-ROIUpperLeftX_)*
    GetChannelCount() * GetDepth();
  char *dst = new char[(ROILowerRightY_ - ROIUpperLeftY_)*bytesInRow];  
  char **src = (char**)GetImageDataArray();
  
  register unsigned int y;
  unsigned int start=ROIUpperLeftX_*GetChannelCount()*GetDepth();
 
  for (y = ROIUpperLeftY_; y < ROILowerRightY_; y++){
    memcpy(dst+(y-ROIUpperLeftY_)*bytesInRow, &(src[y][start]),
           bytesInRow);
  }

  Width_ =  ROILowerRightX_ - ROIUpperLeftX_;
  Height_ = ROILowerRightY_ - ROIUpperLeftY_;
  PixelCount_ = Width_ * Height_;
  WidthStep_=Width_*ChannelCount_*Depth_;
  UnsetROI();
  ReleaseImageDataPointer();
  RedirectImageDataPointer((void*)dst);
  return 0;
}

void ImageBase::SetOutsideROIZero()
{
  unsigned tlx, tly, brx, bry;
  GetROI(tlx, tly, brx, bry);
  
  if (tlx>0 || tly>0 || brx<GetWidth() || bry<GetWidth()){
    void **ida=GetImageDataArray();
    // above ROI
    memset(ida[0], 0, tly*GetWidthStep()+tlx*GetChannelCount()*GetDepth());

    // from right ROI border to left ROI border
    const int size=(GetWidth()-(brx-tlx))*GetChannelCount()*GetDepth();
    const int x=brx*GetChannelCount()*GetDepth();
    for (unsigned y=tly; y<bry; y++){
      memset((void*)&(((char*)(ida[y]))[x]), 0, size); 
    }

    // below ROI
    memset((void*)&(((char*)(ida[bry]))[x]), 0, 
           (GetHeight()-bry)*GetWidthStep());
  }
}

//////////////////////////////////////////////////////////////////////////////
//
// Assignment Operator
//
//////////////////////////////////////////////////////////////////////////////
ImageBase& ImageBase::operator=(const ImageBase& Source)
{
  /// always copy the mebers that do not depend on image data
  /// to allow feeding Image* into ImageBase*. JW 01/2005 
#ifdef BIAS_DEBUG
  _liDebugLevel = Source.GetDebugLevel();
  BIASGDOUT(D_IMAGE_TRACE, 
            "called ImageBase::operator=(const ImageBase& Source)"
            <<" source->GetStorageType = "<<Source.GetStorageType()
            <<" this->GetStorageType   = "<<GetStorageType()
            <<" source->GetColorModel  = "<<Source.GetColorModel()
            <<" this->GetColorModel    = "<<GetColorModel()
            );
#endif
  StorageType_  = Source.GetStorageType();
  
  /// now copy the image data dependant stuff if appropriate JW
  if (Source.IsEmpty()){
    if (!IsEmpty())
      Release();
    return *this;
  }

  if (!IsEmpty()) 
    if ( !SamePixelAndChannelCount(Source)
         || (Source.GetStorageType()!=GetStorageType()) )
      Release();
  if (IsEmpty())
    Init(Source.GetWidth(), Source.GetHeight(), Source.GetChannelCount(),
         Source.GetStorageType());
  
  memcpy(GetImageData() , Source.GetImageData(), 
         Source.GetPixelCount() * Source.GetDepth() * 
         Source.GetChannelCount());

  ColorModel_   = Source.GetColorModel();
  ChannelCount_ = Source.GetChannelCount();
  Depth_        = Source.GetDepth();
  BitDepth_     = Source.GetBitDepth();
  Width_        = Source.GetWidth();
  Height_       = Source.GetHeight();
  PixelCount_   = Source.GetPixelCount();
  WidthStep_    = Source.WidthStep_;
  InterleavedDataOrder_ = ! Source.IsPlanar();
  _UID          = Source.GetUID();

  Source.GetROI(ROIUpperLeftX_, ROIUpperLeftY_, ROILowerRightX_, 
                ROILowerRightY_);
  _MetaData=Source._MetaData;
  return *this;
}


void ImageBase::CopyIn_NoInit(void *data)
{
  BIASDOUT(D_COPYIN," entering CopyIn_NoInit()");
  memcpy(GetImageData(), data, GetPixelCount()*GetChannelCount()*GetDepth());  
}


int ImageBase::Flip() 
{
  int size=GetPixelCount() * GetChannelCount() * GetDepth(); // in byte
  if (size<=0) return -1;
  
  char *newData = new char[size];
  register unsigned int length =  GetWidth() * GetChannelCount() * GetDepth();
  if (InterleavedDataOrder_) {
    register char *targetData = newData + size - length;
    for (register unsigned int i=0; i < Height_; i++) {
      memcpy(targetData,(char*)(ImageDataArray_[i]),length);
      targetData -= length;
    }    
  }    
  else { // planar
    BIASERR("Flip not implemented for planar images.");
    /*
    register int rowSize=Width_*sizeof(StorageType);  
    register StorageType *targetData = newData+
      (Height_-1)*rowSize;
    register StorageType *sourceData = ImageData_;    
    for (int channel=0;channel<ChannelCount_;channel++) {      
      for (register int i=0; i < Height_; i++) {      
        memcpy(targetData,sourceData,rowSize);
        targetData-=rowSize;      
        sourceData+=rowSize;      
      }    
      targetData+=(2*Height_-1)*rowSize;      
      } */   
  }
  ReleaseImageDataPointer();
  RedirectImageDataPointer((void*)newData);
  return 0;
}

int ImageBase::FlipHorizontal() 
{
  int size=GetPixelCount() * GetChannelCount() * GetDepth(); // in byte
  if (size<=0) return -1;
  int pixelsize=GetChannelCount() * GetDepth();
  char *newData = new char[size];
  if (InterleavedDataOrder_) {
    register char *targetData;
    register char *srcData;
    for (register unsigned int i=0; i < Height_; i++) {
      targetData = newData + (i+1)*WidthStep_ - pixelsize;
      srcData = (char *)(ImageDataArray_[i]);
      for (register int x=0; x<(int)Width_; x++){
        memcpy(targetData-x*pixelsize, srcData+x*pixelsize, pixelsize);
      }
    }    
  } else { // planar
    BIASERR("FlipHorizontal not implemented for planar images.");
  }
  ReleaseImageDataPointer();
  RedirectImageDataPointer((void*)newData);
  return 0;
}




void ImageBase::PrintHeader(ostream& os) const
{
  os << "\nVersion_ : \t\t" << Version_;
  os << "\nStorageType_ : \t\tImageBase::"<<StorageType_;
  os << "\nColorModel_ : \t\tImageBase::"<< ColorModel_
     << " ("<<(int)ColorModel_<<")"
     << "\nWidth_: \t\t" << Width_
     << "\nHeight_: \t\t" << Height_
     << "\nnChannelCount_: \t" << ChannelCount_
     << "\nWidthStep_: \t\t" << WidthStep_
     << "\nDepth_: \t\t" << Depth_
     << "\nBitDepth_: \t\t" << BitDepth_
     << "\nPixelCount_: \t\t" << PixelCount_
     << "\nInterleavedDataOrder_: \t" << InterleavedDataOrder_
     << "\nROI: \t\t\t("<<ROIUpperLeftX_<<", "<<ROIUpperLeftY_<<") <--> ("
     <<      ROILowerRightX_<<", "<<ROILowerRightY_<<")"
     << "\nUUID: \t\t\t"<<GetUID()<<" ("
     <<      (GetUID().IsValid()?"valid":"invalid")<<")"
#ifdef BIASDEBUG
     << "\nDebuglevel_:\t" << Debuglevel_;
#else
  ;
#endif
}

std::ostream& BIAS::operator<<(std::ostream& os, 
                               const ImageBase::EStorageType &st){
  switch (st){
  case ImageBase::ST_invalid: os<< "ST_invalid"; break;
  case ImageBase::ST_unsignedchar: os<< "ST_unsignedchar"; break;
  case ImageBase::ST_char: os<< "ST_char"; break;
  case ImageBase::ST_unsignedshortint: os<< "ST_unsignedshortint"; break;
  case ImageBase::ST_shortint: os<< "ST_shortint"; break;
  case ImageBase::ST_unsignedint: os<< "ST_unsignedint"; break;
  case ImageBase::ST_int: os<< "ST_int"; break;
  case ImageBase::ST_float: os<< "ST_float"; break;
  case ImageBase::ST_double: os<< "ST_double"; break;
  default: os<<"unknown StorageType"; break;
  }
  return os;
}
std::ostream& BIAS::operator<<(std::ostream& os, 
                               const ImageBase::EColorModel &st){
  switch (st){
  case ImageBase::CM_invalid: os<< "CM_invalid"; break;
  case ImageBase::CM_Grey: os<< "CM_Grey"; break;
  case ImageBase::CM_RGB: os<< "CM_RGB"; break;
  case ImageBase::CM_BGR: os<< "CM_BGR"; break;
  case ImageBase::CM_YUYV422: os<< "CM_YUYV422"; break;
  case ImageBase::CM_UYVY422: os<< "CM_UYVY422"; break;
  case ImageBase::CM_YUV420P: os<< "CM_YUV420P"; break;
  case ImageBase::CM_YUV444: os<< "CM_YUV444"; break;
  case ImageBase::CM_YUV411: os<< "CM_YUV411"; break;
  case ImageBase::CM_HSV: os<< "CM_HSV"; break;
  case ImageBase::CM_HSI: os<< "CM_HSI"; break;
  case ImageBase::CM_DV: os<< "CM_DV"; break;
  case ImageBase::CM_RGBA: os<< "CM_RGBA"; break;
  case ImageBase::CM_Bayer_RGGB: os<< "CM_Bayer_RGGB"; break;
  case ImageBase::CM_Bayer_GBRG: os<< "CM_Bayer_GBRG"; break;
  case ImageBase::CM_Bayer_GRBG: os<< "CM_Bayer_GRBG"; break;
  case ImageBase::CM_Bayer_BGGR: os<< "CM_Bayer_BGGR"; break;
  case ImageBase::CM_HSL: os << "CM_HSL"; break;
  case ImageBase::CM_hsL: os << "CM_hsL"; break;
  default: os<<"unknown ColorModel"; break;
  }
  return os;
}


// JW 09/2003
bool
ImageBase::PowerOfTwoSize() const {
  //unsigned int w2, h2;
  //return PowerOfTwoSize(w2,h2);
  int w=(int)GetWidth();
  int h=(int)GetHeight();
  // JW 09/03: more efficient ist this with bitmasks :-)
  return (((w&(w-1))==0) && ((h&(h-1))==0));
}


// Jan Woetzel
bool
ImageBase::PowerOfTwoSize(unsigned int & width2, unsigned int & height2) const {


  // HACK.
  // TODO
  /// \todo faster with shifts, like (JW): (but not yet tested): 
  ///  unsigned int pow2Width = 1;
  ///  while (pow2Width < GetWidth()) pow2Width <<= 1;
  ///  return pow2Width;
  /// workaround: slower but works,too (JW): 
  /// compute next power of two exponent for width: 
  /// compute dual logarithm of size and round up: 
  int exponent = (int)ceil( log((double)GetWidth())/log(2.0) ); // ld
  // compute 2^exponent
  width2 = (unsigned int)ldexp( 1.0, exponent );
  
  // same for height:
  exponent = (int) ceil( log((double)GetHeight())/log(2.0) );
  height2 = (unsigned int)ldexp( 1.0, exponent );
  return ( (GetWidth()==width2) && (GetHeight()==height2));
}


// JW
bool ImageBase::DimensionMatch(const BIAS::ImageBase & other) const {
  // return treu of width and height matches. don't are about types or channel count
  return (  (this->GetWidth() == other.GetWidth())
	    && (this->GetHeight() == other.GetHeight()) );
}

 


void *ImageBase::PixelValueBase(unsigned x, unsigned y, unsigned channel)
{
  BIASERR("warnig, PixelValueBase is deprecated and will be removed in the "
          "future,\nuse GetImageDataArray() instead!");
  if (InterleavedDataOrder_)
    return (void *)(&((char*)ImageData_)[(y*WidthStep_*Depth_+
                                          x*Depth_*ChannelCount_+
                                          channel)]);
  else 
    return (void *)(&((char*)ImageData_)[channel*Width_*Height_*Depth_+
                                         y*Width_*Depth_+x*Depth_]);
}


 ostream& BIAS::operator<<(ostream& os,const ImageBase& img2)
{
  // to use reinterpret_cast<> we first have to unconst:
  ImageBase &img = const_cast<ImageBase&>(img2);
  //enum BIAS::ImageBase::EStorageType storageType = img.GetStorageType();
  int Version = img.GetVersionNumber();
  int depthondisk;
  // added in Version 300 to identify mipimage on disk similar to magic number
  os.write(IDENTIFIER, IDENTIFIER_LENGTH); 
  os.write(reinterpret_cast<char*>(&img.StorageType_), sizeof(BIAS::ImageBase::EStorageType));
  os.write(reinterpret_cast<char*>(&Version), sizeof(int));
  os.write(reinterpret_cast<char*>(&img.ChannelCount_), sizeof(int));
  os.write(reinterpret_cast<char*>(&img.Width_), sizeof(int));
  os.write(reinterpret_cast<char*>(&img.Height_), sizeof(int));
  os.write(reinterpret_cast<char*>(&img.ColorModel_), sizeof(BIAS::ImageBase::EColorModel));
  depthondisk = img.Depth_ * 8;
  os.write(reinterpret_cast<char*>(& depthondisk), sizeof(int));
  os.write(reinterpret_cast<char*>(&img.BitDepth_), sizeof(int));
  os.write(reinterpret_cast<char*>(&img.InterleavedDataOrder_), sizeof(bool));
  os.write(reinterpret_cast<char*>(&img.PixelCount_), sizeof(int));
  os.write(reinterpret_cast<char*>(&img.WidthStep_), sizeof(int));
  os.write(reinterpret_cast<char*>(&img.ROIUpperLeftX_), sizeof(unsigned int));
  os.write(reinterpret_cast<char*>(&img.ROIUpperLeftY_), sizeof(unsigned int));
  os.write(reinterpret_cast<char*>(&img.ROILowerRightX_), sizeof(unsigned int));
  os.write(reinterpret_cast<char*>(&img.ROILowerRightY_), sizeof(unsigned int));
 
  // version 311: now writing unique image id as a string,
  // make sure that id is valid, since IsValid_ is not written
#ifdef BIAS_DEBUG
  if (!img.GetUID().IsValid()) BIASERR("Writing invalid UID! Generate a valid "
                                       <<"one before writing image !");
#endif
  std::string id;
  img._UID.GetString(id);
  os.write(id.c_str(), 36);

  // finally write image data:
  os.write(reinterpret_cast<char*>(img.GetImageData()), img.GetPixelCount() * 
           img.GetChannelCount() * img.Depth_);
  return os;
}




 std::istream& BIAS::operator>>(std::istream& is, ImageBase& img)
{
  enum BIAS::ImageBase::EStorageType FileStorageType;
  int Version;
  int Width, Height, ChannelCount;
  char identifier[IDENTIFIER_LENGTH];
  char cBuffer[37];
  bool roidefined;

  is.read(identifier, IDENTIFIER_LENGTH);
  if (identifier[0] != 'M' || identifier[1] != 'I' || identifier[2] != 'P'){
    BIASERR("no identifier in image, maybe old version or no MIP image"
            <<"\ntry mipconvertversion in Image/test");
    is.setstate(std::ios::badbit);
    return is;
  }
  if ( ! img.IsEmpty() )
    img.Release();

  is.read(reinterpret_cast<char*>(&FileStorageType),sizeof(BIAS::ImageBase::EStorageType));
  is.read(reinterpret_cast<char*>(&Version), sizeof(int));
  switch (Version){
  case 313: // removed ROIIsDefined_
    //  std::cout <<"Opening file with version MIP311"<<std::endl;
      is.read(reinterpret_cast<char*>(&ChannelCount), sizeof(int));
      is.read(reinterpret_cast<char*>(&Width), sizeof(int));
      is.read(reinterpret_cast<char*>(&Height), sizeof(int));
      img.Init(Width, Height, ChannelCount, FileStorageType);
      is.read(reinterpret_cast<char*>(&img.ColorModel_), sizeof(BIAS::ImageBase::EColorModel));
      is.read(reinterpret_cast<char*>(&img.Depth_), sizeof(int));
      img.Depth_ /= 8;
      is.read(reinterpret_cast<char*>(&img.BitDepth_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.InterleavedDataOrder_), sizeof(bool));
      is.read(reinterpret_cast<char*>(&img.PixelCount_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.WidthStep_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftY_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightY_), sizeof(unsigned int));
      // read uid:
      is.read(cBuffer, 36);
      cBuffer[36]=0;
      img._UID.SetFromString(cBuffer);
            
      // read image data
      is.read(reinterpret_cast<char*>(img.ImageData_), 
              img.PixelCount_ * img.ChannelCount_ * img.Depth_);
      break;
  case 312: // removed ROIIsDefined_
    //  std::cout <<"Opening file with version MIP311"<<std::endl;
      is.read(reinterpret_cast<char*>(&ChannelCount), sizeof(int));
      is.read(reinterpret_cast<char*>(&Width), sizeof(int));
      is.read(reinterpret_cast<char*>(&Height), sizeof(int));
      img.Init(Width, Height, ChannelCount, FileStorageType);
      is.read(reinterpret_cast<char*>(&img.ColorModel_), sizeof(BIAS::ImageBase::EColorModel));
      is.read(reinterpret_cast<char*>(&img.Depth_), sizeof(int));
      img.Depth_ /= 8;
      is.read(reinterpret_cast<char*>(&img.InterleavedDataOrder_), sizeof(bool));
      is.read(reinterpret_cast<char*>(&img.PixelCount_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.WidthStep_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftY_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightY_), sizeof(unsigned int));
      // read uid:
      is.read(cBuffer, 36);
      cBuffer[36]=0;
      img._UID.SetFromString(cBuffer);
            
      // read image data
      is.read(reinterpret_cast<char*>(img.ImageData_), img.PixelCount_ * img.ChannelCount_
              * img.Depth_);
      break;
  case 311: // solved stack corruption problem due to wrong id writing/loading
    //  std::cout <<"Opening file with version MIP311"<<std::endl;
      is.read(reinterpret_cast<char*>(&ChannelCount), sizeof(int));
      is.read(reinterpret_cast<char*>(&Width), sizeof(int));
      is.read(reinterpret_cast<char*>(&Height), sizeof(int));
      img.Init(Width, Height, ChannelCount, FileStorageType);
      is.read(reinterpret_cast<char*>(&img.ColorModel_), sizeof(BIAS::ImageBase::EColorModel));
      is.read(reinterpret_cast<char*>(&img.Depth_), sizeof(int));
      img.Depth_ /= 8;
      is.read(reinterpret_cast<char*>(&img.InterleavedDataOrder_), sizeof(bool));
      is.read(reinterpret_cast<char*>(&img.PixelCount_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.WidthStep_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftY_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightY_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&roidefined), sizeof(bool));
      if (!roidefined){
        img.ROIUpperLeftX_=img.ROIUpperLeftY_=0;
        img.ROILowerRightX_=img.Width_; img.ROILowerRightY_=img.Height_;
      }
      // read uid:
      is.read(cBuffer, 36);
      cBuffer[36]=0;
      img._UID.SetFromString(cBuffer);
            
      // read image data
      is.read(reinterpret_cast<char*>(img.ImageData_), img.PixelCount_ * img.ChannelCount_
              * img.Depth_); 
      break;
  case 310: // introduced unique image id in this version
    //   std::cout <<"Opening file with version MIP310"<<std::endl;
      is.read(reinterpret_cast<char*>(&ChannelCount), sizeof(int));
      is.read(reinterpret_cast<char*>(&Width), sizeof(int));
      is.read(reinterpret_cast<char*>(&Height), sizeof(int));
      img.Init(Width, Height, ChannelCount, FileStorageType);
      is.read(reinterpret_cast<char*>(&img.ColorModel_), sizeof(BIAS::ImageBase::EColorModel));
      is.read(reinterpret_cast<char*>(&img.Depth_), sizeof(int));
      img.Depth_ /= 8;
      is.read(reinterpret_cast<char*>(&img.InterleavedDataOrder_), sizeof(bool));
      is.read(reinterpret_cast<char*>(&img.PixelCount_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.WidthStep_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftY_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightY_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&roidefined), sizeof(bool));
      if (!roidefined){
        img.ROIUpperLeftX_=img.ROIUpperLeftY_=0;
        img.ROILowerRightX_=img.Width_; img.ROILowerRightY_=img.Height_;
      }
      // do not read uid:
      { /// \bug fake reading of uuid because of bug, id is DISABLED !
        char *tmp=new char[sizeof(UUID)];
        is.read(tmp, sizeof(UUID));
        delete[] tmp;
        BIASERR("Cannot read image id in MIP version 310.");
      }
      
      // read image data
      is.read(reinterpret_cast<char*>(img.ImageData_), img.PixelCount_ * img.ChannelCount_
              * img.Depth_); 
      break;
    case 300:
      //std::cout <<"Opening file with version MIP300"<<std::endl;
      is.read(reinterpret_cast<char*>(&ChannelCount), sizeof(int));
      is.read(reinterpret_cast<char*>(&Width), sizeof(int));
      is.read(reinterpret_cast<char*>(&Height), sizeof(int));
      img.Init(Width, Height, ChannelCount, FileStorageType);
      is.read(reinterpret_cast<char*>(&img.ColorModel_), sizeof(BIAS::ImageBase::EColorModel));
      is.read(reinterpret_cast<char*>(&img.Depth_), sizeof(int));
      img.Depth_ /= 8;
      is.read(reinterpret_cast<char*>(&img.InterleavedDataOrder_), sizeof(bool));
      is.read(reinterpret_cast<char*>(&img.PixelCount_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.WidthStep_), sizeof(int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROIUpperLeftY_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightX_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&img.ROILowerRightY_), sizeof(unsigned int));
      is.read(reinterpret_cast<char*>(&roidefined), sizeof(bool));
      if (!roidefined){
        img.ROIUpperLeftX_=img.ROIUpperLeftY_=0;
        img.ROILowerRightX_=img.Width_; img.ROILowerRightY_=img.Height_;
      }
//        img.PrintHeader(cerr);
//         cerr << endl << "size to read "<< img.PixelCount_ * img.ChannelCount_ *  img.Depth_<< endl;
      is.read(reinterpret_cast<char*>(img.ImageData_), img.PixelCount_ * img.ChannelCount_
	      * img.Depth_); 
      break;
    default:
      BIASERR("operator>> invalid or unsupported version number"
	     <<"\ntry mipconvertversion in Image/test");
      
      break;
  }
  return is;
}
