/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _IMAGE_BASE_H_INCLUDED_
#define _IMAGE_BASE_H_INCLUDED_
#include "bias_config.h" 
#include <exception>
#include <string>

#include <malloc.h>
#include <string.h>
#include <stdio.h>

#include <Base/Debug/Debug.hh>
#include <Base/Debug/Error.hh>
#include "MetaData.hh"
#include "UUID.hh"

#define UNIQUE_IMAGE_ID BIAS::UUID
#define DEF_UNIQUE_IMAGE_ID BIAS::UUID()


// Begin namespace BIAS
namespace BIAS {

#define MIP_DV_WIDTH 500
#define MIP_DV_HEIGHT 288


#define D_COPYIN        1<<1
#define D_CONVERT       1<<2 
#define D_CAST_CHECK    1<<3
#define D_MI_TOPLANAR   1<<4
#define D_MI_IOSTREAM   1<<5
#define D_MI_CONVGREY   1<<6

// global debug level
#define D_IMAGE_TRACE 1<<25

#define VERSION_NUMBER 313

#define IDENTIFIER "MIP"
#define IDENTIFIER_LENGTH 3

  class ImageConvert;
  template<class StorageType> class Image;

  /** @class ImageBase 
      @ingroup g_image_base
      \brief This is the base class for images in BIAS. 
      
      It is a non template class, so you can write image type 
      independet functions.
      Use this class if you deal with different storage types or
      you don't know the storage type, eg. when loading an image from disk.	
  */ 
  class DLLEXPORTAPI ImageBase : public Debug 
  {
    friend class ImageConvert;
    friend class ImageIO;
  public:      
    enum EStorageType {
      ST_invalid =-1,       ///< invalid not set image storage type 
      ST_unsignedchar,     ///< (8bit) unsigned char image storage type
      ST_char,             ///< (8bit) signed char image storage type
      ST_unsignedshortint, ///< (16bit) unsigned integer image storage type
      ST_shortint,         ///< (16bit) signed integer image storage type
      ST_unsignedint,      ///< (32bit) unsigned integer image storage type
      ST_int,              ///< (32bit) signed integer image storage type
      ST_float,            ///< float image storage type
      ST_double            ///< double image storage type
    };
    
    /** @enum EColorModel
        \brief These are the most often used color models. 
    */
    enum EColorModel 
      {
        CM_invalid = -1, ///< invalid (not set) image format
        CM_Grey,        ///< gray values (one channel: gray value)
        CM_RGB,         ///< color values (three channel order: red,green,blue)
        CM_BGR,         ///< color values (three channel order: blue,green,red)
        CM_YUYV422,     ///< YUYV422 (please insert a description)
        CM_UYVY422,     ///< (please insert a description)
        CM_YUV420P,     ///< (please insert a description)
        CM_YUV444,      ///< (please insert a description)
        CM_YUV411,      ///< (please insert a description)
        CM_HSV,         ///< HSV (three channel order: hue, sat , value)
        CM_HSI,         ///< HSI (three channel order: hue, sat , intensity)
        CM_DV,          ///< DV (please insert a description)
        CM_RGBA,        ///< RGBA (four channel order: red,green,blue,alpha)
        CM_GreyA,       ///< Grey  (or Gray) plus Alpha channel
        CM_Bayer_RGGB,  ///< One channel RGB image Bayer tile
        CM_Bayer_GBRG,  ///< One channel RGB image Bayer tile
        CM_Bayer_GRBG,  ///< One channel RGB image Bayer tile
        CM_Bayer_BGGR,  ///< One channel RGB image Bayer tile
        CM_HSL,         ///< similar to HSV but space is a double tipped cone
        CM_hsL,         ///< similar to HSL but euclidean (h,s) for CNCC
        CM_SymTensor2x2,///< The image contains a 2x2 symmetric tensor
        CM_BGRA        ///< BGRA (four channel order: blue,green,red,alpha)
      };

    ImageBase();
    ~ImageBase();

    ImageBase(const ImageBase& im);

    /** @brief Constructor, initializing image size and channels       
        @author Felix Woelk
        @status tested
        @param width  image width  of the new image
        @param height image height of the new image
        @param channels number of color planes      
        @note Example: Defining an RGB image.
        @verbatim
        ImageBase im; 
        im.init(width, height, 3);
        @endverbatim **/
    ImageBase(unsigned int width, unsigned int height, unsigned int channels=1,
              enum EStorageType storageType=  ST_unsignedchar, 
              const bool interleaved=true);
    
    /**  @brief Initialize image size and channels
         
         @warning call this only for empty images,
         calls new for the image data array. Overwrites the color model with
         some kind of best guess
         @author Felix Woelk
         @status tested
         @param width  image width  of the new image
         @param height image height of the new image
         @param channels number of color planes **/
    void Init(unsigned int width, unsigned int height, unsigned int channels=1,
              enum EStorageType storageType = ST_unsignedchar,
              const bool interleaved=true);
  
    /** @brief check if ImageData_ points to allocated image buffer or not */
    inline bool IsEmpty() const
    { return (ImageData_==NULL); }; 
        
    /** @brief Take some data and fill it into the Image 
        @warning Image must be initialized already !
        @author Evers  **/
    void CopyIn_NoInit(void *data);

    /** @brief Free the allocated data structures*/
    void Release();

    /** @brief steals the image data array from source,
        after releasing the actual image data 
        and sets source image data to NULL

        @warning Important for template: Steal images only, if u know the 
        storageTypes are euqal!
        @author Daniel Grest
        @status tested */
    int StealImage(ImageBase &source);

    /////////////////////Get Functions ///////////////////////////////////////
    /** @name Get functions
        @{
    */        
    inline void* GetImageData() const
    { return ImageData_; }; 
    
    inline void* GetImageData()
    { return ImageData_; };

    /** @brief Get an array of pointers to image data. 
        
        A pixel can be accessed  by 
        GetImageDataArray()[row][col] or GetImageDataArray()[y][x] <br>
        For multi channel images with interleaved data order, access with 
        GetImageDataArray()[row][col * GetChannelCount() + Channel] <br>
        For multi channel images with planar data order, access with 
        GetImageDataArray()[row + GetHeight * Channel][col] <br>
        
        @author Felix Woelk
        @return array with pointer to the first pixels in every row. **/
    inline void** GetImageDataArray() const
    { return ImageDataArray_; }; 
 
    inline unsigned int GetWidth() const
    { return Width_; }; 

    inline unsigned int GetHeight() const
    { return Height_; };
    
    /** returns the bytes per channel, which is the sizeof(StorageType) */
    inline unsigned int GetDepth() const 
    { return Depth_; };

    /** returns the bits per channel, which is not necessairily 
        8*sizeof(StorageType).
        @author woelk 12/2004 */
    inline unsigned int GetBitDepth() const 
    { return BitDepth_; };

    /** returns the image size in bytes = count - NOT the dimension */
    inline unsigned int GetSize() const
    { return PixelCount_ * Depth_ * ChannelCount_; };

    
    /** @return true if teh image dimension w x h of thsi and other 
        match = are equal
        @author JW */
    bool DimensionMatch(const BIAS::ImageBase & other) const;    

    
    /** @brief returns the number of Color channels, e.g. RGB 3
 
    the value returned is independent of the StorageType */
    inline unsigned int GetChannelCount() const
    { return ChannelCount_; };

    /** @brief returns the number of bytes per line
        
    returns the number of bytes per line, not the number of StorageTypes<br>   
    interleaved data order: = Width * Depth * ChannelCount 
    planar data order: = Width * Depth 
    */
    inline unsigned int GetWidthStep() const
    { return WidthStep_; };

    inline enum EColorModel GetColorModel() const
    { return ColorModel_; };

    inline enum EStorageType GetStorageType() const 
    { return StorageType_; };

    /** @brief returns number of pixels in image */
    inline unsigned long int GetPixelCount() const
    { return PixelCount_; }; 
     
    /** 
	returns the Position (index from array start) in ImageData 
	Warning: until 2005-01-04 the position inbyte was returned
	but used as position in <StorageType> (ImageDraw).
	Changed this method to return position in StorageType
	Jan-Friso Evers
     */
    inline unsigned int GetPixelPosition(unsigned int x, unsigned int y, 
                                         unsigned short int channel=0) const;
  
    /** copies one specific channel from source to Image
        can only be called from an planar image, so call ToPlanar() befor
        so far only implemented for CM_YUYV422, BGR  and RGB        
        @author Felix Woelk
        @status untested (17/04/2002) */
    int GetChannel(ImageBase &source, unsigned int channel);
     
    /** @brief returns the UUID of the image */
    inline const UNIQUE_IMAGE_ID& GetUID() const
    {  return _UID; };

    inline MetaData *GetMetaData()
    { return &_MetaData; };

    inline void SetMetaData(const MetaData& m)
    { _MetaData = m; };

    inline int GetVersionNumber() const 
    { return Version_; };

    inline bool IsPlanar() const 
    { return (! InterleavedDataOrder_); };
    
    inline bool IsInterleaved() const
    { return (InterleavedDataOrder_); };
  
    //@}       
    

    /** @name Set functions
        @{
    */
    inline void SetColorModel(EColorModel Model)
    { ColorModel_ = Model; };

    inline void SetInterleaved(bool interleaved) 
    { InterleavedDataOrder_=interleaved;UpdateImageDataArrayFromImageData_();};

    /// needed by ImageIO
    inline void SetBitDepth(unsigned bitdepth) 
    { BIASASSERT(bitdepth<=(8*Depth_)); BitDepth_=bitdepth; };

    inline void SetUID(const UNIQUE_IMAGE_ID& id)
    { _UID = id; };
    
    /** @brief sets the image's uid to invalid */
    inline void InvalidateUID()
    { _UID.Invalidate(); };

    //@} 


    /** @name Region of Interest Functions
        Important: the ROI has a size of (lrx-ulx)*(lry-uly) that means
        the lower right point is not part of the ROI!
        @{
    */    
    /** @brief Sets a region of interest
        @author Ingo Thomsen
        @status tested */
    int SetROI(unsigned int UpperLeftX, unsigned int UpperLeftY, 
               unsigned int LowerRightX, unsigned int LowerRightY);

    /** @brief Delete a region of interest
        @author Ingo Thomsen
        @status tested **/
    inline void UnsetROI() 
    {  ROIUpperLeftX_=ROIUpperLeftY_=0;
       ROILowerRightX_=Width_; ROILowerRightY_=Height_; }
  
    /** @brief Return the location of a region of interest, 
        by saving the coordinates
        within the variables defined by the parameters
        @author Ingo Thomsen, woelk
        @status tested **/    
    inline void GetROI(unsigned int& UpperLeftX, unsigned int& UpperLeftY,
                  unsigned int& LowerRightX, unsigned int& LowerRightY) const;
    inline void GetROI(int& UpperLeftX, int& UpperLeftY,
                       int& LowerRightX, int& LowerRightY) const;  

    /** prints the ROI to the given output stream */
    void PrintROI(std::ostream& os=std::cout) const; 

    /** @brief returns a copy of ROI (a new image) lower right point excluded,
        only interleaved images !
        @author Felix Woelk, grest
        @status untested **/
    int GetCopyOfROI(ImageBase& copy) const;

    /** @brief like GetCopyOfROI, but with lower right point included
        only interleaved images !
        @note GetCopyOfROI returns an image where the lower right ROI point
        is not included. This version does it. Correction are not made in
        the original method, because extensive corrections would become
        necessary in other methods.
        @author Ingo Thomsen  **/
    int GetCopyOfROI2(ImageBase& copy) const;


    /** @brief paste Image to current ROI
        @author Felix Woelk 02 2003 **/
    int Paste2ROI(const ImageBase& Image);

    /** @brief reduces image to current ROI,
        !!! image size changes !!!
        @author Felix Woelk
        @status untested
        @todo Make it work on colored Images **/
    int Cut2ROI();
    
    /** @brief sets all pixel not in ROI to zero
        @author woelk 10/2004 */
    void SetOutsideROIZero();


    inline const unsigned int& GetROIUpperLeftX() const 
    { return ROIUpperLeftX_; }
        
    inline const unsigned int& GetROIUpperLeftY() const 
    { return ROIUpperLeftY_; }
    
    inline const unsigned int& GetROILowerRightX() const 
    { return ROILowerRightX_; }

    inline const unsigned int& GetROILowerRightY() const 
    { return ROILowerRightY_; }
    //@}

    /** @name Operators
        @{
    */
    /** @brief assignment operator, 
        allocates memory structure via Init only if necessary
        @author Felix Woelk **/				
    ImageBase& operator=(const ImageBase& Source);

    /** @brief comparison operator, 
        based upon the adresses of the actual image data
        @author Ingo Thomsen
        @status tested **/
    inline bool operator== (const ImageBase &img) 
    { return (GetImageData() == img.GetImageData()); }
    
     friend 
    std::ostream& operator<<(std::ostream& os, const ImageBase &img);
    
     friend 
    std::istream& operator>>(std::istream& is, ImageBase& img);
    
    //@} 
    /** @name Misc functions
        @{
    */
    /** @brief flips the image vertically (row order is inverted)
        In place function
        return 0 in case of success, -1 otherwise
        @author Daniel Grest
        @date Jun 2002		
        @status Interleaved tested, planar untested
        Aug 2002: Jan-Friso Evers: fixed interleaved for float and other */
    int Flip();

    /** @brief flips the image horizontal (column order is inverted)
        In place function
        return 0 in case of success, -1 otherwise
        @author woelk 10/2004 */
    int FlipHorizontal();

    /** @brief checks if data area has same "size" as Image of other type 
     */
    inline bool SamePixelAndChannelCount(const ImageBase& Image) const;

    /** @brief Releases ImageData_ 
        (to be used together with RedirectImageDataPointer)
        @author Felix woelk
        @status untested (17/04/2002) */
    inline void ReleaseImageDataPointer();

    /**  @brief This method takes data and set the internal 
         image data pointer to this.

         You should really know what you are doing using this method.
         Don't say you haven't been warned.
         @author Friso Evers
         @status untested **/
    inline void RedirectImageDataPointer(void *data) 
    {  ImageData_ = data; UpdateImageDataArrayFromImageData_(); };

    /** @return true if image size is a power of two JW */
    bool PowerOfTwoSize() const;
    
    /** @return true if both, width and height are already a power of two.
        The next power of two size >= actual size is computed and stored 
        in args.
        useful for OpenGL texture enlargement for 'power of two' textures.
        @author Jan Woetzel 09/2003 */
    bool PowerOfTwoSize(unsigned int & width2, unsigned int & height2) const;
    
    void *PixelValueBase(unsigned x, unsigned y, unsigned channel=0);
    
    void PrintHeader(std::ostream& os) const;
    
    /// zeroes the image
    inline void SetZero();

    /** @brief check if image contains that pixel position */
    inline bool IsPositionInImage(const int& x, const int& y) const;

    /** @brief find closest valid pixel position to x,y 
        @author koeser 01/2005 
        @return 0=was correct 3,4=y out of range, 1,2 x out of range  */
    inline int MoveToClosestPositionInImage(int& x, int& y) const;

    // @}
  protected:    
    /** @brief recomputes pointer array returned by GetImageDataArray */
    int UpdateImageDataArrayFromImageData_();

    /// version number of imagebase class
    static const int Version_ =  VERSION_NUMBER;
    /// the color model of the pixels
    enum EColorModel ColorModel_; 
    /// the storage type in a pixel channel
    enum EStorageType StorageType_; 
    /// number of channels per pixel
    unsigned int ChannelCount_;
    /// size of one channel of one pixel in bytes
    unsigned int Depth_; 
    /// relevant bits per pixel per channel
    unsigned int BitDepth_; 
    /// image width in pixels
    unsigned int Width_; 
    /// image height in pixels
    unsigned int Height_; 
    /// planar or interleaved: planar means we have several image planes,
    /// e.g. for a 3x3 RGB image data is RRRRRRRRRGGGGGGGGGGBBBBBBBB
    /// while interleaved means RGBRGBRGBRGBRGBRGBRGBRGBRGB
    bool InterleavedDataOrder_; 
    /// number of pixels in image
    unsigned int PixelCount_;
    /// size of a line in bytes
    unsigned int WidthStep_; 
    /// region of interest (define upper left corner)
    unsigned int ROIUpperLeftX_, ROIUpperLeftY_; 
    /// lower right corner of ROI (which is NOT part of the ROI)
    unsigned int ROILowerRightX_, ROILowerRightY_; 
    /// array of pointers to the first byte in an image row
    void** ImageDataArray_; 
    /// a pointer to the image data 
    void* ImageData_;
    /// unique id for every image
    UNIQUE_IMAGE_ID _UID;
    /// additional data block, handled by derived classes
    MetaData _MetaData; 
  };

  std::ostream& operator<<(std::ostream& os, 
                           const ImageBase::EStorageType &st);
  std::ostream& operator<<(std::ostream& os, 
                           const ImageBase::EColorModel  &st);

#include "./ImageBaseInline.hh"

} // end namespace
#endif
