/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_IMAGEIO_H__
#define __BIAS_IMAGEIO_H__

#include "bias_config.h" 

//#include <unistd.h>
//#include <sys/mman.h>
#include <string>
#include "Image.hh"
#include "MetaData.hh"

#define MIP_IDENTIFIER "MIP"
#define MIP_IDENTIFIER_LENGTH 3


#define UNKNOWN_META_DATA -5

// default quality for saving e.g. jpg images from  0..100
#define BIAS_DEFAULT_IMAGE_QUALITY 80

// return code if no error occured
#define BIAS_IO_SUCCESS 0

namespace BIAS {
  /** @class ImageIO
      @ingroup g_image_io
      @brief routines for loading and writing all kinds of image formats
      
      - use Load() and ExportImage() for foreign file formats

      - for the binary mip image format use Read/Write 

      As all functions are static call them like ImageIO::Load(). */
  class DLLEXPORTAPI ImageIO {
  public:
    /** @enum TFileFormat
        @brief format specifier when writing an image to a file */
    enum TFileFormat {FF_mip, FF_ppm, FF_pgm, FF_jpg, FM_tif, FF_png,
                      FF_viff};
    
    /** @brief Write a mip binary Image to disk
        @param FileName output file name. Extension .mip is added if extension
        isn't already (case insensitive) 'mip' (JW 04/2003)
        @author Felix Woelk 
        @param sync tries to flush all open buffers/caches **/
    static int Write(const std::string& FileName, ImageBase &img, 
                     bool sync=false, bool forceNewID = true);
      
    /** @brief Write a mip binary image to disk circumventing OS caches
	as good as possible. (O_DIRECT)
	@author Jan-Friso Evers
     */
    static int WriteUnbuffered(const std::string& FileName, ImageBase &img);


    /** @brief Read a mip binary Image from disk and allocate memory structure
        @param FileName complete name of file to be loaded
        @param img object to receive the image data fro the file
        @param id The image is assigned id if there is no id in the file
        @author Felix Woelk **/
    static int Read(const std::string& FileName, ImageBase &img);
 
    /** @brief first tries a call to Read if it fails, tries to ImportImage
        @return -1 in case of error 
        @return UNKNOWN_META_DATA if meta data could not be read 
        @author Felix Woelk, Jan Woetzel **/
    static int Load(const std::string& FileName, ImageBase& img);

   
    /** @brief Try to import an image from a file. 

        Formats supported are listed in TFileFormat.
        More over a few other formats may be includeable like GIF.
        The imported image is always of the type RGB.
        @author Ingo Thomsen
        @status tested
        @date 17.06.2002
        @param assumeGrey Set it to grey, if the image to be imported is 
        certainly a grey image, to speed up loading process
        @param filename name of the file to be loaded
        @return -1 in case of error 
        @return UNKNOWN_META_DATA if meta data could not be read **/
    static int ImportImage(const std::string& filename, ImageBase &result,
                           bool assumeGrey = false);

    /** @brief Export image as file.
        @attention Do NOT scale or shift in this routine, 
                   do it by hand if you need it. (JW 04/2003) 
	    \todo add ImageMagick support in ExportImage
        @author Ingo Thomsen
        @status tested
        @date 17.06.2002
        @param FileFormat What file format shall be saved
        @param filename name of the file to be loaded
        @param jpeg_quality Quality of a jpeg image to be saved, 
               ignored in all other cases 
        @param sync does only worlk with pgm/ppm (woelk)
        @return 0 in case of no error **/
    static int ExportImage(const std::string& filename, ImageBase &img,
                           enum TFileFormat FileFormat, bool sync=false,
                           int jpeg_quality= BIAS_DEFAULT_IMAGE_QUALITY, 
                           bool forceNewID = true);

    /** @brief appends extension to name if extension is not already extension,
        avoids double extension like ".pgm.pgm"  

        @param oldName filename (maybe selected by user)
        @param newExtension desired extension
        @return a filename with extension
        @author Jan Woetzel 04/2003 */
    static std::string ExtensionName(const std::string & oldName, 
                                     const std::string & newExtension);
    


#ifdef BIAS_HAVE_IMAGEMAGICKLIB
    /** @brief Reads an image from disk using the ImageMagick++ library.
        is called automatically by ImportImage.
        @return 0 in case of succes, negative value in case of error.
        @author woelk 12/2004 */
    static int ImportMagickPP(const std::string& FileName, ImageBase& result);
#endif // BIAS_HAVE_IMAGEMAGICKLIB

#ifdef BIAS_HAVE_IMLIB
     /** @brief Reads an image from disk using the imlib library.
         encapsuled from ImportImage 10/2004
 
         !!! Deprecated, ImportMagickPP is succesor. !!!
 
         @return 0 in case of succes, negative value in case of error.
         @param filename the input filename to read from.
         @param result is teh returned read image
         @author Jan Woetzel 10/2004 */
    static int ImportImlib(const string& filename, ImageBase &result, 
                           bool assumeGrey);
#endif // BIAS_HAVE_IMLIB




#ifdef BIAS_HAVE_IMAGEMAGICKLIB
    /** @brief Reads an image from disk using the ImageMagick++ library.
        \todo Implement assumeGrey for MagickPP like for Imlib (JW).
        
        result image must be initialized. Th read image is automatically 
        converted to match the result image.

        @return 0 in case of succes, negative value in case of error.
        @param FileName the input filename to read from.
        @param img contains the read image if success.
        @param assumeGrey not implemented for now, just for interface 
        compatibility with Bias Import functions
        @param readMetaData try to read ASCII metadata from file comment.
        @author Jan Woetzel 10/2004 */
    static int ImportMagickPPAutoconvert(const std::string& FileName, 
                                         ImageBase& result,
                                         const bool &assumeGrey=false,
                                         const bool &readMetaData=false );

  /** @brief Reads an image from disk using the ImageMagick++ library.      
      \todo Implement 16bit short write with  Magick++ using BLOB.
      \todo Implement quality directive for jpg saving with Magick++.
	  \bug IMageMagick PPM files are alwyas written as ASCII instaed of binary (JW)
      @return 0 in case of succes, negative value in case of error.
      @param FileName the input filename to read from.
      @param img contains the image to be written to disk. 
      @param quality for saving, e.g.1-100 for jpeg, tested fro jpeg.
      @param writeMetaData try to write ASCII MetaData as comment to file.
      Output file type is determined automatically by file extension.
      @author Jan Woetzel 10/2004 */
    static int ExportMagickPP(const std::string& FileName,const ImageBase& img,
                              const int &quality=BIAS_DEFAULT_IMAGE_QUALITY,
                              const bool &writeMetaData=false );
#endif // BIAS_HAVE_IMAGEMAGICKLIB

  protected:
    static int ImportImagePnm_(std::ifstream& ifs, ImageBase &result,
                               int depth);
    static int ExportImagePnm_(const std::string &FileName, ImageBase &img,
                               bool sync=false);
    
    /** @brief export a one channel grey float viff image (Khoros imaeg format)
        for compatibility to old stereo code
        @param FileName destination filename
        @param img BIAS ImageBase to be written to disk 
        @author Jan Woetzel 01/2003 */ 
    static int ImportImageViff_(std::ifstream& ifs, ImageBase &result);
      

    // to WriteUnbuffered two members are needed
    static char *MemAlignedBuffer_;
    static unsigned long int MemAlignedBufferSize_;


  }; // end of class
} // end of namespace BIAS

#endif
