/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _Vector2_hh_
#define _Vector2_hh_


#include <iostream>
#include <sstream>
#include <iomanip>


//additional math constants are not used by default in WIN32 - so set usage
#ifdef WIN32
#define _USE_MATH_DEFINES
#endif

#include <math.h>


#include "bias_config.h" 


#include <Base/Debug/Error.hh>
#include "Operators.hh"

// dim of this vector
#define VECTOR2SIZE 2
namespace BIAS {

  // forward declarations to avoid mutual inclusion
  template <class T> class Vector;
  template <class T> class Matrix;
  template <class T> class Matrix2x2;
  
  template <class T> 

  /** @class Vector2
      @ingroup g_math
      @brief class Vector2 contains a Vector of dim. 2 and fixed dimension
       
      It's elment type is templated.
      manual loop unrolling is used if possible. 
      The Vector is in row-major order (2 rows, 1 column)
      the indizes begin with zero (to size-1)
      @author Jan Woetzel
      @status untested (03/04/2002) **/
  class DLLEXPORTAPI Vector2 
  {
 
  public:
    /** destructor
        @author Jan Woetzel
        @status untested (02/28/2002) **/
    ~Vector2() {};


    /**	default constructor
        @author Jan Woetzel
        @status alpha (02/28/2002) **/
    Vector2() {};

    /**	copy constructor
        @author Jan Woetzel
        @status alpha (03/05/2002)**/
    inline Vector2(const Vector2<T> &vec) {
      copy(vec.GetData());
    };

    /** assignment with a constant value for all elements
        @author Jan Woetzel
        @status alpha (03/01/2002) **/
    explicit inline Vector2(const T & scalar) {
      Set(scalar);
    };


    /** constructor with element assigment using explicit values
        @author Ingo Thomsen
        @status tested (03/04/2002) **/
    inline Vector2(const T v0, const T v1) {
      data_[0] = v0;
      data_[1] = v1;
    }


    /**	assignment with an array of values
        which is copied into this ones class members
        @author Jan Woetzel
        @status untested (02/28/2002) **/
    explicit inline Vector2(const T * pv) {
      copy(pv);
    };
    
    /**	constructor with element assignment
        @author Jan Woetzel
        @status alpha (03/01/2002) **/
    explicit inline Vector2(char *s) {
      std::istringstream inst(s);
      inst >> data_[0];
      inst >> data_[1];
    };


    /** cast constructor
        @author Jan Woetzel
        @status alpha (03/22/2002) **/
    Vector2(const Vector < T > &v);


    /** copy the array of vectorsize beginning at *T to this->data_
        @author Jan Woetzel
        @status untested (02/28/2002) **/
    inline void copy(const T * pv) {
      // better use memcopy ?
      data_[0] = pv[0];
      data_[1] = pv[1];
    };


    /** set all elements to a scalar value
        @author Jan Woetzel
        @status untested (02/28/2002) **/
    inline void Set(const T & scalar) {
      Set(scalar, scalar);
    };

    
    /** @author Jan Woetzel */
    inline void SetZero() {
      Set(0,0);
    };
    
    
    /** @author Jan Woetzel */
    inline bool IsZero() const {
      return ( (data_[0]==(T)0)
               && (data_[1]==(T)0)
               && (data_[2]==(T)0) );
    }
    
    
    /** set elementwise with given scalar values
        @author Jan Woetzel
        @status alpha (02/28/2002) **/
    inline void Set(const T & x, const T & y) {
      data_[0] = x;
      data_[1] = y;
    };


    /**	@return the size of this vector
        @author Jan Woetzel
        @status untested (02/28/2002) **/
    inline const int size() const {
      return VECTOR2SIZE;
    };


    /** get the data pointer
        the member function itself is const (before {..}) because it doesn't 
        change the this object.
        @return the const pointer to the data array for reading
        @author Jan Woetzel
        @status untested (02/28/2002) **/
    inline const T *GetData() const {
      return data_;
    };

    inline T *GetData() {
      return data_;
    };

    
    /** assignment operator
        set the vector elementwise to scalar value
        @author Jan Woetzel
        @status alpha (02/28/2002) **/
    inline Vector2<T> &operator=(const T& scalar) {
      Set(scalar);
      return *this;
    };
    
    /** assignment operator
        @author grest
    */
    inline Vector2<T> &operator=(const Vector2<T> &vec) {
      copy(vec.GetData());
      return *this;
    };

    /** access an element  of the vector with 0-based indizes.
        read only (no write)
        member funciton const because it doesn't change this object
        @author Jan Woetzel
        @status untested (03/01/2002) **/
    inline const T & operator[] (const int i) const {
      BIASASSERT(0 <= i);
      BIASASSERT(i < VECTOR2SIZE);
      return data_[i];
    };

    /** access an element  of the vector with 0-based indizes.
        write allowed
        @author Jan Woetzel
        @status untested (03/01/2002) **/
    inline T & operator[] (const int i) {
      BIASASSERT(0 <= i);
      BIASASSERT(i < VECTOR2SIZE);
      return data_[i];
    };


    /** read only access functions to use names instead of indizes
        @return the projective coordinate of the vector 
        (they are usually not euclidian!)
        @author Jan Woetzel
        @status untested (03/04/2002) **/
    inline  T & GetX() {
      return data_[0];
    };
    inline  T & GetY() {
      return data_[1];
    };
    inline  T & GetW() {
      return data_[1];
    };
    
    // const non reference of the above JW 05/2003: 
    inline  T  GetX() const {
      return data_[0];
    };
    inline  T  GetY() const {
      return data_[1];
    };
    inline  T  GetW() const {
      return data_[1];
    };
   
    
    
    /** set (write) access functions to use names instead of indizes
        @author Jan Woetzel
        @status untested (03/04/2002) **/
    inline void SetX(const T x) {
      data_[0] = x;
    };
    inline void SetY(const T y) {
      data_[1] = y;
    };
    inline void SetW(const T w) {
      data_[1] = w;
    };



    /** scalar product of two vectors, storing the result in result
        @author Jan Woetzel
        @status untested **/
    inline void ScalarProduct(const Vector2<T> &argvec, T & result) const {
      result = data_[0] * argvec[0] + data_[1] * argvec[1];
    }

    inline T ScalarProduct(const Vector2<T> &argvec) const {
      return ((data_[0] * argvec[0]) + (data_[1] * argvec[1]));
    }
    

    /** multiply two vectors elementwise, storing the result in destvec
        @author Jan Woetzel
        @status untested (03/06/2002) **/
    inline void ElementwiseProduct(const Vector2<T> &argvec,
                                   Vector2<T> &destvec) const {
      destvec[0] = data_[0] * argvec[0];
      destvec[1] = data_[1] * argvec[1];
    }


    /** Return the L1 norm: |a| + |b|
        @author Ingo Thomsen
        @date 04/11/2002
        @status untested **/
    inline T norm_L1() const {
      return (data_[0] > 0 ? data_[0] : data_[0] * -1)
        + (data_[1] > 0 ? data_[1] : data_[1] * -1);
    }
    inline T NormL1() const 
    { return norm_L1(); };

    /** Return the L2 norm: sqrt(a^2 + b^2)
        @author Ingo Thomsen
        @date 04/11/2002
        @status untested **/
    inline double norm_L2() const {
      return sqrt( (double) ( data_[0] * data_[0] + data_[1] * data_[1] ) );
    }

    inline double NormL2() const 
    { return norm_L2(); };

    /** Return the euclidian distance of 2 vectors.
        @author Birger Streckel
        @date 08/2002
    **/
    inline double Dist(const Vector2<T> &vec) const {
      return sqrt( (double) ( (data_[0]-vec[0])*(data_[0]-vec[0]) + 
                              (data_[1]-vec[1])*(data_[1]-vec[1]) ));
    }

    ///////////////////////////////////////////////////////////////////////
    ///////////////// Comparison operators using scalars //////////////////
    ///////////////////////////////////////////////////////////////////////


    /** Comparison operator 'equal'
        @author Ingo Thomsen
        @status tested **/
    inline bool operator== (const Vector2<T>& arg) const {
      return ( (data_[0] == arg.data_[0]) && (data_[1] == arg.data_[1]) );
    }

    /** Comparison operator 'not equal'
        @author Ingo Thomsen
        @status tested **/
    inline bool operator!= (const Vector2<T>& arg) const {
      return ( (data_[0] != arg.data_[0]) || (data_[1] != arg.data_[1]) );
    }



    /////////////////////////////////////////////////////////////////////
    //////////////////// Basic Operators using scalars //////////////////
    /////////////////////////////////////////////////////////////////////
    

    /** Addition (in place) of an scalar
        @author Ingo Thomsen **/
    inline void AddIP (const T& scalar) {
      Add(scalar, *this);
    }

    /** Addition with a scalar, storing results in destionation vector
        @author Ingo Thomsen **/
    inline void Add(const T& scalar, Vector2<T>& dest) const{
      dest.data_[0] = data_[0] + scalar;
      dest.data_[1] = data_[1] + scalar;
    }

    /** Substraction (in place) of an scalar
        @author Ingo Thomsen **/
    inline void SubIP (const T& scalar) {
      Sub(scalar, *this);
    }

    /** Substraction with a scalar, storing results in destionation vector
        @author Ingo Thomsen **/
    inline void Sub (const T& scalar, Vector2<T>& dest) const {
      dest.data_[0] = data_[0] - scalar;
      dest.data_[1] = data_[1] - scalar;
    }

    /** Multiplication (in place) of an scalar
        @author Ingo Thomsen **/
    inline void MultiplyIP (const T& scalar) {
      Multiply(scalar, *this);
    }

    /** Multiplication with a scalar, storing results in destionation vector
        @author Ingo Thomsen **/
    inline void Multiply (const T& scalar, Vector2<T>& dest) const {
      dest.data_[0] = data_[0] * scalar;
      dest.data_[1] = data_[1] * scalar;
    }

    /** Division (in place) of an scalar
        @author Ingo Thomsen **/
    inline void DivideIP (const T& scalar) {
#ifdef BIAS_DEBUG
      if (scalar == 0 )
        BIASERR("Division by Zero!");
#endif
      Divide(scalar, *this);
    }

    /** Division with a scalar, storing results in destionation vector
        @author Ingo Thomsen **/
    inline void Divide (const T& scalar, Vector2<T>& dest) const {
#ifdef BIAS_DEBUG
      if (scalar == 0 )
        BIASERR("Division by Zero!");
#endif
      dest.data_[0] = data_[0] / scalar;
      dest.data_[1] = data_[1] / scalar;
    }



    /////////////////////////////////////////////////////////////////////
    //////////////////// Basic Operators using vectors //////////////////
    /////////////////////////////////////////////////////////////////////

    /** in place adding
        @author Ingo Thomsen
        @status tested **/
    inline void AddIP(const Vector2<T> &argvec) {
      Add(argvec, *this);
    }


    /** adding of two vectors, storing the result in destvec
        @author Ingo Thomsen
        @status tested **/
    inline void Add(const Vector2<T> &argvec, Vector2<T> &destvec) const {
      destvec[0] = data_[0] + argvec[0];
      destvec[1] = data_[1] + argvec[1];
    }

    /** in place subtraction
        @author Ingo Thomsen
        @status tested **/
    inline void SubIP(const Vector2<T> &argvec) {
      Sub(argvec, *this);
    }


    /** subtracting of two Vectors, storing the result in destvec
        @author Ingo Thomsen
        @status tested **/
    inline void Sub(const Vector2<T> &argvec, Vector2<T> &destvec) const {
      destvec[0] = data_[0] - argvec[0];
      destvec[1] = data_[1] - argvec[1];
    }

    inline void Fill(T value)
    { data_[0] = data_[1] = value; };

  
    /** coordinate transform
        compute the euclidian coord p=(x,y) 
        for (this) given polarcoord (radius, phi) with
        radius : distance to center,
        phi    : radian angle between positive x-axis and p=(x,y)
        assume (x,z) is in orthogonal right-hand-system.
        @author Jan Woetzel 09/2003 */
    BIAS::Vector2<T> CoordPolarToEuclidian() const;
    
    
    /** coordinate transform
        compute the polar coordinates (radius, phi) for (this) given 
        eucldian point (x,y).  with
        radius : distance to center,
        phi    : radian angle between positive x-axis and p=(x,y)
        assume (x,z) is in orthogonal right-hand-system.
        @author Jan Woetzel 09/2003 */
    BIAS::Vector2<T> CoordEuclidianToPolar() const;
        
  protected:

    T data_[VECTOR2SIZE];	// the elements of the vector

  };
  // ==============================================================
  // ===  END OF CLASS Vector2  ===================================
  // ==============================================================
  



  // --------------- stream I/O ---------------------------------------

  /** Output with streams
      @author Jan Woetzel
      @status alpha (03/01/2002) **/
  template <class T>
  std::ostream& operator<<(std::ostream & os, const Vector2<T>& vec) 
  {
    /// @todo is this using namespace inside a function  OK?
    using namespace std;
#ifdef BIAS_DEBUG
    std::streambuf *OutBuf_=cout.rdbuf(), *OsBuf_= os.rdbuf(), 
      *ErrBuf_=std::cerr.rdbuf();

    // print size at first to be compatible to Vector<>
    if (OutBuf_!=OsBuf_ && ErrBuf_!=OsBuf_){
      // not cout and not cerr
#endif
      // print size at first to be compatible to Vector<>
      os << VECTOR2SIZE << endl;	
      for (int row = 0; row < VECTOR2SIZE; row++) {
        os << setw(8);		// set width for output format
        //os << setprecision(10);
        // print out one element per line
        os << vec.GetData()[row] << " " << endl;
      }
#ifdef BIAS_DEBUG
    } else {
      // cout or cerr
      os << "[ ";
      for (int row = 0; row < VECTOR2SIZE; row++)
        os << setprecision(10) << vec.GetData()[row] << " ";
      os << "] ";
    }
#endif
    return os;
  }


  /** Input with streams
      @author Jan Woetzel
      @status untested (03/01/2002) **/
  template <class T>
  std::istream& operator>>(std::istream &is, Vector2 <T>& vec) 
  {
    int N;				// first argument of the stream is the size
    is >> N;			// read size first to be compatible to Vector<>
    if (!(N == vec.size())) {
      // wrong size
      BIASERR("can't assign a Vector2<T> with a dim " << N << " vector");
    } else {
      // correct size, sequential read elementwise
      for (int row = 0; row < N; row++) {
        is >> vec.GetData()[row];
      };
    };
    return is;
  }

  // ------------------ operators with vectors---------------------------



  /** add operator for two Vectors
      @author Ingo Thomsen, tested **/
  template <class T>
  inline Vector2<T>& operator+=(Vector2<T> &vec, const Vector2<T> &argvec) {
    vec.Add(argvec, vec);
    return vec;
  }

  /** sub operator for two Vectors
      @author Ingo Thomsen
      @status tested **/
  template <class T>
  inline Vector2<T> &operator-= (Vector2<T> &vec, const Vector2<T> &argvec) {
    vec.Sub(argvec, vec);
    return vec;
  }

  // --------------- operators with scalars -----------------------------

 
  /** Addition operator with scalar argument
      @author Ingo Thomsen, tested */
  template <class T>
  inline Vector2<T>& operator+=(Vector2<T>& vec, const T& scalar) {
    vec.AddIP(scalar);
    return vec;
  }

  /** Substraction operator with scalar argument
      @author Ingo Thomsen, tested **/
  template <class T>
  inline Vector2<T>& operator-=(Vector2<T>& vec, const T& scalar) {
    vec.SubIP(scalar);
    return vec;
  }

  /** Multiplication operator with scalar argument
      @author Ingo Thomsen, tested **/
  template <class T>
  inline Vector2<T>& operator*= (Vector2<T>& vec, const T& scalar) {
    vec.MultiplyIP(scalar);
    return vec;
  }

  /** Division operator with scalar argument
      @author Ingo Thomsen, tested **/
  template <class T>
  inline Vector2<T>& operator/= (Vector2<T>& vec, const T& scalar) {
    BIASASSERT(scalar != (T)0);
    vec.DivideIP(scalar);
    return vec;
  }



} // namespace BIAS
#endif

