/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Binomial.hh"

#ifdef BIAS_DEBUG
#include <Base/Image/ImageIO.hh>
#endif

using namespace BIAS;
using namespace std;


//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////

#define FILTER_INSTANTIATION_CLASS Binomial
#include "Filterinst.hh"

//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////

template <class InputStorageType, class OutputStorageType>
Binomial<InputStorageType, OutputStorageType>::
Binomial()
  : Convolution<InputStorageType, OutputStorageType>() 
{
  // no param object !
  this->_ParamsAllocated = true;
  _HalfWinSize = new int;
  *_HalfWinSize = 1;
  //AddDebugLevel(D_BINOMIAL_KERNEL);
  _CalculateKernel(*_HalfWinSize);
}

template <class InputStorageType, class OutputStorageType>
Binomial<InputStorageType, OutputStorageType>::
Binomial(Param& para, const std::string& postfix)
  : Convolution<InputStorageType, OutputStorageType>(para, "-Binomial"+postfix)
{
  int grp = para.GetFreeGroupID();
  _AddParam(para, grp, "-Binomial"+postfix);
  if (para.IsUsedGroupID(grp)) para.SetGroupName(grp, "Binomial");
  //AddDebugLevel(D_BINOMIAL_KERNEL);
  _CalculateKernel(*_HalfWinSize);
}

template <class InputStorageType, class OutputStorageType>
Binomial<InputStorageType, OutputStorageType>::
Binomial(Param& para, const int grp, const std::string& postfix)
  : Convolution<InputStorageType,OutputStorageType>(para,grp, "-Binomial"+postfix) 
{
  _AddParam(para, grp, "-Binomial"+postfix);
  //AddDebugLevel(D_BINOMIAL_KERNEL);
  _CalculateKernel(*_HalfWinSize);
}

template <class StorageType, class KernelType>
Binomial<StorageType, KernelType>::~Binomial()
{
  if (this->_ParamsAllocated) {
    delete _HalfWinSize; _HalfWinSize = NULL;
  }
}

////////////////////////////////////////////////////////////////////////////

template <class InputStorageType, class OutputStorageType>
int Binomial<InputStorageType, OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Binomial::Filter\n");

  int res=0;
  
  switch (*_HalfWinSize){
  case 1:
    res = Filter3x3(src, dst);
    break;
  case 2:
    res = Filter5x5(src, dst);
    break;
  default:
    if ((*_HalfWinSize != _LastHalfWinSize))
      _CalculateKernel(*_HalfWinSize);
    res = Convolution<InputStorageType, OutputStorageType>::Filter(src,dst);
    break;
  }
  
  return res;
}

template <class InputStorageType, class OutputStorageType>
int Binomial<InputStorageType, OutputStorageType>::
Filter3x3(const Image<InputStorageType>& src, Image<OutputStorageType>& dst) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Binomial::Filter3x3\n");
  int res=0;
  if (src.GetChannelCount()==1 && *_FilterBorderHandling==TBH_valid){
    if (!src.SamePixelAndChannelCount(dst)) 
      if (!dst.IsEmpty()) dst.Release();
    if (dst.IsEmpty()) dst.Init(src.GetWidth(), src.GetHeight(), 
                                src.GetChannelCount());
    if (CalculationInFloat()){
      res = Filter3x3ValidGreyFloat(src, dst);
    } else {
      *_HalfWinSize=1;
      res = Filter(src, dst); 
    }
  } else {
    *_HalfWinSize=1;
    res = Filter(src, dst); 
  }
  return res;
}
  
template <class InputStorageType, class OutputStorageType>
int Binomial<InputStorageType, OutputStorageType>::
Filter5x5(const Image<InputStorageType>& src, Image<OutputStorageType>& dst) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Binomial::Filter5x5\n");
  int res=0;
  if (src.GetChannelCount()==1 && *_FilterBorderHandling==TBH_valid){
    if (!src.SamePixelAndChannelCount(dst)) 
      if (!dst.IsEmpty()) dst.Release();
    if (dst.IsEmpty()) dst.Init(src.GetWidth(), src.GetHeight(), 
                                src.GetChannelCount());
    if (CalculationInFloat()){
      res = Filter5x5ValidGreyFloat(src, dst);
    } else {
      BIASERR("unfinished");
      *_HalfWinSize=2;
      res = Filter(src, dst);
    }
  } else {
    *_HalfWinSize=2;
    res = Filter(src, dst); 
  }
  return res;
}

template <class InputStorageType, class OutputStorageType>
int Binomial<InputStorageType, OutputStorageType>::
Filter3x3ValidGreyFloat(const Image<InputStorageType>& src, 
                        Image<OutputStorageType>& dst) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Binomial::Filter3x3ValidGreyFloat\n");
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount() == 1);
  BIASASSERT(dst.SamePixelAndChannelCount(src));
#endif
  //  AddDebugLevel(D_BINOMIAL_WRITE_IM);
  const unsigned width=src.GetWidth();
  int nwidth = - int(width);
  const unsigned height=src.GetHeight();

  unsigned minx, miny, maxx, maxy;
  src.GetROI(minx, miny, maxx, maxy);
  dst.SetROI(minx+1, miny+1, maxx-1, maxy-1);

  if (width>_tmpFloat.GetWidth() || 
      height>_tmpFloat.GetHeight()){
    if (!_tmpFloat.IsEmpty()) _tmpFloat.Release();
    _tmpFloat.Init(width, height, src.GetChannelCount());
  }

  InputStorageType *im=src.GetImageData();
  InputStorageType *p, *end, *lend;
  OutputStorageType *pg, *oend, *olend;
  CONV_FLOAT *ph, tmp;
  int offset, step;

  // separated binomial 3x3 mask = 1 2 1
  offset=(int)minx+(int)miny*(int)width+1;
  p=im+offset;
  ph=_tmpFloat.GetImageData()+offset;
  step=width-(int)maxx+(int)minx+2;
  end=im+(int)maxx+((int)maxy-1)*(int)width-1;
  lend=im+(int)maxx+((int)miny*(int)width)-1;
  // vertical part
  while (p<end){
    while (p<lend){
      *ph++ = (CONV_FLOAT)(p[-1]) +(CONV_FLOAT)(*p) * 2.0f + (CONV_FLOAT)(p[1]);
      p++;
    }
    ph+=step;
    p+=step;
    lend+=width;
  }

#ifdef BIAS_DEBUG
  if (DebugLevelIsSet(D_BINOMIAL_WRITE_IM)){
    Image<InputStorageType> im=src;
    ImageIO::Write("binomial3x3-input.mip", im);
    ImageIO::Write("binomial3x3-horiz.mip", _tmpFloat);
  }
#endif

  offset=(int)minx+((int)miny+1)*(int)width+1;
  ph=_tmpFloat.GetImageData()+offset;
  OutputStorageType *s=dst.GetImageData();
  pg=s+offset;
  oend=s+(int)maxx+((int)maxy-2)*(int)width-1;
  olend=s+(int)maxx+((int)miny+1)*(int)width-1;
  CONV_FLOAT div = 1.0 / 16.0;
  // horizontal part
  while (pg<oend){
    while (pg<olend){
      tmp = ph[nwidth] + *ph * 2.0f + ph[width];
      *pg++ = (OutputStorageType)(tmp * div);
      ph++;
    }
    pg+=step;
    ph+=step;
    olend+=width;
  }
#ifdef BIAS_DEBUG
  if (DebugLevelIsSet(D_BINOMIAL_WRITE_IM)){
    ImageIO::Write("binomial3x3-result.mip", dst);
  }
#endif

  return 0;
}

template <class InputStorageType, class OutputStorageType>
int Binomial<InputStorageType, OutputStorageType>::
Filter5x5ValidGreyFloat(const Image<InputStorageType>& src, 
                        Image<OutputStorageType>& dst) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Binomial::Filter5x5ValidGrey\n");
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount() == 1);
  BIASASSERT(dst.SamePixelAndChannelCount(src));
#endif
  //AddDebugLevel(D_BINOMIAL_WRITE_IM);
  const unsigned width=src.GetWidth();
  int nwidth = - int(width);
  const unsigned height=src.GetHeight();

  unsigned minx, miny, maxx, maxy;
  src.GetROI(minx, miny, maxx, maxy);
  dst.SetROI(minx+2, miny+2, maxx-2, maxy-2);

  // init internal memory if necessairy
  if (width>_tmpFloat.GetWidth() || 
      height>_tmpFloat.GetHeight()){
    if (!_tmpFloat.IsEmpty()) _tmpFloat.Release();
    _tmpFloat.Init(width, height, src.GetChannelCount());
  }

  InputStorageType *im=src.GetImageData();
  InputStorageType *p, *end, *lend;
  OutputStorageType *pg, *oend, *olend;
  CONV_FLOAT *ph, tmp;
  int offset, step;
  // separated binomial 5x5 mask = 1 4 6 4 1
  offset=(int)minx+(int)miny*(int)width+2;
  p=im+offset;
  ph=_tmpFloat.GetImageData()+offset;
  step=width-(int)maxx+(int)minx+4;
  end=im+(int)maxx+((int)maxy-1)*(int)width-2;
  lend=im+(int)maxx+((int)miny*(int)width)-2;
  // vertical part
  while (p<end){
    while (p<lend){
      *ph++ = (CONV_FLOAT)(p[-2]) + (CONV_FLOAT)(p[-1])*4.0f + (CONV_FLOAT)(*p) 
        * 6.0f + (CONV_FLOAT)(p[+1])*4.0f + (CONV_FLOAT)(p[2]);
      p++;
    }
    ph+=step;
    p+=step;
    lend+=width;
  }

#ifdef BIAS_DEBUG
  if (DebugLevelIsSet(D_BINOMIAL_WRITE_IM)){
    Image<InputStorageType> im=src;
    ImageIO::Write("binomial5x5-input.mip", im);
    ImageIO::Write("binomial5x5-horiz.mip", _tmpFloat);
  }
#endif

  offset=(int)minx+((int)miny+2)*(int)width+2;
  ph=_tmpFloat.GetImageData()+offset;
  OutputStorageType *s=dst.GetImageData();
  pg=s+offset;
  oend=s+(int)maxx+((int)maxy-3)*(int)width-2;
  olend=s+(int)maxx+((int)miny+2)*(int)width-2;
  int dw=width<<1;
  CONV_FLOAT div = 1.0 / 256.0;
  // horizontal part
  while (pg<oend){
    while (pg<olend){
      tmp = ph[-dw] + ph[dw] + ph[nwidth]*4.0f + ph[width]*4.0f + (*ph)*6.0f;
      *pg++ = (OutputStorageType)(tmp * div);
      ph++;
    }
    pg+=step;
    ph+=step;
    olend+=width;
  }

#ifdef BIAS_DEBUG
  if (DebugLevelIsSet(D_BINOMIAL_WRITE_IM)){
    ImageIO::Write("binomial5x5-result.mip", dst);
  }
#endif

  return 0;
}

////////////////////////////////////////////////////////////////////////////

template <class StorageType, class KernelType>
void Binomial<StorageType, KernelType>::
_CalculateKernel(int HalfWinSize)
{
  Vector<FM_INT> vec;
  int shift;
  const int size=(HalfWinSize<<1)+1;
  vec.newsize(size);
  vec[0]=vec[size-1]=1;

  int numerator=size-1, denominator=1;
  for (int i=1; i<=HalfWinSize; i++){
    vec[i]=vec[size-i-1]=(numerator/denominator);
    BIASCDOUT(D_BINOMIAL_KERNEL, "num "<<numerator<<" denom "<<denominator
              <<"  =  "<<vec[i]<<endl);
    denominator*=(i+1);
    numerator*=(size-i-1);
  }
  shift=(HalfWinSize<<1);

  BIASCDOUT(D_BINOMIAL_KERNEL, "size "<<size<<" : "<< vec << "  shift "
            <<shift<<endl<< "-------------\n");
  
  _fm.Init(vec, vec, 0, shift*shift, true);
  Vector<FM_FLOAT> fvec;
  fvec.newsize(size);
  FM_FLOAT div=(FM_FLOAT)(1<<shift);
  for (int i=0; i<size; i++)
    fvec[i]=(FM_FLOAT)(vec[i]/div);
  _fm.Init(fvec, fvec, false);
 
  _LastHalfWinSize = HalfWinSize;
}

template <class StorageType, class KernelType>
void Binomial<StorageType, KernelType>::
_AddParam(Param& para, const int grp, const std::string& postfix)
{

  if (para.CheckParam("HalfWinSize"+postfix)){ 
    _HalfWinSize=para.GetParamInt("HalfWinSize"+postfix);
  } else { 
    _HalfWinSize=para.AddParamInt("HalfWinSize"+postfix, 
                                  "half size of convolution kernel", 
                                  2, 0, INT_MAX, 0, grp);
  }

}
