/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Rescale.hh"

#ifdef BIAS_DEBUG
#include <Base/Image/ImageIO.hh>
#endif

using namespace BIAS;
using namespace std;

//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////
#define FILTER_INSTANTIATION_CLASS Rescale
#include "Filterinst.hh"


//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////

template <class InputImageType, class OutputImageType>
Rescale<InputImageType, OutputImageType>::
Rescale()
  : FilterBase<InputImageType, OutputImageType>(), _gauss(), _binomial(), _mean()
{  
  _ParamsAllocated=true;
   _RescaleMeanSigmaFactor= new double;    
   *_RescaleMeanSigmaFactor=0.9;
   _RescaleFactor= new double;
   *_RescaleFactor= 2.0; 
   _RescaleMeanType= new int;
   *_RescaleMeanType = LPT_Mean;
   
}

template <class InputImageType, class OutputImageType>
Rescale<InputImageType, OutputImageType>::
Rescale(Param& para, const std::string& postfix)
  : FilterBase<InputImageType, OutputImageType>(para, "-Rescale"+postfix), 
    _gauss(para, "-Rescale"+postfix), 
    _binomial(para, "-Rescale"+postfix), 
    _mean(para, "-Rescale"+postfix)
{
  int grp=para.GetFreeGroupID();
  _AddParam(para, grp, "-Rescale"+postfix);
  if (para.IsUsedGroupID(grp)) para.SetGroupName(grp, "Rescale");
  //  AddDebugLevel(D_RESCALE_INIT);
}

template <class InputImageType, class OutputImageType>
Rescale<InputImageType, OutputImageType>::
Rescale(Param& para, const int grp, const std::string& postfix)
  : FilterBase<InputImageType, OutputImageType>(para, grp, "-Rescale"+postfix), 
    _gauss(para, grp, "-Rescale"+postfix), 
    _binomial(para, grp, "-Rescale"+postfix), 
    _mean(para, grp, "-Rescale"+postfix)
{
  _AddParam(para, grp, "-Rescale"+postfix);
  //  AddDebugLevel(D_RESCALE_INIT);
}

template <class InputImageType, class OutputImageType>
Rescale<InputImageType, OutputImageType>::~Rescale()
{
  if (this->_ParamsAllocated) {
    delete _RescaleMeanSigmaFactor;
    delete _RescaleFactor;
    delete _RescaleMeanType;
  }
}


template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
Filter(const Image<InputImageType>& src, Image<OutputImageType>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::Filter\n");
  int res=0;
  if (*_RescaleFactor<1.0){
    res=Upsample(src, dst);
  } else if (*_RescaleFactor>1.0){
    res=Downsample(src, dst);
  } else {
    dst=src;
    res=0;
  }
  return res;
}

//////////////////////////////////////////////////////////////////////////
//               downsampling functions
//////////////////////////////////////////////////////////////////////////

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
Downsample(const Image<InputImageType>& src, Image<OutputImageType>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::Downsample(src, dst)\n");
  double pow2 = log(*_RescaleFactor)/log(2.0);

  if (*_RescaleMeanType == LPT_Mean){
    unsigned newwidth, newheight;
    if (fabs(*_RescaleFactor-2.0)<=DBL_EPSILON){
      newwidth=src.GetWidth()>>1;
      newheight=src.GetHeight()>>1;
      if (dst.GetWidth()!=newwidth || dst.GetHeight()!=newheight
          || dst.GetChannelCount()!=src.GetChannelCount()){
        if (!dst.IsEmpty()) dst.Release();
        dst.Init(newwidth, newheight, src.GetChannelCount() );
      }
      return DownsampleBy2(src, dst);
    } else if (fabs(*_RescaleFactor-4.0)<=DBL_EPSILON ){
      newwidth=src.GetWidth()>>2;
      newheight=src.GetHeight()>>2;
      if (dst.GetWidth()!=newwidth || dst.GetHeight()!=newheight
          || dst.GetChannelCount()!=src.GetChannelCount()){
        if (!dst.IsEmpty()) dst.Release();
        dst.Init(newwidth, newheight, src.GetChannelCount());
      }
      return DownsampleBy4(src, dst);
    } else if ( pow2 - floor(pow2) <DBL_EPSILON ) {
      return DownsampleBPoT(src,dst);
    }
  } 
  return  Downsample(src, dst,
                     (unsigned int)floor((double)src.GetWidth()/
                                         *_RescaleFactor),
                     (unsigned int)floor((double)src.GetHeight()/
                                         *_RescaleFactor));
}



template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
Downsample(const Image<InputImageType>& src, Image<OutputImageType>& dst,
           const unsigned newwidth, const unsigned newheight)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::Downsample(src, dst, width, height)\n");
  const double hfactor = (double)src.GetWidth() / (double)newwidth;
  const double vfactor = (double)src.GetHeight() / (double)newheight;
  const double factor = (vfactor>hfactor)?vfactor:hfactor;

  Image<OutputImageType> meanim;
  if (_ApplyMeanFilter(src, meanim, factor)!=0){
    BIASERR("error applying mean filter");
  }

  int tlx, tly, brx, bry;
  meanim.GetROI(tlx, tly, brx, bry);

  tlx=(int)ceil((double)tlx/hfactor);
  tly=(int)ceil((double)tly/vfactor);
  brx=(int)floor((double)brx/hfactor);
  bry=(int)floor((double)bry/vfactor);
  
  if (dst.GetWidth()!=newwidth || dst.GetHeight()!=newheight){
    if (!dst.IsEmpty()) dst.Release();
    dst.Init(newwidth, newheight, src.GetChannelCount(), src.GetStorageType(),
             src.IsInterleaved());
  }
  dst.SetROI(tlx, tly, brx, bry);

  int res=0;
  if (src.GetChannelCount()==1 && src.GetColorModel()==ImageBase::CM_Grey){
    res=_FillInterpolatedGrey(meanim, dst);
  } else {
    res=_FillInterpolatedColor(meanim, dst);
  }
  return res;
}

/*
template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
DownsampleROI(const Image<InputImageType>& src, Image<OutputImageType>& dst,
              const unsigned newwidth, const unsigned newheight)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "DownsampleROI(width, height)\n");
  unsigned int src_width = src.GetWidth();
  unsigned int src_height = src.GetHeight();

  
  // The ROI boundaries
  unsigned int ul_x;
  unsigned int ul_y;
  unsigned int lr_x;
  unsigned int lr_y;  
  src.GetROI(ul_x, ul_y, lr_x, lr_y);

#ifdef BIAS_DEBUG
  if(ul_x==0 && ul_y==0 && lr_x==src_width && lr_y==src_height){
    BIASERR("No ROI defined in source image. Nothing to downsample.");
    return -1;
  }
#endif

  unsigned int roi_width = lr_x - ul_x + 1;

  // So what is the factor anyway? (Based on newwidth)
  float factor = 1.0 / ((float) newwidth / (float) roi_width);
  
#ifdef BIAS_DEBUG
  unsigned int roi_height = lr_y - ul_y + 1;
  // Scaling factor in x and y direction equal?
  if (  roi_width * newheight != roi_height * newwidth) {
    BIASERR("Downsample to new width AND height for ROI width and height must "
            <<"be of equal factor in both directions. ");
    // ... or adapt the mean filter!
    return -1;
  }

  if (factor<1.0){
    BIASERR("New width and height would suggest an upsampling. Nothing done "
            <<factor);
    return -1;
  }
#endif
  
  unsigned int halfsize = (factor <= 2.0) ? 1 : 
    (unsigned int) ceil((factor-1.0)/2.0);
  
  // Check the ROI boundaries
  if (ul_x < halfsize || 
      ul_y < halfsize || 
      lr_x > src_width - 1 - halfsize ||
      lr_y > src_height - 1 - halfsize 
      )
    {
      BIASERR("ROI is to near to the edge of the source image.");
      return -1;
    }

  // Create an image containing the ROI plus edge
  Image<InputImageType> roiwithedge, srcroi=src;
  srcroi.SetROI(ul_x - halfsize, ul_y - halfsize, lr_x + halfsize, 
                lr_y + halfsize);
  srcroi.GetCopyOfROI2(roiwithedge);

  Image<OutputImageType> meanim;
  if (_ApplyMeanFilter(roiwithedge, meanim, factor)!=0){
    BIASERR("error applying mean filter");
  }

  meanim.SetROI(halfsize, halfsize, 
		meanim.GetWidth() - 1 - halfsize, 
		meanim.GetHeight() - 1 - halfsize);

  Image<OutputImageType> filtered_image;
  meanim.GetCopyOfROI(filtered_image);

  return Upsample(filtered_image, dst, newwidth, newheight);  
}
*/
////////////////////////////////////////////////////////////////////////////

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
DownsampleBy2(const Image<InputImageType>& src, Image<OutputImageType>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy2(src, dst)\n");
  int res=0;
  unsigned nwidth=src.GetWidth()>>1;
  unsigned nheight=src.GetHeight()>>1;
  if (dst.GetWidth()!=nwidth || dst.GetHeight()!=nheight){
    if (!dst.IsEmpty()) dst.Release();
    dst.Init(nwidth, nheight, src.GetChannelCount(), src.GetStorageType(), 
             src.IsPlanar());
  }
  if (src.GetColorModel()==ImageBase::CM_Grey){
    res = DownsampleBy2Grey(src, dst);
  } else {
    if (src.IsInterleaved() && src.GetChannelCount()==3){
      res = DownsampleBy2Color(src, dst);
    } else {
      // this is slow
      int newwidth=src.GetWidth()>>1;
      int newheight=src.GetHeight()>>1;
      res = Downsample(src, dst, newwidth, newheight);
    }
  }
  
  return res;
}

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
DownsampleBy2Grey(const Image<InputImageType>& Source,
                  Image<OutputImageType>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy2Grey\n");
  unsigned minx, miny, maxx, maxy, dx, dy, mdx, mdy;
  Source.GetROI(minx, miny, maxx, maxy);

  // make sure the image is not shifted
  if (minx%2!=0) minx++;
  if (maxx%2!=0) maxx--;
  if (miny%2!=0) miny++;
  if (maxy%2!=0) maxy--;

  dx=minx>>1;
  dy=miny>>1;
  mdx=maxx>>1;
  mdy=(maxy>>1)-1;

  Destination.SetROI(dx, dy, mdx, mdy+1);

  register double sum;
  register InputImageType *srcu, *srcl;
  register OutputImageType *dst, *dstend, *dstlend;
  register int srcwidth=Source.GetWidth(), dstwidth=Destination.GetWidth(),
    dststep;
  dst = Destination.GetImageData();

  srcl = srcu = Source.GetImageData()+minx+miny*srcwidth;
  srcl += srcwidth;
  dst = Destination.GetImageData() + dy*dstwidth;
  dstlend = dst + mdx;
  dstend = Destination.GetImageData() + mdx + mdy*dstwidth;
  dst+=dx;

  srcwidth+=(srcwidth-maxx+minx);
  dststep=dstwidth-mdx+dx;

  if (srcwidth%2!=0){
    srcwidth+=1;
  }
  while (dst < dstend){
    while (dst < dstlend){
      sum = 0.0;
      sum += (double)*srcu++;
      sum += (double)*srcu++;
      sum += (double)*srcl++;
      sum += (double)*srcl++;
      *dst++ = (OutputImageType)(sum / 4.0);
    }
    srcu += srcwidth;
    srcl += srcwidth;
    dst += dststep;
    dstlend += dstwidth;
  }
  return 0;
}


template <>
int Rescale<unsigned char, unsigned char>::
DownsampleBy2Grey(const Image<unsigned char>& Source, 
                  Image<unsigned char>& Destination) {
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy2Grey\n");
  register unsigned  sum;
  unsigned minx, miny, maxx, maxy, dx, dy, mdx, mdy;
  Source.GetROI(minx, miny, maxx, maxy);

  // make sure the image is not shifted
  if (minx%2!=0) minx++;
  if (maxx%2!=0) maxx--;
  if (miny%2!=0) miny++;
  if (maxy%2!=0) maxy--;

  dx=minx>>1;
  dy=miny>>1;
  mdx=maxx>>1;
  mdy=(maxy>>1)-1;

  Destination.SetROI(dx, dy, mdx, mdy+1);

  register unsigned char *srcu, *srcl, *dst, *dstend, *dstlend;
  register int srcwidth=Source.GetWidth(), dstwidth=Destination.GetWidth(),
    dststep;

  srcl = srcu = Source.GetImageData()+minx+miny*srcwidth;
  srcl += srcwidth;
  dst = Destination.GetImageData() + dy*dstwidth;
  dstlend = dst + mdx;
  dstend = Destination.GetImageData() + mdx + mdy*dstwidth;
  dst+=dx;

  srcwidth+=(srcwidth-maxx+minx);
  dststep=dstwidth-mdx+dx;

  if (srcwidth%2!=0){
    srcwidth+=1;
  }
  while (dst < dstend){
    while (dst < dstlend){
      sum = 0;
      sum += *srcu++;
      sum += *srcu++;
      sum += *srcl++;
      sum += *srcl++;
      *dst++ = sum >> 2;
    }
    srcu += srcwidth;
    srcl += srcwidth;
    dst += dststep;
    dstlend += dstwidth;
  }

  return 0;
}


#ifdef BUILD_IMAGE_USHORT
template <>
int Rescale<unsigned short, unsigned short>::
DownsampleBy2Grey(const Image<unsigned short>& Source,
                      Image<unsigned short>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy2Grey\n");
  register unsigned  sum;
  unsigned minx, miny, maxx, maxy, dx, dy, mdx, mdy;
  Source.GetROI(minx, miny, maxx, maxy);

  // make sure the image is not shifted
  if (minx%2!=0) minx++;
  if (maxx%2!=0) maxx--;
  if (miny%2!=0) miny++;
  if (maxy%2!=0) maxy--;

  dx=minx>>1;
  dy=miny>>1;
  mdx=maxx>>1;
  mdy=(maxy>>1)-1;

  Destination.SetROI(dx, dy, mdx, mdy+1);

  register unsigned short *srcu, *srcl, *dst, *dstend, *dstlend;
  register int srcwidth=Source.GetWidth(), dstwidth=Destination.GetWidth(),
    dststep;

  srcl = srcu = Source.GetImageData()+minx+miny*srcwidth;
  srcl += srcwidth;
  dst = Destination.GetImageData() + dy*dstwidth;
  dstlend = dst + mdx;
  dstend = Destination.GetImageData() + mdx + mdy*dstwidth;
  dst+=dx;

  srcwidth+=(srcwidth-maxx+minx);
  dststep=dstwidth-mdx+dx;

  if (srcwidth%2!=0){
    srcwidth+=1;
  }
  while (dst < dstend){
    while (dst < dstlend){
      sum = 0;
      sum += *srcu++;
      sum += *srcu++;
      sum += *srcl++;
      sum += *srcl++;
      *dst++ = sum >> 2;
    }
    srcu += srcwidth;
    srcl += srcwidth;
    dst += dststep;
    dstlend += dstwidth;
  }

  return 0;
}
#endif

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
DownsampleBy2Color(const Image<InputImageType>& Source, 
                   Image<OutputImageType>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy2Color\n");
  
#ifdef BIAS_DEBUG
  if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
    ImageBase im=Source;
    ImageIO::Write("filtered-source.mip", im);
  }
  unsigned tlx, tly, brx, bry;
  Source.GetROI(tlx, tly, brx, bry);
  if (tlx!=0 || tly!=0 || brx!=Source.GetWidth() || bry!=Source.GetHeight()){
    BIASERR("ROI not yet supported");
    BIASASSERT(false);
  }
#endif
  register double sum1, sum2, sum3;
  register InputImageType *srcu, *srcl;
  register OutputImageType   *dst, *dstend, *dstlend;
  register int srcwidth=Source.GetWidth()*3, dstwidth=Destination.GetWidth()*3;
  dst = Destination.GetImageData();
  srcl = srcu = Source.GetImageData();
  srcl += srcwidth;
  dstend = dst + Destination.GetPixelCount() * 3;
  dstlend = dst + dstwidth * 3;

  if (srcwidth%2!=0){
    srcwidth+=3;
  }
  while (dst < dstend){
    while (dst < dstlend){
      sum1 = sum2 = sum3 = 0.0;
      sum1 += (double)*srcu++;
      sum2 += (double)*srcu++;
      sum3 += (double)*srcu++;
      sum1 += (double)*srcu++;
      sum2 += (double)*srcu++;
      sum3 += (double)*srcu++;
      sum1 += (double)*srcl++;
      sum2 += (double)*srcl++;
      sum3 += (double)*srcl++;
      sum1 += (double)*srcl++;
      sum2 += (double)*srcl++;
      sum3 += (double)*srcl++;
      *dst++ = (OutputImageType)(sum1 / 4.0);
      *dst++ = (OutputImageType)(sum2 / 4.0);
      *dst++ = (OutputImageType)(sum3 / 4.0);
    }
    srcu += srcwidth;
    srcl += srcwidth;
    dstlend += dstwidth;
  }
  return 0;
}

template <> 
int Rescale<unsigned char, unsigned char>::
DownsampleBy2Color(const Image<unsigned char>& Source, 
                   Image<unsigned char>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy2Color\n");
#ifdef BIAS_DEBUG
  if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
    ImageBase im=Source;
    ImageIO::Write("filtered-source.mip", im);
  }
  unsigned tlx, tly, brx, bry;
  Source.GetROI(tlx, tly, brx, bry);
  if (tlx!=0 || tly!=0 || brx!=Source.GetWidth() || bry!=Source.GetHeight()){
    BIASERR("ROI not yet supported");
    BIASASSERT(false);
  }
#endif
  register unsigned int sum1, sum2, sum3;
  register unsigned char *srcu, *srcl, *dst, *dstend, *dstlend;
  register int srcwidth=Source.GetWidth() * 3, 
    dstwidth=Destination.GetWidth() *3;
  dst = Destination.GetImageData();
  srcl = srcu = Source.GetImageData();
  srcl += srcwidth;
  dstend = dst + Destination.GetPixelCount() * 3;
  dstlend = dst + dstwidth;

  if (srcwidth%2!=0){
    srcwidth+=3;
  }
  while (dst < dstend){
    while (dst < dstlend){
      sum1 = sum2 = sum3 = 0;
      sum1 += *srcu++;
      sum2 += *srcu++;
      sum3 += *srcu++;
      sum1 += *srcu++;
      sum2 += *srcu++;
      sum3 += *srcu++;
      sum1 += *srcl++;
      sum2 += *srcl++;
      sum3 += *srcl++;
      sum1 += *srcl++;
      sum2 += *srcl++;
      sum3 += *srcl++;
      *dst++ = sum1 >> 2;
      *dst++ = sum2 >> 2;
      *dst++ = sum3 >> 2;
    }
    srcu += srcwidth;
    srcl += srcwidth;
    dstlend += dstwidth;
  }

//   for (unsigned int y=0;y<Destination.GetHeight(); y++){
//     for (unsigned int x=0;x<Destination.GetWidth(); x++){
//       for (unsigned int c=0;c<3;c++){
// 	sum1 = srcu[y * 
//       }
//     }
//   } 


  return 0;
}

#ifdef BUILD_IMAGE_USHORT
template <> 
int Rescale<unsigned short, unsigned short>::
DownsampleBy2Color(const Image<unsigned short>& Source, 
                   Image<unsigned short>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy2Color\n");
#ifdef BIAS_DEBUG
  if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
    ImageBase im=Source;
    ImageIO::Write("filtered-source.mip", im);
  }
  unsigned tlx, tly, brx, bry;
  Source.GetROI(tlx, tly, brx, bry);
  if (tlx!=0 || tly!=0 || brx!=Source.GetWidth() || bry!=Source.GetHeight()){
    BIASERR("ROI not yet supported");
    BIASASSERT(false);
  }
#endif
  register unsigned int sum1, sum2, sum3;
  register unsigned short *srcu, *srcl, *dst, *dstend, *dstlend;
  register int srcwidth=Source.GetWidth() * 3, 
    dstwidth=Destination.GetWidth() *3;
  dst = Destination.GetImageData();
  srcl = srcu = Source.GetImageData();
  srcl += srcwidth;
  dstend = dst + Destination.GetPixelCount() * 3;
  dstlend = dst + dstwidth;

  if (srcwidth%2!=0){
    srcwidth+=3;
  }
  while (dst < dstend){
    while (dst < dstlend){
      sum1 = sum2 = sum3 = 0;
      sum1 += *srcu++;
      sum2 += *srcu++;
      sum3 += *srcu++;
      sum1 += *srcu++;
      sum2 += *srcu++;
      sum3 += *srcu++;
      sum1 += *srcl++;
      sum2 += *srcl++;
      sum3 += *srcl++;
      sum1 += *srcl++;
      sum2 += *srcl++;
      sum3 += *srcl++;
      *dst++ = sum1 >> 2;
      *dst++ = sum2 >> 2;
      *dst++ = sum3 >> 2;
    }
    srcu += srcwidth;
    srcl += srcwidth;
    dstlend += dstwidth;
  }
  return 0;
}
#endif

/////////////////////////////////////////////////////////////////////////////

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
DownsampleBy4(const Image<InputImageType>& src, Image<OutputImageType>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy4\n");
  int res=0;
  unsigned nwidth=(int)src.GetWidth()>>2;
  unsigned nheight=(int)src.GetHeight()>>2;
  if (dst.GetWidth()!=nwidth || dst.GetHeight()!=nheight){
    if (!dst.IsEmpty()) dst.Release();
    dst.Init(nwidth, nheight, src.GetChannelCount(), src.GetStorageType(), 
             src.IsPlanar());
  }
  if (src.GetColorModel()==ImageBase::CM_Grey){
    res = DownsampleBy4Grey(src, dst);
  } else {
    if (src.IsInterleaved() && src.GetChannelCount()==3){
      res = DownsampleBPoT(src, dst);
    } else {
      // this is slow
      int newwidth=src.GetWidth()>>2;
      int newheight=src.GetHeight()>>2;
      res = Downsample(src, dst, newwidth, newheight);
    }
  }
  return res;
}

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
DownsampleBy4Grey(const Image<InputImageType>& Source,
                  Image<OutputImageType>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy4Grey\n");
  BIASASSERT(Source.GetChannelCount()==1 && 
             Source.GetColorModel()==ImageBase::CM_Grey);
  BIASASSERT(Destination.GetChannelCount()==1 && 
             Destination.GetColorModel()==ImageBase::CM_Grey);
  BIASASSERT(Source.GetWidth()>>2==Destination.GetWidth());
  BIASASSERT(Source.GetHeight()>>2==Destination.GetHeight());

  /// \todo warning kerneltype sum needed here    
    
  double  sum;
  unsigned minx, miny, maxx, maxy, dx, dy, mdx, mdy;
  Source.GetROI(minx, miny, maxx, maxy);

  // make sure the image is not shifted
  if (minx%4!=0) minx+=(4-(minx%4));
  if (maxx%4!=0) maxx-=(maxx%4);
  if (miny%4!=0) miny+=(4-(miny%4));
  if (maxy%4!=0) maxy-=(maxy%4);

  dx=minx>>2;
  dy=miny>>2;
  mdx=maxx>>2;
  mdy=maxy>>2;

  Destination.SetROI(dx, dy, mdx, mdy);

  register InputImageType *src1, *src2, *src3, *src4;
  register OutputImageType *dst, *dstend, *dstlend;
  register int srcwidth=(int)Source.GetWidth();
  register const int dstwidth=(int)Destination.GetWidth();
  register int dststep;

  src1 = Source.GetImageData()+minx+miny*srcwidth;
  src2 = src1 + srcwidth;
  src3 = src2 + srcwidth;
  src4 = src3 + srcwidth;
  
  dst = Destination.GetImageData() + dy*dstwidth;
  dstlend = dst + mdx;
  dst+=dx;
  dstend = Destination.GetImageData() + mdx + (mdy-1)*dstwidth;

  srcwidth+=(3*srcwidth-maxx+minx);
  dststep=dstwidth-mdx+dx;

  if (srcwidth%2!=0){
    srcwidth+=1;
  }
  
  /// \todo warning kerneltype div needed here

  const double div=(1.0/16.0);
  while (dst < dstend){
    while (dst < dstlend){
      sum = 0.0;
      sum += *src1++;
      sum += *src1++;
      sum += *src1++;
      sum += *src1++;

      sum += *src2++;
      sum += *src2++;
      sum += *src2++;
      sum += *src2++;

      sum += *src3++;
      sum += *src3++;
      sum += *src3++;
      sum += *src3++;

      sum += *src4++;
      sum += *src4++;
      sum += *src4++;
      sum += *src4++;

      *dst++ = (OutputImageType)(sum * div);
    }
    src1 += srcwidth;
    src2 += srcwidth;
    src3 += srcwidth;
    src4 += srcwidth;
    dst += dststep;
    dstlend += dstwidth;
  }

  return 0;
}


template <>
int Rescale<unsigned char, unsigned char>::
DownsampleBy4Grey(const Image<unsigned char>& Source, 
                  Image<unsigned char>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy4Grey\n");
  BIASASSERT(Source.GetChannelCount()==1 && 
             Source.GetColorModel()==ImageBase::CM_Grey);
  BIASASSERT(Destination.GetChannelCount()==1 && 
             Destination.GetColorModel()==ImageBase::CM_Grey);
  BIASASSERT(Source.GetWidth()>>2==Destination.GetWidth());
  BIASASSERT(Source.GetHeight()>>2==Destination.GetHeight());
  register unsigned  sum;
  unsigned minx, miny, maxx, maxy, dx, dy, mdx, mdy;
  Source.GetROI(minx, miny, maxx, maxy);

  // make sure the image is not shifted
  if (minx%4!=0) minx+=(4-(minx%4));
  if (maxx%4!=0) maxx-=(maxx%4);
  if (miny%4!=0) miny+=(4-(miny%4));
  if (maxy%4!=0) maxy-=(maxy%4);

  dx=minx>>2;
  dy=miny>>2;
  mdx=maxx>>2;
  mdy=maxy>>2;

  Destination.SetROI(dx, dy, mdx, mdy);

  register unsigned char *src1, *src2, *src3, *src4, *dst, *dstend, *dstlend;
  register int srcwidth=(int)Source.GetWidth();
  register const int dstwidth=(int)Destination.GetWidth();
  register int dststep;

  src1 = Source.GetImageData()+minx+miny*srcwidth;
  src2 = src1 + srcwidth;
  src3 = src2 + srcwidth;
  src4 = src3 + srcwidth;
  
  dst = Destination.GetImageData() + dy*dstwidth;
  dstlend = dst + mdx;
  dst+=dx;
  dstend = Destination.GetImageData() + mdx + (mdy-1)*dstwidth;

  srcwidth+=(3*srcwidth-maxx+minx);
  dststep=dstwidth-mdx+dx;

  if (srcwidth%2!=0){
    srcwidth+=1;
  }
  while (dst < dstend){
    while (dst < dstlend){
      sum = 0;
      sum += *src1++;
      sum += *src1++;
      sum += *src1++;
      sum += *src1++;

      sum += *src2++;
      sum += *src2++;
      sum += *src2++;
      sum += *src2++;

      sum += *src3++;
      sum += *src3++;
      sum += *src3++;
      sum += *src3++;

      sum += *src4++;
      sum += *src4++;
      sum += *src4++;
      sum += *src4++;

      *dst++ = sum >> 4;
    }
    src1 += srcwidth;
    src2 += srcwidth;
    src3 += srcwidth;
    src4 += srcwidth;
    dst += dststep;
    dstlend += dstwidth;
  }

  return 0;
}

#ifdef BUILD_IMAGE_USHORT
template <>
int Rescale<unsigned short, unsigned short>::
DownsampleBy4Grey(const Image<unsigned short>& Source, 
                      Image<unsigned short>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBy4Grey\n");
  BIASASSERT(Source.GetChannelCount()==1 && 
             Source.GetColorModel()==ImageBase::CM_Grey);
  BIASASSERT(Destination.GetChannelCount()==1 && 
             Destination.GetColorModel()==ImageBase::CM_Grey);
  BIASASSERT(Source.GetWidth()>>2==Destination.GetWidth());
  BIASASSERT(Source.GetHeight()>>2==Destination.GetHeight());
  register unsigned  sum;
  unsigned minx, miny, maxx, maxy, dx, dy, mdx, mdy;
  Source.GetROI(minx, miny, maxx, maxy);

  // make sure the image is not shifted
  if (minx%4!=0) minx+=(4-(minx%4));
  if (maxx%4!=0) maxx-=(maxx%4);
  if (miny%4!=0) miny+=(4-(miny%4));
  if (maxy%4!=0) maxy-=(maxy%4);

  dx=minx>>2;
  dy=miny>>2;
  mdx=maxx>>2;
  mdy=maxy>>2;

  Destination.SetROI(dx, dy, mdx, mdy);

  register unsigned short *src1, *src2, *src3, *src4, *dst, *dstend, *dstlend;
  register int srcwidth=(int)Source.GetWidth();
  register const int dstwidth=(int)Destination.GetWidth();
  register int dststep;

  src1 = Source.GetImageData()+minx+miny*srcwidth;
  src2 = src1 + srcwidth;
  src3 = src2 + srcwidth;
  src4 = src3 + srcwidth;
  
  dst = Destination.GetImageData() + dy*dstwidth;
  dstlend = dst + mdx;
  dst+=dx;
  dstend = Destination.GetImageData() + mdx + (mdy-1)*dstwidth;

  srcwidth+=(3*srcwidth-maxx+minx);
  dststep=dstwidth-mdx+dx;

  if (srcwidth%2!=0){
    srcwidth+=1;
  }
  while (dst < dstend){
    while (dst < dstlend){
      sum = 0;
      sum += *src1++;
      sum += *src1++;
      sum += *src1++;
      sum += *src1++;

      sum += *src2++;
      sum += *src2++;
      sum += *src2++;
      sum += *src2++;

      sum += *src3++;
      sum += *src3++;
      sum += *src3++;
      sum += *src3++;

      sum += *src4++;
      sum += *src4++;
      sum += *src4++;
      sum += *src4++;

      *dst++ = sum >> 4;
    }
    src1 += srcwidth;
    src2 += srcwidth;
    src3 += srcwidth;
    src4 += srcwidth;
    dst += dststep;
    dstlend += dstwidth;
  }

  return 0;
}
#endif





template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
DownsampleBPoT(const Image<InputImageType>& Source, 
		Image<OutputImageType>& Destination) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::DownsampleBPoT\n");
  int res = 0;
  unsigned int channels = Source.GetChannelCount();
  int PowerOfTwo = (int)rint(log(*_RescaleFactor)/log(2.0));
  if (PowerOfTwo == 0){
    // No downsample required, just copy
    if (Destination.GetWidth() != Source.GetWidth() 
        || Destination.GetHeight() != Source.GetHeight() 
        ||  Destination.GetChannelCount() != channels ){
      if ( ! Destination.IsEmpty())
        Destination.Release();
      Destination.Init(Source.GetWidth(),Source.GetHeight(),channels);
    }
    Destination = Source;
  } else { // downsample
    unsigned int newwidth, newheight;
    unsigned int oldwidth,oldheight;
    
    oldwidth = Source.GetWidth();
    oldheight = Source.GetHeight();
    
    newwidth = (oldwidth >> PowerOfTwo);
    newheight = (oldheight >> PowerOfTwo);
    
    if (Destination.GetWidth() != newwidth 
        || Destination.GetHeight() != newheight 
        ||  Destination.GetChannelCount() != channels ){
      if ( ! Destination.IsEmpty())
        Destination.Release();
      Destination.Init(newwidth,newheight,channels);
    }
    OutputImageType *sink;
    InputImageType *src;
    bool IdenticalImages = false;
    
    sink = Destination.GetImageData();
    src = Source.GetImageData(); 
    
    if ((void*)src==(void*)sink){
      IdenticalImages = true;
      sink = new OutputImageType[newwidth * newheight];
    }
 
   
    // calculate offset array
    int *offset;
    register int x, y, length=(1<<PowerOfTwo);    
    register int index=0;
    offset = new int[length*length]; // 2x2 for PowerofTwo==1
    length=(length>>1);
    for(y=-length; y<length; y++)
      for(x=-length; x<length; x++){
        offset[index++]=(int)(y*oldwidth*channels + x*channels);
      }
    
    double div=pow(2.0,PowerOfTwo+PowerOfTwo);
    register int inc = (length<<1); 
    
    int min_y = (int)(length*oldwidth*channels);
    int max_y = (int)(oldheight*oldwidth*channels);
    register int inc_y = (int)(oldwidth*inc*channels);
    
    int min_x = length*channels;
    int max_x = oldwidth*channels;
    register int inc_x =(int)(inc*channels);
    //     cout <<"channels: "<<channels<<endl;    
    //     cout <<"y-loop: "<<min_y<<" <"<<max_y<<" inc: "<<inc_y<<endl;
    //     cout <<"x-loop: "<<min_x<<" <"<<max_x<<" inc: "<<inc_x<<endl;
    
    register int off_ind;

    /// \todo warning kerneltype here
      
    register double sum;
    register int image_index=0;
    for(y=min_y; y<max_y;y+=inc_y)    
      for(x=min_x; x<max_x; x+=inc_x){
        for (unsigned int c = 0; c<channels; c++) {
          sum = 0.0;
          for(off_ind=0;off_ind<index;off_ind++){
            sum +=  src[y+x+offset[off_ind]+c];
          }
          sink[image_index]=(OutputImageType)(sum/div);
          // 	    sink[image_index]=src[y+x+c];
          image_index ++;
        }
      }
  
  
    
    delete[] offset;
    
    if (IdenticalImages) {
      Destination.Release();
      Destination.Init(newwidth, newheight, channels);
      Destination.ReleaseImageDataPointer();

      Destination.RedirectImageDataPointer(sink);
    }
  }

  return res;
}




//////////////////////////////////////////////////////////////////////////
//               upsampling functions
//////////////////////////////////////////////////////////////////////////

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
Upsample(const Image<InputImageType>& src, Image<OutputImageType>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::Upsample(src, dst)\n");
  return Upsample(src, dst,
                  (unsigned int)floor((double)src.GetWidth()/
                                      *_RescaleFactor),
                  (unsigned int)floor((double)src.GetHeight()/
                                      *_RescaleFactor));
}

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
Upsample(const Image<InputImageType>& src, Image<OutputImageType>& dst,
         const unsigned newwidth, const unsigned newheight)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::Upsample(src, dst, w, h)\n");
#ifdef BIAS_DEBUG
  unsigned tlx, tly, brx, bry;
  src.GetROI(tlx, tly, brx, bry);
  if (tlx!=0 || tly!=0 || brx!=src.GetWidth() || bry!=src.GetHeight()){
    BIASERR("ROI not yet supported");
    BIASASSERT(false);
  }
#endif
  BIASASSERT(src.GetWidth()<newwidth);
  BIASASSERT(src.GetHeight()<newheight);
  int res=0;
  float oldx, oldy;
  float scalex, scaley;
  unsigned int channelcount=src.GetChannelCount();
  bool identicalimages=false;
  OutputImageType *sink;

  if ((void *)src.GetImageData() == (void *)dst.GetImageData())
    identicalimages=true;

  if (!identicalimages){
    if (!src.SamePixelAndChannelCount(dst))
      if (!dst.IsEmpty()) dst.Release();
    
    if (dst.IsEmpty())
      dst.Init(newwidth, newheight, channelcount);

    sink = dst.GetImageData();
  } else {
    sink = new OutputImageType[ newwidth *  newheight * channelcount];
  }
  
  scalex = (float)(src.GetWidth() - 1)/ (float) (newwidth);
  scaley = (float)(src.GetHeight() - 1) / (float) (newheight);
  if (src.IsInterleaved()) {    
    for (unsigned int y = 0; y<newheight; y++)
      for (unsigned int x = 0; x<newwidth; x++)
        for (unsigned int c = 0; c<channelcount; c++) {
          oldx = (float)x * scalex;
          oldy = (float)y * scaley;
//           oldx = (float)x / factor;
//           oldy = (float)y / factor;
          sink[(y * newwidth + x)* channelcount + c] = 
            (OutputImageType)rint(src.BilinearInterpolation(oldx, oldy, c));
        }
  } else { // planar
    BIASERR("Upscale only implemented on interleaved images");
    res = -1;
  }
  if (identicalimages && res==0){
    dst.Release();
    dst.Init(newwidth, newheight, channelcount);
    
    dst.ReleaseImageDataPointer();
    dst.RedirectImageDataPointer(sink);
  }
  
  return res; 
}

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
UpsampleGrey(const Image<InputImageType>& src, Image<OutputImageType>& dst,
             const unsigned newwidth, const unsigned newheight)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::UpsampleGrey(src, dst, w, h)\n");
#ifdef BIAS_DEBUG
  unsigned tlx, tly, brx, bry;
  src.GetROI(tlx, tly, brx, bry);
  if (tlx!=0 || tly!=0 || brx!=src.GetWidth() || bry!=src.GetHeight()){
    BIASERR("ROI not yet supported");
    BIASASSERT(false);
  }
#endif
#ifdef BIAS_DEBUG
  if (src.GetChannelCount()!=1 || src.GetColorModel()!=ImageBase::CM_Grey){
    BIASERR("only for grey images");
    BIASASSERT(false);
  }
#endif
  BIASASSERT(src.GetWidth()<newwidth);
  BIASASSERT(src.GetHeight()<newheight);
  BIASASSERT(src.GetChannelCount()==1);
  BIASASSERT(src.GetColorModel()==ImageBase::CM_Grey);
  int res=0;
  float oldx, oldy;
  float scalex, scaley;
  bool identicalimages=false;
  OutputImageType *sink;

  if ((void *)src.GetImageData() == (void *)dst.GetImageData())
    identicalimages=true;

  if (!identicalimages){
    if (!src.SamePixelAndChannelCount(dst))
      if (!dst.IsEmpty()) dst.Release();
    
    if (dst.IsEmpty())
      dst.Init(newwidth, newheight, 1);

    sink = dst.GetImageData();
  } else {
    sink = new OutputImageType[ newwidth *  newheight ];
  }
  
  scalex = (float)(src.GetWidth() - 1)/ (float) (newwidth);
  scaley = (float)(src.GetHeight() - 1) / (float) (newheight);
  if (src.IsInterleaved()) {    
    for (unsigned int y = 0; y<newheight; y++)
      for (unsigned int x = 0; x<newwidth; x++){
        oldx = (float)x * scalex;
        oldy = (float)y * scaley;
        sink[(y * newwidth + x)] = 
          src.FastBilinearInterpolationGrey(oldx, oldy);
      }
  } else { // planar
    BIASERR("Upscale only implemented on interleaved images");
    res = -1;
  }
  if (identicalimages && res==0){
    dst.Release();
    dst.Init(newwidth, newheight, 1);
    
    dst.ReleaseImageDataPointer();
    dst.RedirectImageDataPointer(sink);
  }
  
  return res;
}

////////////////////////////////////////////////////////////////////////////

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
UpsampleBy2Grey(const Image<InputImageType>& Source,
                Image<OutputImageType>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::UpsampleBy2Grey\n");
  register OutputImageType *dstu=NULL, *dstl=NULL;
  register InputImageType *src=NULL, *srcend=NULL,    *srclend=NULL;
  register int srcwidth=Source.GetWidth(), dstwidth=Destination.GetWidth();
#ifdef BIAS_DEBUG
  if (srcwidth*2!=dstwidth) {
      BIASERR("wrong image width :"<<srcwidth<<"*2 != "<<dstwidth);
      abort();
  }
#endif
  dstu = dstl = Destination.GetImageData();
  src = Source.GetImageData();
  dstl += dstwidth;
  srcend = src + Source.GetPixelCount();
  srclend = src + srcwidth;

  while (src < srcend){
    while (src < srclend){
      *dstu++ = *dstl++ = *src;
      *dstu++ = *dstl++ = *src++;
    }
    dstu += dstwidth;
    dstl += dstwidth;
    srclend += srcwidth;
  }
  return 0;
}

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
UpsampleBy2(const Image<InputImageType>& Source, 
            Image<OutputImageType>& Destination)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::UpsampleBy2\n");
#ifdef BIAS_DEBUG
  if (!Source.IsInterleaved()){
    BIASERR("only for interleaved images");
    BIASASSERT(false);
  }
#endif
  register OutputImageType *dstu=NULL, *dstl=NULL;
   register InputImageType *src=NULL, *srcend=NULL,  *srclend=NULL;
  register int srcwidth=Source.GetWidthStep(), 
    dstwidth=Destination.GetWidthStep(),
    cc=Source.GetChannelCount();
  dstu = dstl = Destination.GetImageData();
  src = Source.GetImageData();
  dstl += dstwidth;
  srcend = src + Source.GetPixelCount() * cc - cc;
  srclend = src + srcwidth - cc;

  while (src < srcend){
    while (src < srclend){
      for (int i=0; i<cc; i++){
        *dstu = *dstl = dstu[cc] = dstl[cc] = *src;
        src++; dstu++; dstl++;
      }
      dstu+=cc;
      dstl+=cc;
    }
    dstu += dstwidth;
    dstl += dstwidth;
    srclend += srcwidth;
  }
  return 0;
}


//////////////////////////////////////////////////////////////////////////
//               protected functions
//////////////////////////////////////////////////////////////////////////

template <class InputImageType, class OutputImageType>
void Rescale<InputImageType, OutputImageType>::
_AddParam(Param& para, const int grp, const std::string& postfix)
{
  
  if (para.CheckParam("LowPassType"+postfix)){ 
    _RescaleMeanType =
      para.GetParamEnum("LowPassType"+postfix);
  } else { 
    vector<string> lptypes;
    lptypes.push_back(string("LPT_Mean"));
    lptypes.push_back(string("LPT_Gauss"));
    
    _RescaleMeanType =
      para.AddParamEnum("LowPassType"+postfix,
                        "kind of low pass filter used",
                        lptypes, LPT_Mean, NULL, 0, grp); 
  }

  if (para.CheckParam("MeanSigmaFactor"+postfix)){ 
    _RescaleMeanSigmaFactor=
      para.GetParamDouble("MeanSigmaFactor"+postfix);
  } else { 
    _RescaleMeanSigmaFactor=
      para.AddParamDouble("MeanSigmaFactor"+postfix,
         "sigma of mean filter = MeanSigmaFactor * RescaleFactor",
                                       0.9, 0.0, DBL_MAX, 0, grp); 
  }
  para.SetEnhancedFlag("MeanSigmaFactor"+postfix, true);

  if (para.CheckParam("Factor"+postfix)){ 
    _RescaleFactor=para.GetParamDouble("Factor"+postfix);
  } else { 
    _RescaleFactor=
      para.AddParamDouble("Factor"+postfix,
         "the rescale factor used in most of the functions. Factor>1.0 means downsampling.",
                          2.0, 0.0, DBL_MAX, 0, grp); 
  }

  ///////////////////////////////////////////////////////////////////////////

  _MeanSize = para.GetParamInt("Size-Mean"+postfix);

  _GaussSigma = para.GetParamDouble("Sigma-Gauss"+postfix);

  _GaussRatio = para.GetParamDouble("Ratio-Gauss"+postfix);

}

template <class InputImageType, class OutputImageType>
void Rescale<InputImageType, OutputImageType>::
_UpdateMean()
{
  switch(*_RescaleMeanType){
  case LPT_Gauss:
    _lowpass = &_gauss;
    BIASCDOUT(D_RESCALE_INIT, "using Gauss low pass filter\n");
    break;
  case LPT_Binomial:
    _lowpass = &_binomial;
    BIASCDOUT(D_RESCALE_INIT, "using Binomial low pass filter\n");
    break;
  case LPT_Mean:
    _lowpass = &_mean;
    BIASCDOUT(D_RESCALE_INIT, "using mean low pass filter \n");
    break;
  default:
    BIASERR("unknown low pass filter type");
    _lowpass = NULL;
    BIASASSERT(false);
    break;
  }
}

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
_ApplyMeanFilter(const Image<InputImageType>& src, 
                 Image<OutputImageType>& dst, const double factor)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::_ApplyMeanFilter\n");
  switch (*_RescaleMeanType){
  case LPT_Gauss:
    _lowpass = &_gauss;
    if (this->_ParamsAllocated) _gauss.SetSigma(*_RescaleMeanSigmaFactor 
                                                * factor);
    else {
      *_GaussSigma = *_RescaleMeanSigmaFactor * factor;
      BIASCDOUT(D_RESCALE_INIT, "using Gauss low pass filter with sigma "
                <<*_GaussSigma<<"\n");
    }
    break;
  case LPT_Mean:
    _lowpass = &_mean;
    if (this->_ParamsAllocated) _mean.SetSize((int)ceil(1.0*factor));
    else {
      *_MeanSize = (int)ceil(1.0*factor);
      BIASCDOUT(D_RESCALE_INIT, "using mean low pass filter with size "
                <<*_MeanSize<<"\n");
    }
    break;
  default:
    BIASERR("unknown low pass filter type");
    _lowpass = NULL;
    BIASASSERT(false);
    break;
  }
  //_lowpass->AddDebugLevel(D_CONV_KERNEL);
  int res=_lowpass->Filter(src, dst);
#ifdef BIAS_DEBUG
  if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
    ImageIO::Write("rescale-lowpass-filtered.mip", dst);
  }
#endif
  return res;
}


template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
//_FillInterpolatedGrey(const Image<InputImageType>& src, 
_FillInterpolatedGrey(const Image<OutputImageType>& src, 
                      Image<OutputImageType>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::_FillInterpolatedGrey\n");
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1 && 
             src.GetColorModel()==ImageBase::CM_Grey);
  BIASASSERT(!dst.IsEmpty());
  
  if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
    ImageBase im=src;
    ImageIO::Write("filtered.mip", im);
  }
#endif
  int tlx, tly, brx, bry;
  dst.GetROI(tlx, tly, brx, bry);
  const int minx=tlx, maxx=brx, miny=tly, maxy=bry;
  OutputImageType **idst=dst.GetImageDataArray();
  const double facx = (double)src.GetWidth() / (double)dst.GetWidth();
  const double facy = (double)src.GetHeight() / (double)dst.GetHeight();
  register int x, y; 
  for (y=miny; y<maxy; y++){
    for (x=minx; x<maxx; x++){
      idst[y][x]=(OutputImageType)
        src.FastBilinearInterpolationGrey(x*facx, y*facy);
      //      cerr << "( "<<x<<", "<<y<<")  ->  ( "<<x*facx<<", "<<y*facy<<")\n";
    }
  }
  return 0;
} 

template <>
int Rescale<unsigned char, unsigned char>::
_FillInterpolatedGrey(const Image<unsigned char>& src, 
                      Image<unsigned char>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::_FillInterpolatedGrey\n");
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1 && 
             src.GetColorModel()==ImageBase::CM_Grey);
  BIASASSERT(!dst.IsEmpty());
  
  if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
    ImageBase im=src;
    ImageIO::Write("filtered.mip", im);
  }
#endif
  int tlx, tly, brx, bry;
  dst.GetROI(tlx, tly, brx, bry);
  const int minx=tlx, maxx=brx, miny=tly, maxy=bry;
  unsigned char **idst=dst.GetImageDataArray();
  const double facx = (double)src.GetWidth() / (double)dst.GetWidth();
  const double facy = (double)src.GetHeight() / (double)dst.GetHeight();
  register int x, y; 
  for (y=miny; y<maxy; y++){
    for (x=minx; x<maxx; x++){
      idst[y][x]=(unsigned char)
        rint(src.FastBilinearInterpolationGrey(x*facx, y*facy));
    }
  }
  return 0;
} 

#ifdef BUILD_IMAGE_USHORT
template <>
int Rescale<unsigned short, unsigned short>::
_FillInterpolatedGrey(const Image<unsigned short>& src, 
                      Image<unsigned short>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::_FillInterpolatedGrey\n");
#ifdef BIAS_DEBUG
  BIASASSERT(src.GetChannelCount()==1 && 
             src.GetColorModel()==ImageBase::CM_Grey);
  BIASASSERT(!dst.IsEmpty());
  
//   if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
//     ImageIO::Write("filtered.mip", src);
//   }
#endif
  int tlx, tly, brx, bry;
  dst.GetROI(tlx, tly, brx, bry);
  const int minx=tlx, maxx=brx, miny=tly, maxy=bry;
  unsigned short **idst=dst.GetImageDataArray();
  const double facx = (double)src.GetWidth() / (double)dst.GetWidth();
  const double facy = (double)src.GetHeight() / (double)dst.GetHeight();
  register int x, y; 
  for (y=miny; y<maxy; y++){
    for (x=minx; x<maxx; x++){
      idst[y][x]=(unsigned short)
        rint(src.FastBilinearInterpolationGrey(x*facx, y*facy));
    }
  }
  return 0;
} 
#endif

template <class InputImageType, class OutputImageType>
int Rescale<InputImageType, OutputImageType>::
//_FillInterpolatedColor(const Image<InputImageType>& src,
_FillInterpolatedColor(const Image<OutputImageType>& src,
                       Image<OutputImageType>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::_FillInterpolatedColor\n");
#ifdef BIAS_DEBUG
  BIASASSERT(!dst.IsEmpty());
  BIASASSERT(src.IsInterleaved());
  
  if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
    ImageBase im=src;
    ImageIO::Write("filtered.mip", im);
  }
#endif
  int tlx, tly, brx, bry;
  dst.GetROI(tlx, tly, brx, bry);
  const int minx=tlx, maxx=brx, miny=tly, maxy=bry;
  OutputImageType **idst=dst.GetImageDataArray();
  const double facx = (double)src.GetWidth() / (double)dst.GetWidth();
  const double facy = (double)src.GetHeight() / (double)dst.GetHeight();
  register int x, y; 
  const unsigned int cc = src.GetChannelCount();
  unsigned int c;
  for (y=miny; y<maxy; y++){
    for (x=minx; x<maxx; x++){
      for (c=0;c<cc; c++)
      idst[y][x*cc+c]=(OutputImageType)
        src.BilinearInterpolationRGBInterleaved(x*facx, y*facy, c);
    }
  }
  return 0;
}

template <>
int Rescale<unsigned char, unsigned char>::
_FillInterpolatedColor(const Image<unsigned char>& src, 
                       Image<unsigned char>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::_FillInterpolatedColor\n");
#ifdef BIAS_DEBUG
  BIASASSERT(!dst.IsEmpty());
  BIASASSERT(src.IsInterleaved());
  
  if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
    ImageBase im=src;
    ImageIO::Write("filtered.mip", im);
  }
#endif
  int tlx, tly, brx, bry;
  dst.GetROI(tlx, tly, brx, bry);
  const int minx=tlx, maxx=brx, miny=tly, maxy=bry;
  unsigned char **idst=dst.GetImageDataArray();
  const double facx = (double)src.GetWidth() / (double)dst.GetWidth();
  const double facy = (double)src.GetHeight() / (double)dst.GetHeight();
  register int x, y; 
  const unsigned int cc = src.GetChannelCount();
  unsigned int c;
  for (y=miny; y<maxy; y++){
    for (x=minx; x<maxx; x++){
      for (c=0;c<cc; c++)
        idst[y][x*cc+c]=(unsigned char)rint
          (src.BilinearInterpolationRGBInterleaved(x*facx, y*facy, c));
    }
  }
  return 0;
}

#ifdef BUILD_IMAGE_USHORT
// exactly the same as <unsigned char, int>, but partial specializations
// are not supported by many compilers
template <>
int Rescale<unsigned short,unsigned short>::
_FillInterpolatedColor(const Image<unsigned short>& src, 
                       Image<unsigned short>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Rescale::_FillInterpolatedColor\n");
#ifdef BIAS_DEBUG
  BIASASSERT(!dst.IsEmpty());
  BIASASSERT(src.IsInterleaved());
  
//   if (this->DebugLevelIsSet(D_RESCALE_WRITE_IMG)) {
//     ImageIO::Write("filtered.mip", src);
//   }
#endif

  int tlx, tly, brx, bry;
  dst.GetROI(tlx, tly, brx, bry);
  const int minx=tlx, maxx=brx, miny=tly, maxy=bry;
  unsigned short **idst=dst.GetImageDataArray();
  const double facx = (double)src.GetWidth() / (double)dst.GetWidth();
  const double facy = (double)src.GetHeight() / (double)dst.GetHeight();
  register int x, y; 
  const unsigned int cc = src.GetChannelCount();
  unsigned int c;
  for (y=miny; y<maxy; y++){
    for (x=minx; x<maxx; x++){
      for (c=0;c<cc; c++)
        idst[y][x*cc+c]=(unsigned char)rint
          (src.BilinearInterpolationRGBInterleaved(x*facx, y*facy, c));
    }
  }


  return 0;
}
#endif
