/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __Rescale_hh__
#define __Rescale_hh__

#include "FilterBase.hh"
#include "FilterNToN.hh"
#include "Gauss.hh"
#include "Binomial.hh"
#include "Mean.hh"

#define D_RESCALE_INIT       0x00000001
#define D_RESCALE_WRITE_IMG  0x00000002

namespace BIAS 
{
  /** @class Rescale
      @ingroup g_filter
      @brief Down-, Upsampling routines and Resize
      
      @author woelk 09/2004 */
  template <class InputStorageType, class OutputStorageType>
  class Rescale : public FilterBase<InputStorageType, OutputStorageType>
  {
  public:
    /** uses internal param object */
    Rescale();

    Rescale(Param& para, const std::string& postfix="");
    
    Rescale(Param& para, const int grp, const std::string& postfix);
    
    virtual ~Rescale();
    
    /** virtual function 
        calls FilterRescale */
    virtual int Filter(const Image<InputStorageType>& src,
                       Image<OutputStorageType>& dst);

 
    /////////////////////////////////////////////////////////////////////////

    /** generic downsample function
        @author woelk 09/2004 */
    int Downsample(const Image<InputStorageType>& src, 
                   Image<OutputStorageType>& dst);

    /** downsamples src to dst of size newwidth x newheight
        @author woelk 09/2004 **/
    int Downsample(const Image<InputStorageType>& src, 
                   Image<OutputStorageType>& dst,
                   const unsigned newwidth, const unsigned newheight);

    /** Takes the source image that has to have a defined ROI. Based on this 
        ROI it will  store a  downsampled version in the destination image. 
        The pixel values surounding the ROI will be  taken into account while 
        executing the mean filter. Therefore it is not possible to downsample 
        the whole image.
        @param Source Image with a defined ROI
        @param Destination Image where the rescaled image will be stored. Will 
        initialized if required.
        @param newwidth Desired width of the target image
        @param newheight Desired height of the target image
        @return -1 In any case of error (e.g. no ROI defined, ROI to close to
        the edge or desired width/height does no work)
        @date 26.09.2003
        @author Ingo Thomsen **/
    // int DownsampleROI(const Image<InputStorageType>& src, 
    //                  Image<OutputStorageType>& dst,
    //                  const unsigned newwidth, const unsigned newheight);

    /////////////////////////////////////////////////////////////////////////
    //              specialized downsampling functions
    /////////////////////////////////////////////////////////////////////////

    /** calls the appropriate function for the color model.
        mean low pass filter is used implicitly.
        @author woelk 09/2004 */
    int DownsampleBy2(const Image<InputStorageType>& src, 
                      Image<OutputStorageType>& dst);

     /** assumes that destination is initialized.
         mean low pass filter is used implicitly.
         Destination width = floor(source.GetWidth()/2)
         Destination height = floor(source.GetHeight()/2)
         only for one channel images
         also make use of ROI in Source image (fast :-)) )
         not tested with odd size images
         @author woelk 01/2003 */
    int DownsampleBy2Grey(const Image<InputStorageType>& src, 
                          Image<OutputStorageType>& dst);

    /** assumes that destination is initialized. 
        mean low pass filter is used implicitly.
        Destination width = floor(source.GetWidth()/2)
        Destination height = floor(source.GetHeight()/2)
        only for three channel interleaved images
        not tested with odd size images
        @author woelk 01/2003 */
    int DownsampleBy2Color(const Image<InputStorageType>& src, 
                           Image<OutputStorageType>& dst);

    /////////////////////////////////////////////////////////////////////////

     /** calls the appropriate function for the color model
         mean low pass filter is used
         @author woelk 09/2004 */
    int DownsampleBy4(const Image<InputStorageType>& src, 
                      Image<OutputStorageType>& dst);

     /** assumes that destination is initialized 
         mean low pass filter is used implicitly.
         Destination width = floor(source.GetWidth()/4)
         Destination height = floor(source.GetHeight()/4)
         only for one channel grey images
         also make use of ROI in Source image (fast :-)) )
         @author woelk 09/2004 */
    int DownsampleBy4Grey(const Image<InputStorageType>& src, 
                          Image<OutputStorageType>& dst);


    /** @author evers
     */
    int DownsampleBPoT(const Image<InputStorageType>& Source,
                       Image<OutputStorageType>& Destination);
    /////////////////////////////////////////////////////////////////////////
    //              upsampling functions
    /////////////////////////////////////////////////////////////////////////

    /** generic upsample function
        @author woelk 09/2004 */
    int Upsample(const Image<InputStorageType>& src, 
                 Image<OutputStorageType>& dst);

    /** Upsamples src to dst of size width by height.
        width > src.GetWidth() && height > src.GetHeight() is asserted.
        ROI not supported.
        @author ? */
    int Upsample(const Image<InputStorageType>& src, 
                 Image<OutputStorageType>& dst,
                 const unsigned newwidth, const unsigned newheight);

    /** Upsamples src to dst of size width by height.
        Only for one channel grey images.
        ROI not supported.
        width > src.GetWidth() && height > src.GetHeight() is asserted.
        @author ? */
    int UpsampleGrey(const Image<InputStorageType>& src, 
                     Image<OutputStorageType>& dst,
                     const unsigned newwidth, const unsigned newheight);

    /////////////////////////////////////////////////////////////////////////

         /** assumes that destination is initialized 
        dst width = 2*src.GetWidth()
        dst height = 2*src.GetHeight()
        only for interleaved images
        for images with arbitrary channel count
        @author woelk 04/2004 */
    int UpsampleBy2(const Image<InputStorageType>& src, 
                    Image<OutputStorageType>& dst);

    /** assumes that destination is initialized 
        dst width = 2*src.GetWidth()
        dst height = 2*src.GetHeight()
        only for one channel images
        @author woelk 01/2003 */
    int UpsampleBy2Grey(const Image<InputStorageType>& src, 
                        Image<OutputStorageType>&dst);

    /////////////////////////////////////////////////////////////////////////
    //              get/set functions
    /////////////////////////////////////////////////////////////////////////

    inline void SetFactor(double factor) { *_RescaleFactor=factor; }

    inline double GetFactor() const { return *_RescaleFactor; }

    inline void SetLowPassType(int lptype) 
    { *_RescaleMeanType=lptype; _UpdateMean();}
    
    inline int GetLowPassType() const { return *_RescaleMeanType; }
    
  protected:
    // the mean filter
    FilterNToN<InputStorageType, OutputStorageType> *_lowpass;
    Gauss<InputStorageType, OutputStorageType> _gauss;
    Binomial<InputStorageType, OutputStorageType> _binomial;
    Mean<InputStorageType, OutputStorageType> _mean;
    
    // parameter 
    // *_RescaleFactor<1.0 means upsampling
    // *_RescaleFactor>1.0 means downsampling
    double *_RescaleFactor; 
    int *_RescaleMeanType;
    double *_RescaleMeanSigmaFactor;
    // parameters of Mean and Gauss
    int *_MeanSize;
    double *_GaussSigma;
    double *_GaussRatio;

    /** fills the parameter
        @author woelk 09/2004 */
    void _AddParam(Param& para, const int grp, const std::string& postfix);

    /** updates _mean using _RescaleMeanType
        @author woelk 09/2004 */
    void _UpdateMean();

    /** updates _mean using _RescaleMeanType
        and calls _mean->Filter(src, dst)
        @author woelk 09/2004 */
    int _ApplyMeanFilter(const Image<InputStorageType>& src, 
                         Image<OutputStorageType>& dst, const double factor);

    /** Assumes dst is initialized and ROI is set.
        For every pixel (x, y) in ROI of dst, a bilinear lookup at the 
        coordinates (x*src.Width/dst.Width, y*src.Height/dst.GetHeight)
        is performed.
        Only for one channel grey images
        @author woelk 09/2004 */
    int _FillInterpolatedGrey(const Image<OutputStorageType>& src, 
                              Image<OutputStorageType>& dst);

    /** Same as _FillInterpolatedGrey but for multi channel interleaved 
        images. 
        @author woelk 09/2004 */
    int _FillInterpolatedColor(const Image<OutputStorageType>& src, 
                               Image<OutputStorageType>& dst);
 
  }; // class

#ifdef WIN32
  template <> int Rescale<unsigned char, unsigned char>::
  _FillInterpolatedGrey(const Image<unsigned char>& src, 
			Image<unsigned char>& dst);

template <>
int Rescale<unsigned char, unsigned char>::
_FillInterpolatedColor(const Image<unsigned char>& src, 
                       Image<unsigned char>& dst);

template <> 
int Rescale<unsigned char, unsigned char>::
DownsampleBy2Color(const Image<unsigned char>& Source, 
                   Image<unsigned char>& Destination);

template <>
int Rescale<unsigned char, unsigned char>::
DownsampleBy2Grey(const Image<unsigned char>& Source, 
		  Image<unsigned char>& Destination);



template <>
int Rescale<unsigned char, unsigned char>::
DownsampleBy4Grey(const Image<unsigned char>& Source, 
		  Image<unsigned char>& Destination);

//#ifdef BUILD_IMAGE_USHORT
template <>
int Rescale<unsigned short, unsigned short>::
_FillInterpolatedGrey(const Image<unsigned short>& src,
                      Image<unsigned short>& dst);

template <> 
int Rescale<unsigned short, unsigned short>::
DownsampleBy2Color(const Image<unsigned short>& Source, 
                   Image<unsigned short>& Destination);
template <>
int Rescale<unsigned short, unsigned short>::
DownsampleBy2Grey(const Image<unsigned short>& Source, 
		  Image<unsigned short>& Destination);
template <>
int Rescale<unsigned short, unsigned short>::
DownsampleBy4Grey(const Image<unsigned short>& Source, 
		  Image<unsigned short>& Destination);

template <>
int Rescale<unsigned short, unsigned short>::
_FillInterpolatedColor(const Image<unsigned short>& src, 
                       Image<unsigned short>& dst);

//#endif //BUILD_IMAGE_USHORT
//
//
//
#endif // WIN32


} // namespace

#endif // __Rescale_hh__
