/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Conic2D.hh"
#include <float.h>
#include "MathAlgo/PolynomialSolve.hh"
#include "MathAlgo/SVD.hh"
#define SVD_ZERO_THRESH DBL_EPSILON * 16

using namespace BIAS;
using namespace std;

double Conic2D::GetDiscriminant() {
  return ((*this)[0][0]*(*this)[1][1]-(*this)[0][1]*(*this)[0][1]);
}

bool Conic2D::IsEllipse() {
  // degenerate case ?
  if (fabs(det())<=DBL_EPSILON) {
    // rank < 3
    //cout << "rank<3, Det=" <<det()<<endl;
    return false;
  }

  if (GetSignature() != 1) {
    // Degenerated case 
    //cout << "Degenerated case" <<GetSignature()<<endl;
    return false;
  }

  // parabola or hyperbola ?
  if (GetDiscriminant()<-DBL_EPSILON) {
    //cerr <<"GetDiscriminant"<<endl;
    return false;
  }
  if ((*this)[0][0]+(*this)[1][1]>0.0) {
    // ok, should be ellipse
    return true;
  }
  // must be empty set
  //cerr <<(*this)[0][0] << "/" << (*this)[1][1]<<endl;
  return false;
}

void Conic2D::Draw(Image<unsigned char>& img) {
  const unsigned int uiHeight = img.GetHeight();
  const unsigned int uiWidth  = img.GetWidth();
  register unsigned char** ppI = img.GetImageDataArray();

  HomgPoint2D p(0,0);
  p.Homogenize();
  register int LastSign = 0;
  double dist;
  // dumb drawing algorithm (improve this using parameterized drawing)
  // run over all pixels and check where the sign changes
  // wont catch all ellipses with some radius<1
  for (register unsigned int y=0;y<uiHeight; y++) {
    p[0] = -1; p[1] = y;
    dist = LocatePoint(p);
    if (dist>0.0) LastSign = 1;
    else if (dist<0.0) LastSign = -1;
    else LastSign = 0 ;
    for (register unsigned int x=0; x<uiWidth; x++) {
      p[0] = x; p[1] = y;
      dist = LocatePoint(p);
      if (dist<0.0) {
        if (LastSign>0) { 
          ppI[y][x] = (ppI[y][x]>127)?0:255;
          cout << "D";
        }
        LastSign = -1;
      } else if (dist>0.0) {
        if (LastSign<0) {
          ppI[y][x] = (ppI[y][x]>127)?0:255;
          cout << "D";
        }
        LastSign = 1;
      } else {
        LastSign = 0;
        ppI[y][x] = (ppI[y][x]>127)?0:255;
        cout << "D";
      }
    }
  }
  // run into perpendicular direction to catch x-parallel lines
  for (register unsigned int x=0;x<uiWidth; x++) {
    p[0] = x; p[1] = -1;
    dist = LocatePoint(p);
    if (dist>0.0) LastSign = 1;
    else if (dist<0.0) LastSign = -1;
    else LastSign = 0 ;
    for (register unsigned int y=0; y<uiHeight; y++) {
      p[0] = x; p[1] = y;
      dist = LocatePoint(p);
      // cout << dist <<" ";
      if (dist<0.0) {
        if (LastSign>0) { 
          ppI[y][x] = (ppI[y][x]>127)?0:255;
           cout << "D";
        }
        LastSign = -1;
      } else if (dist>0.0) {
        if (LastSign<0) {
          ppI[y][x] = (ppI[y][x]>127)?0:255;
          cout << "D";
        }
        LastSign = 1;
      } else {
        LastSign = 0;
        ppI[y][x] = (ppI[y][x]>127)?0:255;
        cout << "D";
      }
    }
  }
}

void Conic2D::SetEllipse(const HomgPoint2D& Center, const double& dAngle,
                         const double& radius_a, const double& radius_b) {
  
  SetZero();
  (*this)[0][0] = 1.0 / (radius_a*radius_a);
  (*this)[1][1] = 1.0 / (radius_b*radius_b);
  (*this)[2][2] = -1;
  
  Matrix3x3<double> Transform, Rotation;

  // transfer ellipse center into origin
  Transform.SetIdentity();
  Transform[0][2] = -Center[0];
  Transform[1][2] = -Center[1];
  Transform[2][2] =  Center[2];

  // rotate by dAngle around origin
  Rotation.SetIdentity();
  Rotation[1][1] =   Rotation[0][0] = cos(dAngle);
  Rotation[0][1] = -(Rotation[1][0] = sin(dAngle));
    
  // first transfer, than rotate
  Transform = Rotation * Transform;
 
  (*this) = Transform.Transpose() * (*this) * Transform;
  // Point * (*this) * Point now means:
  // (R * (T * Point)) * (this_canonic) * (R * (T * Point))
  // where this_canonic = diag(radius_a^2, radius_b^2, -1) is the unrotated
  // ellipse around the origin
}

bool Conic2D::GetEllipseParameters(HomgPoint2D& Center, 
                                   double& dAngle,
                                   double& radius_a, 
                                   double& radius_b) {
  if (!IsEllipse()) {
    BIASERR("Conic is no ellipse! Make sure Det=-1");
    return false;
  }

  /* ellipse conic is symmetric matrix of form:

     a b d
     b c e
     d e f

  */ 

  double&  a = (*this)[0][0];
  const double   b = 0.5*((*this)[1][0]+(*this)[0][1]);
  double&  c = (*this)[1][1];
  const double  d = 0.5*((*this)[2][0]+(*this)[0][2]); 
  const double  e = 0.5*((*this)[2][1]+(*this)[1][2]); 
  // double&  f = (*this)[2][2];

  double dDisc = GetDiscriminant();
  if (fabs(dDisc) <= DBL_EPSILON)   {
    // degenerate case
    return false;
  }

  Center[0] = -(c * d - b * e) / dDisc;
  Center[1] = -(a * e - b * d) / dDisc;
  Center[2] = 1;

  double q0 = LocatePoint(Center);

  double theta, lambda1, lambda2;
  double p = (a + c) / 2.0;
  double q = (a - c) / 2.0;
  double r = sqrt(q * q + b * b);
  if ((fabs(b) <= DBL_EPSILON) && (fabs(q) <=DBL_EPSILON))
    theta = 0.0;
  else
    theta = 0.5 * atan2(b, q);
 
  dAngle = -theta;

  lambda1 = p + r;
  lambda2 = p - r;


  if ((fabs(lambda1) <= DBL_EPSILON) || 
      (fabs(lambda2) <= DBL_EPSILON) || 
      (fabs(q0) <= DBL_EPSILON)) {
    // degenerate
    return false;
  }

  lambda1 /= -q0;
  lambda2 /= -q0;

  radius_a = 1.0 / sqrt(fabs(lambda1));
  radius_b = 1.0 / sqrt(fabs(lambda2));
  return true;
}


void Conic2D::SetQuadricProjection(const Quadric3D &Q, const PMatrix& P) {
  // get the dual quadric of Q since its projection is simple
  Matrix4x4<double> DualQuadric;
  DualQuadric = Q.GetDualQuadric();
 
  // "project" the dual quadric to the dual conic
  Conic2D DualConic;
  DualConic = P * DualQuadric * P.Transpose();
  
  // get the actual conic from the dual conic
  (*this) = DualConic.GetDualConic();

  // set determinant to -1, so we can separate inside/outside in ellipses
  Normalize();
}

unsigned int Conic2D::GetSignature() {
  // do singular value decomposition of this
  SVD svd(*this);
  Vector3<double> S;
  Matrix3x3<double> VT;
  S = svd.GetS();
  VT = svd.GetVT();
  // now we need the eigenvalues (with sign !)
  // compute this by projecting the (normalized) eigenvectors onto their images
  Vector3<double> eigenvector, eigenimage;
  int Signature = 0;
  for (unsigned int i=0; i<3; i++) {
    // get ith eigenvector and its image
    VT.GetRow(i,eigenvector);
    eigenimage = (*this) * eigenvector; 
    // compute scalar product, we are not "thinking" homogenized now
    // this is simply an analysis of a 3x3 matrix !
    S[i] = eigenimage * eigenvector;
    if (S[i]<-SVD_ZERO_THRESH) Signature--;
    else if (S[i]>SVD_ZERO_THRESH) Signature++;
  }
  
  // return number of eigenvalues with one sign minus number of eigenvalues
  // with opposite sign, choosing signs such that difference is not negative
  if (Signature<0) return (unsigned int)(-Signature);
  return (unsigned int)Signature;
}



bool Conic2D::IntersectsCircle(const HomgPoint2D& C, double Radius) {
  // code has been copied from maple, see below

  vector<double> Coefficients;
    
  const double r = Radius;
  const double u = C[0]/C[2];
  const double v = C[1]/C[2];
  
  const double a = (*this)[0][0];
  const double b = (*this)[1][1];
  const double c = (*this)[0][1] + (*this)[1][0];
  const double d = (*this)[0][2] + (*this)[2][0];
  const double e = (*this)[1][2] + (*this)[2][1];
  const double f = (*this)[2][2];

  // solve quartic equation
  Coefficients.resize(5);
    
  // coefficients copied from maple, see below
  Coefficients[4] = -2*a*b + b*b + a*a + c*c;   

  Coefficients[3] = 2*u*c*a + 2*u*c*b - 4*a*a*v + 2*d*c + 2*e*b + 4*b*a*v 
    - 2*v*c*c  - 2*a*e;

  Coefficients[2] = 2*f*b + 2*u*c*e - 2*a*a*r*r  + 2*a*a*u*u  + v*v*c*c  
    - r*r*c*c  + u*u*c*c  + e*e + 2*u*d*a + d*d  + 2*b*a*u*u 
    + 4*e*a*v - 2*a*f - 4*v*c*d - 4*u*c*a*v
    + 6*a*a*v*v  + 2*u*d*b - 2*b*a*v*v  + 2*b*a*r*r;

  Coefficients[1] = - 2*r*r*c*d - 4*a*a*u*u*v + 2*u*c*f - 2*e*a*v*v
    + 2*e*a*r*r  + 2*v*v*c*d + 2*f*e - 4*u*d*a*v + 4*f*a*v
    - 2*v*d*d + 2*u*u*u*c*a + 2*u*c*a*v*v - 2*u*c*a*r*r 
    + 2*e*a*u*u + 2*u*d*e + 4*a*a*v*r*r  - 4*a*a*v*v*v 
    + 2*u*u*c*d;

  Coefficients[0] =  2*u*d*f + 2*a*a*u*u*v*v - 2*a*a*u*u*r*r - 2*a*a*v*v*r*r
    + 2*u*u*u*d*a + 2*f*a*u*u - 2*f*a*v*v + 2*f*a*r*r + f*f 
    + a*a*u*u*u*u + a*a*v*v*v*v + a*a*r*r*r*r + u*u*d*d 
    + v*v*d*d - r*r*d*d + 2*u*d*a*v*v - 2*u*d*a*r*r;
  
     
  PolynomialSolve ps;
  ps.CheckCoefficients(Coefficients);
  return (ps.HasRealSolution(Coefficients));
  // solution is y component of point

/* maple code that has been used for IntersectsCircle:
   
eqns := { a*x^2+b*y^2+c*x*y+d*x+e*y+f=0, (x-u)*(x-u)+(y-v)*(y-v)=r*r };

                    2       2                              2      2      2
                  %1  b - %1  a + %1 e + 2 %1 a v + f - a u  - a v  + a r
  {y = %1, x = - --------------------------------------------------------}
                                       %1 c + d + 2 a u



                    2    2    2            4
     %1 := RootOf((b  + a  + c  - 2 a b) _Z  +

            2              2
    (- 2 v c  + 2 d c - 4 a  v + 2 u c b + 2 e b + 2 u c a + 4 b a v - 2 a e)

      3         2  2          2                2  2          2      2  2
    _Z  + (- 2 a  r  - 2 b a v  + 2 d a u + 2 a  u  + 2 b a u  + 6 a  v

        2  2                                  2                2  2          2
     + v  c  + 4 v a e + 2 u c e - 4 v d c + e  - 4 u c a v + u  c  + 2 b a r

                          2  2            2    2           2
     - 2 a f + 2 u d b - r  c  + 2 b f + d ) _Z  + (2 e a u  + 4 f a v

          2  3                    2            2        2                2
     - 4 a  v  + 2 f e + 2 u c a v  - 2 u c a r  - 2 v d  + 2 u d e + 2 v  c d

                      3                        2      2              2
     - 4 u d a v + 2 u  c a + 2 u c f + 2 e a r  - 2 r  c d - 2 e a v

          2          2  2        2    2                    3              2
     + 2 u  c d - 4 a  u  v + 4 a  v r ) _Z + 2 u d f + 2 u  d a - 2 f a v

        2      2  2  2    2  4    2  4    2  4    2  2    2  2    2  2
     + f  + 2 a  u  v  + a  u  + a  v  + a  r  + u  d  + v  d  - r  d

                2            2      2  2  2      2  2  2          2          2
     + 2 u d a v  - 2 u d a r  - 2 a  v  r  - 2 a  u  r  + 2 f a u  + 2 f a r )

*/

}
