/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __CovMatrix3x3_hh__
#define __CovMatrix3x3_hh__

#include <Base/Math/Matrix3x3.hh>
#include <Base/Geometry/PMatrixBase.hh>
#include <MathAlgo/SVD.hh>

#define COVMATRIX3X3_TYPE double

namespace BIAS {
  /** @class CovMatrix3x3
      @ingroup g_math
      @ingroup g_geometry
      @brief class for 3x3 covariance matrices

      This class represents a 3x3 convariance matrix with the capability of
      doing a PCA (principle component analysis).

      @attention For performance reasons, the PCA is only done once and saved,
      subsequent calls to GetPCA() only return a copy.
      If the covariance matrix is changed, you must call InvalidatePCA()
      to force a recomputation at the next GetPCA()-Call ! */

  class CovMatrix3x3 : public Matrix3x3<COVMATRIX3X3_TYPE>
  {
  public:
    CovMatrix3x3();

    ~CovMatrix3x3();

    CovMatrix3x3(const CovMatrix3x3& mat);
    
    CovMatrix3x3(const Matrix3x3<COVMATRIX3X3_TYPE> & mat) 
      { (*this) = mat; };
    

    int GetPCA(Vector3<COVMATRIX3X3_TYPE>& S, 
               Matrix3x3<COVMATRIX3X3_TYPE>& VT);
    
    int GetS(Vector3<COVMATRIX3X3_TYPE>& S);
    
    CovMatrix3x3& operator=(const CovMatrix3x3& mat); 
    
    CovMatrix3x3& operator=(const BIAS::Matrix3x3<COVMATRIX3X3_TYPE>& mat)
      { 

         
        Matrix3x3<COVMATRIX3X3_TYPE>::operator=(mat);
      _IsDecomposed = false;  return (*this); };


    /** Computes the bounding box x=[xmin,xmax] , y=[ymin,ymax] in image  as 
        determined from the projected covariance matrix covmat. multiplied with
        scale
        @author frahm, woelk 02 2003 */
    int GetSearchArea(PMatrixBase& P, HomgPoint3D& point3D, 
                      Vector2<double>& x, Vector2<double>& y, double scale);

    /** Computes the bounding box x=[xmin,xmax] , y=[ymin,ymax] in image  as 
        determined from the projected covariance matrix covmat. 
        A min Search Area can be specified with minSearchArea(which is the
        radius of the uncertainty sphere).
        @author Daniel Grest, Feb 2003
    */
    int GetSearchArea(PMatrixBase& P, HomgPoint3D& point3D, 
		      Vector2<double>& x, Vector2<double>& y, 
		      double scale, double minSearchArea);


    /** Call this function to indicate that the matrix fields have changed
        The saved PCA is declared invalid and the next call to GetPCA forces
        a recomputation
        @author koeser 09/2003 */
    inline void InvalidatePCA() { _IsDecomposed = false; };

  protected:
    /// flag indicating whether the stored PCA is still valid
    bool _IsDecomposed;
 
     
      
#ifdef BIAS_DEBUG
    /// saves matrix fields to make consistency checks possible later
    Matrix3x3<COVMATRIX3X3_TYPE> ConsistencyBackup_;
#endif
   
    SVD _svd;
    Vector3<COVMATRIX3X3_TYPE> _S;
    Matrix3x3<COVMATRIX3X3_TYPE> _U, _VT;

    int _Decompose();
  }; // class



} // namespace


#endif // __CovMatrix3x3_hh__
