/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "HMatrix.hh"

#include <MathAlgo/SVD.hh>
#include <MathAlgo/Lapack.hh>




using namespace BIAS;
using namespace std;

HMatrix::HMatrix()
  : HMatrixBase()
{}

HMatrix::HMatrix(Matrix3x3<double> Mat)
  : HMatrixBase(Mat)
{}


HMatrix& HMatrix::operator=(const HMatrix& mat)
  
{
  HMatrixBase::operator=(mat);
  _Hinv=mat.GetHinv();
  return *this;
}

HMatrix::~HMatrix()
{}

int HMatrix::Map(Image<unsigned char> &source, Image<unsigned char> &dest)
{
  int res=0;
#ifdef BIAS_DEBUG
  if (source.GetChannelCount() != 1){
    BIASERR("unable to map image with more than one channel "
	    <<source.GetChannelCount());
    return -2;
  }    
#endif

  if (!dest.IsEmpty())
    dest.Release();
  SVD svd(*this);
  _Hinv=svd.Invert();
  unsigned char **destida;
  HomgPoint2D dstpt, srcpt;

  if (!dest.IsEmpty()){
    if (!source.SamePixelAndChannelCount(dest)){
      dest.Release();
      dest.Init(source.GetWidth(), source.GetHeight(), 
                source.GetChannelCount());
    }
  } else {
    dest.Init(source.GetWidth(), source.GetHeight(), source.GetChannelCount());
  }
  
  destida=dest.GetImageDataArray();

  for (register unsigned int x=0; x<dest.GetWidth(); x++){
    for (register unsigned int y=0; y<dest.GetHeight(); y++){
      dstpt.Set((double)x,(double)y);
      srcpt=_Hinv*dstpt;
      srcpt.Homogenize();
      if (!srcpt.IsAtInfinity()){
        if (srcpt[0] > 0 && srcpt[0] < source.GetWidth()-1 &&
            srcpt[1] > 0 && srcpt[1] < source.GetHeight()-1){
          destida[y][x]=
            (unsigned char)rint(source.BilinearInterpolationGrey(srcpt[0], 
                                                                 srcpt[1]));
        }
      }
    }
  }

  return res;
}



int HMatrix::Map(Image<unsigned char> &source, Image<unsigned char> &dest,
		 int origin[2])
{
  int res=0;
#ifdef BIAS_DEBUG
  if (source.GetChannelCount() != 1){
    BIASERR("unable to map image with more than one channel "
	    <<source.GetChannelCount());
    return -2;
  }    
#endif

  if (!dest.IsEmpty())
    dest.Release();
  HomgPoint2D corners[4];
  HomgPoint2D mappedcorners[4];
  double maxx=0.0, minx=0.0, maxy=0.0, miny=0.0;
  unsigned int newwidth, newheight;
  SVD svd(*this);
  _Hinv=svd.Invert();
  unsigned char **destida;
  HomgPoint2D dstpt, srcpt;

  corners[0][0]=corners[0][1]=0.0;
  corners[1][0]=(double)source.GetWidth();
  corners[1][1]=0.0;
  corners[2][0]=(double)source.GetWidth();
  corners[2][1]=(double)source.GetHeight();
  corners[3][0]=0.0;
  corners[3][1]=(double)source.GetHeight();
  
  for (int i=0; i<4; i++) {
    corners[i][2]=1.0;
    mappedcorners[i]=(*this)*corners[i];
    mappedcorners[i].Homogenize();
    if (i==0){
      maxx = minx = mappedcorners[0][0];
      maxy = miny = mappedcorners[0][1];
    } else {
      maxx = (mappedcorners[i][0]>maxx)?(mappedcorners[i][0]):(maxx);
      minx = (mappedcorners[i][0]<minx)?(mappedcorners[i][0]):(minx);
      maxy = (mappedcorners[i][1]>maxy)?(mappedcorners[i][1]):(maxy);
      miny = (mappedcorners[i][1]<miny)?(mappedcorners[i][1]):(miny);
    }
  }

  cerr << minx << " <= x <= " << maxx << "\t" << miny << " <= y <= " << maxy 
       << endl;

  origin[0]=(int)floor(minx);
  origin[1]=(int)floor(miny);

  newwidth=(unsigned int)ceil(maxx-minx);
  newheight=(unsigned int)ceil(maxy-miny);
  cerr << "width: " << newwidth << "\theight: " << newheight << endl;
  dest.Init(newwidth, newheight);
  destida=dest.GetImageDataArray();

  for (register unsigned int x=0; x<dest.GetWidth(); x++){
    for (register unsigned int y=0; y<dest.GetHeight(); y++){
      dstpt.Set((double)x+(double)origin[0],(double)y+(double)origin[1]);
      srcpt=_Hinv*dstpt;
      srcpt.Homogenize();
      if (!(srcpt[0] < 0 || srcpt[0] > source.GetWidth()-1 ||
            srcpt[1] < 0 || srcpt[1] > source.GetHeight()-1 ||
            fabs(srcpt[2] - 1.0) > 1e-9)){
        destida[y][x]=
          (unsigned char)rint(source.BilinearInterpolationGrey(srcpt[0], 
							       srcpt[1]));
      }
    }
  }

  return res;
}

int HMatrix::Compute(std::vector<BIAS::HomgPoint2D>& frompoint, 
            std::vector<BIAS::HomgPoint2D>& topoint){

  int result = 0;
  if ((int) frompoint.size()<4){
    BIASERR("at least 4 points correspondences are needed for HMatrixBase "
            <<"computation");
    return -1;
  }

  Matrix<double> A(2*frompoint.size(), 9);
  Vector<double> S;
  Matrix<double> U, VT;

  register unsigned int p=0, p2=0;
  for (p=0, p2=0; p<2*frompoint.size(); p+=2, p2++){
 
    A[p][0] = A[p][1] = A[p][2] = 0.0;
    A[p+1][3] = A[p+1][4] = A[p+1][5] = 0.0;

    A[p+1][0] = topoint[p2][2] * frompoint[p2][0];
    A[p+1][1] = topoint[p2][2] * frompoint[p2][1];
    A[p+1][2] = topoint[p2][2] * frompoint[p2][2];

    A[p][3] = - A[p+1][0];
    A[p][4] = - A[p+1][1];
    A[p][5] = - A[p+1][2];

    A[p][6] = topoint[p2][1] * frompoint[p2][0];
    A[p][7] = topoint[p2][1] * frompoint[p2][1];
    A[p][8] = topoint[p2][1] * frompoint[p2][2];
    
    A[p+1][6] = - topoint[p2][0] * frompoint[p2][0];
    A[p+1][7] = - topoint[p2][0] * frompoint[p2][1];
    A[p+1][8] = - topoint[p2][0] * frompoint[p2][2];
  }
  
  result = General_singular_value_decomposition(A, S, U, VT);
 
  if (result == 0){
    for (p=0; p<9; p++){
      GetData()[p] = VT[8][p];
    }
  }

  return result;
}

int 
HMatrix::Compute(unsigned int numpoints, Vector2<double> frompoint[], 
                 Vector2<double> topoint[])
{
  HomgPoint2D *from=new HomgPoint2D[numpoints], *to=new HomgPoint2D[numpoints];
  for (register unsigned int i=0; i< numpoints; i++){
    from[i][0]=frompoint[i][0];
    from[i][1]=frompoint[i][1];
    from[i][2]=1.0;
    to[i][0]=topoint[i][0];
    to[i][1]=topoint[i][1];
    to[i][2]=1.0;
  }
  int res=Compute(numpoints, from, to);
  delete[] from;
  delete[] to;
  return res;
}

int HMatrix::Compute(unsigned int numpoints, HomgPoint2D frompoint[], 
                     HomgPoint2D topoint[])
{
  int result = 0;
  if (numpoints<4){
    BIASERR("at least 4 points correspondences are needed for HMatrixBase "
            <<"computation");
    return -1;
  }

  Matrix<double> A(2*numpoints, 9);
  Vector<double> S;
  Matrix<double> U, VT;

  register unsigned int p=0, p2=0;
  for (p=0, p2=0; p<2*numpoints; p+=2, p2++){ 
    A[p][0] = A[p][1] = A[p][2] = 0.0;
    A[p+1][3] = A[p+1][4] = A[p+1][5] = 0.0;

    A[p+1][0] = topoint[p2][2] * frompoint[p2][0];
    A[p+1][1] = topoint[p2][2] * frompoint[p2][1];
    A[p+1][2] = topoint[p2][2] * frompoint[p2][2];

    A[p][3] = - A[p+1][0];
    A[p][4] = - A[p+1][1];
    A[p][5] = - A[p+1][2];

    A[p][6] = topoint[p2][1] * frompoint[p2][0];
    A[p][7] = topoint[p2][1] * frompoint[p2][1];
    A[p][8] = topoint[p2][1] * frompoint[p2][2];
    
    A[p+1][6] = - topoint[p2][0] * frompoint[p2][0];
    A[p+1][7] = - topoint[p2][0] * frompoint[p2][1];
    A[p+1][8] = - topoint[p2][0] * frompoint[p2][2];
  }
  
  result = General_singular_value_decomposition(A, S, U, VT);
 
  if (result == 0){
    /* the solution is the unit singular vector corresponding to 
       the smallest singular value. Since S is sorted in descending order,
       h is in the last row of VT */
    //     double length = 0.0;
    //     for (p=0; p<9; p++){
    //       length += VT[8][p] * VT[8][p];
    //     }

    //     length = sqrt(length);

    // //     //BIASDOUT(D_HM_COMP," norm of h: " << length);
    // //     cout << " norm of h: " << length << endl;
    
    //     for (p=0; p<9; p++){
    //       GetData()[p] = VT[8][p] / length;
    //     }
    for (p=0; p<9; p++){
      GetData()[p] = VT[8][p];
    }
  }

  return result;
}


int HMatrix::ComputeMLESAC(std::vector<BIAS::HomgPoint2D>& frompoints, 
                           std::vector<BIAS::HomgPoint2D>& topoints,
                           double OutlierThreshold /*=1.25*/){
  BIASERR("HMatrix::ComputeMLESAC only available with vxl !");
  abort();
  // avoid warning of no return value:
  return -1;
}



int HMatrix::ComputeMLESAC(std::vector<BIAS::HomgPoint2D>& frompoints, 
                           std::vector<BIAS::HomgPoint2D>& topoints,
                           std::vector<bool>& inlier,
                           double OutlierThreshold /*=1.25*/){
  BIASERR("HMatrix::ComputeMLESAC only available with vxl !");
  abort();
  // avoid warning of no return value:
  return -1;
}




// JW
void HMatrix::GetMappedRange(const unsigned int OrigWidth, 
                             const unsigned int OrigHeight,
                             BIAS::Vector2<double> & corner,
                             double & sizeWidth, double & sizeHeight) const
{
  double minX, maxX, minY, maxY; // extrema
  
  // rect. coordinates of the 4 src. corners.
  BIAS::HomgPoint2D mapped[4];
  
  // 0  3
  // 1  2
  mapped[0]= (*this) * BIAS::HomgPoint2D(0,         0,          1); 
  // H * '0' left,  upper
  
  mapped[1]= (*this) * BIAS::HomgPoint2D(0,         OrigHeight, 1); 
  // H * '1' left,  lower

  mapped[2]= (*this) * BIAS::HomgPoint2D(OrigWidth, OrigHeight, 1); 
  // H * '2' right, lower

  mapped[3]= (*this) * BIAS::HomgPoint2D(OrigWidth, 0,          1); 
  // H * '3' right, upper
      
  // calc. extrema(xmy) of mapped corners
  // init with first point:
  mapped[0].Homogenize(); // let w==1
  minX=maxX=mapped[0][0]; // x euclid
  minY=maxY=mapped[0][1]; // y suclid  
  // homogenize and calc extrema
  for (unsigned int i=1; i<4; i++) {
    mapped[i].Homogenize();
    // find new extrema
    minX = (mapped[i][0]<minX) ? mapped[i][0] : minX; // x
    maxX = (mapped[i][0]>maxX) ? mapped[i][0] : maxX;    
    minY = (mapped[i][1]<minY) ? mapped[i][1] : minY; // y
    maxY = (mapped[i][1]>maxY) ? mapped[i][1] : maxY;
  };
  
  corner=BIAS::Vector2<double>(minX, minY);
  sizeWidth=maxX-minX;
  sizeHeight=maxY-minY;  
}


double HMatrix::GetMappedError(const HomgPoint2D &PointPicOne,
                               const HomgPoint2D &PointPicTwo) const
{
  
  HomgPoint2D x1 = PointPicOne;
  HomgPoint2D x2 = PointPicTwo;
  x1.Homogenize();
  x2.Homogenize();
  HomgPoint2D mapped((*this) * x1);
  mapped.Homogenize();
  return (mapped[0]-x2[0])*(mapped[0]-x2[0]) +
    (mapped[1]-x2[1])*(mapped[1]-x2[1]);
}

