/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "bias_config.h"

#ifndef __GuiGTK_hh__
#define __GuiGTK_hh__

#ifndef BIAS_HAVE_GTK 	 
#  error __FILE__: __LINE__: BIAS_HAVE_GTK not defined. Please recompile BIAS with USE_GTK enabled if you want to use GuiGTK.
#endif

#include <fstream>
#include <gtk/gtk.h>
#include <string>
#include <pthread.h>


#include "GuiBase.hh"

#define D_GUI_GDK 0x00010000

namespace BIAS{
  /** @class GuiGTK
      @ingroup g_gui
      @brief Gui based on gtk is a simple windowing environment...
      
      Usage:
      - get an object

      - call ShowImage() to visualize an image

      - use CheckForKeyEvent() to have your programm react on pressed keys
  */
  class GuiGTK : public GuiBase
  {
  public:
    GuiGTK(bool ShowMousePos = true);
    /// as standard constructor, also sets Title_ 
    GuiGTK(std::string const& Title, bool ShowMousePos = true);

    GuiGTK(const GuiGTK& gui);

    virtual ~GuiGTK();

    void SetTitle(std::string const& Title);
 
    /// stops the gtk_main loop and closes the window
    int DestroyWindow();
 
    // can optionaly be called to forward options like --geometry to gtk
    void InitGtk(int *argc, char **argv[]);

   protected:
    guchar *Buffer_;
    GtkWidget *Window_;
    GtkWidget *DrawingArea_;
    GtkWidget *Label_;
    GtkWidget *Vbox_;

    static pthread_t MainThread_;
    static bool ThreadsInitialized_;
    static bool MainThreadRunning_;
    static bool GtkInitialized_;
    static unsigned short int WindowCount_;
    static pthread_mutex_t MainThreadRunningMutex_;
    static pthread_mutex_t GtkInitializedMutex_;
    static pthread_mutex_t WindowCountMutex_;
    static pthread_mutex_t BufferMutex_;
    
    void Init_(bool ShowMousePos);

    /// gtk needs to be initialized 
    void InitGtk_();

    /// self explaining
    inline bool IsAllocated_()
    { return ( (Width_!=0) && (Height_!=0) && (nChannels_!=0) ); };

    /// self explaining
    int AllocateBuffer_(Image<unsigned char>& image);

    /// self explaining
    int FreeBuffer_();


    /// updates the image shown in window
    int ShowConvertedImage_(Image<unsigned char> &image);

    /// initialy creates a window for a given image size
    /// and runs the gtk main loop in a separate thread
    /// if not already running
    /// also allocates memory for Buffer_
    int CreateWindow_(Image<unsigned char> &image);

    /// internal function that handles the redraw
    int Redraw_();

    /// test if internal variables Width_, Height_ and nChannels_
    /// are set correct for image
    /// these 3 variables are used as an indicator if Buffer_ is 
    /// allocated in right size
    inline bool TestImage_(Image<unsigned char> &image)
    { return ( ( Width_ == image.GetWidth() ) &&
               ( Height_ == image.GetHeight() ) &&
               ( nChannels_ == image.GetChannelCount() ) ); };

    /// reallocates memory and set internal variables
    int ReInitWindow_(Image<unsigned char> &image);

    /// callbacks and thread functions must be static
    /// the gtk main loop runs in a separate thread 
    static void* GtkMain(void* arg);
    /// callback for key press event
    static gint KeyPressEventCallback_(GtkWidget *Widget, 
                                       GdkEventKey *Event, 
                                       gpointer user_data);
    /// callback function for expose event
    static gboolean DrawingAreaExposeCallback_(GtkWidget *Widget, 
                                               GdkEventExpose *Event, 
                                               gpointer user_data);

    static gboolean LabelExposeCallback_(GtkWidget *Widget, 
                                         GdkEventExpose *Event, 
                                         gpointer user_data);

    static gint DeleteEvent_(GtkWidget *widget, GdkEvent  *event,
                             gpointer  user_data);

    static void Destroy_(GtkWidget *widget, gpointer user_data);

    static gint MouseMotionCallback_(GtkWidget *Widget, GdkEventMotion *Event, 
                                     gpointer user_data);
    
    static gint MouseButtonPressCallback_(GtkWidget *widget, 
                                          GdkEventButton *event, 
                                          gpointer user_data);

    static gint MouseButtonReleaseCallback_(GtkWidget *widget, 
                                            GdkEventButton *event, 
                                            gpointer user_data);

  };

} // namepsace BIAS
#endif // __Gui_hh__













