#ifndef __ImageCanvas_hh__
#define __ImageCanvas_hh__

#include <bias_config.h>
#include <wx/wx.h>
#include <Image/PyramidImage.hh> 
#include <Base/Debug/Debug.hh>

#ifndef BIAS_HAVE_PTHREADS
#error BIAS_HAVE_PTHREADS not defined. But BIAS::ImageCanvas needs pthreads (on Unix and Win32). Please enable USE_PTHREADS in CMake configure step.
#endif
#include <pthread.h>

#define D_IC_IM (1<<0)


namespace BIAS{

  /** @brief used by ImageCanvas and ScaledImageCanvas, should not be used 
      directly
      @author woelk */
  class ZoomImageCanvas : public wxPanel
  {
  public:
    ZoomImageCanvas(wxWindow* parent, wxStatusBar *sbar=NULL, int sbarid=0,
                    wxWindowID id = -1, const wxPoint& pos = wxDefaultPosition,
                    const wxSize& size = wxSize(320, 240), 
                    long style = wxTAB_TRAVERSAL, 
                    const wxString& name = "ZoomImageCanvas");
    ~ZoomImageCanvas();
    
    void NewImage(BIAS::ImageBase& im, BIAS::Image<unsigned char>& ucim);
    void NewCoo(unsigned x, unsigned y);
    void NewZoom(int zoom);

    inline void SetClearDrawingContext(bool clear)
    { ClearDrawingContext_=clear; }

  protected:
    void UpdateStatusBar(unsigned x, unsigned y);
    void GetZoomImage();

    void OnResize(wxSizeEvent & event);
    void OnPaint(wxPaintEvent & event);
    void OnMouseMove(wxMouseEvent &event);
    void OnMouseLeave(wxMouseEvent &event);
    void OnRightMouse(wxMouseEvent &event);
    void OnLeftMouse(wxMouseEvent &event);
    void OnKey(wxKeyEvent &event);

    wxBitmap *Bitmap_;
    wxImage *wximage_;
    BIAS::Image<unsigned char> ucim_, zim_;
    BIAS::ImageBase origim_;
    wxStatusBar *SBar_;
    int SBarID_;
    // if ClearDrawingContext_=false, the drawing context is not cleared 
    // before a repaint, this should prevent the flickering
    bool ClearDrawingContext_;

    unsigned ZoomFactor_, ZoomX_, ZoomY_; // zoom coo, set from NewCoo
    // upper left corner of ZoomX_, ZoomY_ in Window_
    unsigned ZoomXPos_, ZoomYPos_; 
    int wwidth_, wheight_; // width & height of Window_
    int mousex_, mousey_;

    pthread_mutex_t *wxbitmap_mutex_, *origim_mutex_;

    DECLARE_EVENT_TABLE();
  }; // class ZoomImageCanvas

  /** @brief used by ImageCanvas and ScaledImageCanvas, should not be used 
      directly
      @author woelk */
  class ZoomFrame : public wxFrame
  {
    enum { MIA_ZoomQuitButton };
  public:
    ZoomFrame(wxWindow* parent, const wxString& title);
    ~ZoomFrame();
    
    void NewImage(BIAS::ImageBase& im, BIAS::Image<unsigned char>& ucim,
                  const wxString& name);
    void NewCoo(unsigned x, unsigned y);

    void SetZoom(unsigned z);

  protected:
    void Setup();

    void OnQuit(wxCommandEvent & event);
    void OnClose(wxCloseEvent & event);
    void OnZoomFactor(wxScrollEvent & event);
    //void OnKeyPress(wxKeyEvent &event);

    wxString ImageName_;
    wxString Title_;
    wxButton *QuitButton_;
    ZoomImageCanvas *Window_;
    wxSlider *SLScale_;

    DECLARE_EVENT_TABLE();
  }; // class ZoomFrame
  

  /** @class ImageCanvas
      @ingroup g_gui
      @brief display image in wx application, provides zoom and investigation 
       functionality
      @author woelk */
 
  class ImageCanvas : public wxScrolledWindow , public BIAS::Debug
  {
  public:
    ImageCanvas(wxWindow* parent, wxStatusBar* sbar=NULL,
                int sbarid=0, wxWindowID id = -1, 
                const wxPoint& pos = wxDefaultPosition, 
                const wxSize& size = wxSize(320, 240), 
                long style = wxHSCROLL | wxVSCROLL, 
                const wxString& name = "ScrolledWindow");
    ~ImageCanvas();
    

    void Show(BIAS::Image<unsigned char>& im, std::string name);
    void Show(BIAS::ImageBase& origim, BIAS::Image<unsigned char>& scaledrgbim,
              std::string name);
    void Show(BIAS::PyramidImage<unsigned char>& pim, std::string name);

    

    // to let user access classic hide/show function
    virtual bool Show(bool show=true) {return wxScrolledWindow::Show(show);};

    inline void SetZoomed(bool s) { Zoomed_ = s; }
    inline BIAS::ImageBase *GetOrigImage() { return &origim_; } 

    bool GetMousePos(int &x, int &y);

    inline void SetDefaultZoom(unsigned z)
    { DefaultZoom_=z; }

    inline void SetClearDrawingContext(bool clear)
    { ClearDrawingContext_=clear; }

    inline void SetShowScaled(bool on){ShowScaled_ = on;};
    inline bool GetShowScaled(){return ShowScaled_;};

  protected:
    void UpdateStatusBar(unsigned x, unsigned y);
    void Zoom(unsigned x, unsigned y);
    
    void OnPaint(wxPaintEvent& event);
    void OnSize(wxSizeEvent& event);
    void OnMouseMove(wxMouseEvent &event);
    void OnMouseLeave(wxMouseEvent &event);
    void OnRightMouseButton(wxMouseEvent &event);
    //void OnKeyPress(wxKeyEvent &event);

    wxBitmap *Bitmap_;
    wxImage *wximage_;
    wxStatusBar *SBar_;
    int SBarID_;
    ZoomFrame *MyZoomWin_;
    unsigned width_, height_;
    int mousex_, mousey_;
    bool Zoomed_;
    BIAS::ImageBase origim_;
    BIAS::Image<unsigned char> rgbim_;

    // for scaled-down display
    bool ShowScaled_;
    int ScaledImageSizeX_,ScaledImageSizeY_;

    pthread_mutex_t *wxbitmap_mutex_, *origim_mutex_;

    unsigned DefaultZoom_;
    // if ClearDrawingContext_=false, the drawing context is not cleared 
    // before a repaint, this should prevent the flickering
    bool ClearDrawingContext_;

    std::string ImageName_;
    wxString Title_;

    DECLARE_EVENT_TABLE();
  }; // ImageCanvas


  /** @class ScaledImageCanvas
      @ingroup g_gui
      @brief display float images in wx application, provides zoom, scale, 
      shift  and investigation functionality
      @author woelk */
  class ScaledImageCanvas : public wxPanel
  {
  public:
    ScaledImageCanvas(wxWindow* parent, wxStatusBar* sbar=NULL,
                            int sbarid=0, wxWindowID id = -1, 
                            const wxPoint& pos = wxDefaultPosition, 
                            const wxSize& size = wxDefaultSize, 
                            long style = wxTAB_TRAVERSAL, 
                            const wxString& name = "ScaledScrolledWindow");
    ~ScaledImageCanvas();

    void Show(BIAS::ImageBase& im, std::string name);

    void Show(BIAS::PyramidImage<float>& pim, std::string name);
    
    // to let user access classic hide/show function
    virtual bool Show(bool show=true) {return MiaScWi_->Show(show);};

    inline void SetScaleRange(double min, double max)
    { BIASASSERT(max>min); MinScale_=min; MaxScale_=max; }

    inline void SetOffsetRange(double min, double max)
    { BIASASSERT(max>min); MinOffSet_=min; MaxOffSet_=max; }

    int SetScale(double scale);
    int SetOffset(double offset);

    inline void SetDefaultZoom(unsigned z)
    { MiaScWi_->SetDefaultZoom(z); }

    /** @return  (pointer) to class member ImageCanvas used for display JW */
    ImageCanvas* GetImageCanvas();

    /** @return datat image origim_ class member JW */
    BIAS::ImageBase & GetOrigimg();


  protected:
    ImageCanvas *MiaScWi_; // true data image
    wxSlider *ScaleSL_, *OffsSL_;
    BIAS::ImageBase origim_;
    BIAS::Image<float> scaledf_;
    BIAS::Image<unsigned char> rgbim_, ucim_;
    std::string ImageName_;
    double MinScale_, MaxScale_, MinOffSet_, MaxOffSet_;

    void OnScaleOffset(wxScrollEvent & event);

    void UpdateScaleOffset();
    DECLARE_EVENT_TABLE();
  }; // ScaledImageCanvas


} // namespace


#endif // __ImageCanvas_hh__
