
/*
$Id: CameraParam.hh,v 2.0 2005/02/10 07:39:40 evers Exp $
*/


#ifndef __CameraParameter_inc__
#define __CameraParameter_inc__

#include <vector>
#include <string>

#include <Base/Debug/Error.hh>
#include <Base/Common/XMLIO.hh>
#include <Base/Math/Vector2.hh>

#define CAMERA_PARAMETER_CURRENT_XML_VERSION 1.0

namespace BIAS{

  struct CPDiscreteParam{
    double zoom;
    double focallength;
    // radial/tangential distortion parameters.
    // 4th order model of brown.
    double kc1,kc2,kc3,kc4;
  };


  class CameraParam{

    //    static std::string CameraModelName[];
    // static std::string LensModelName[];

  public:
    CameraParam();
    CameraParam(const CameraParam &cp);

#ifdef BIAS_HAVE_XML2
    CameraParam(const std::string &filename);
    CameraParam(XMLIO &myXML, xmlNodePtr &cameraNode);
#endif //BIAS_HAVE_XML2

    ~CameraParam();

    // @brief default initialization
    int Init();

#ifdef BIAS_HAVE_XML2
    // @brief initialization with camera data file
    int Init(const std::string &filename);
    // @brief initialization with camera node from pre-parsed data file
    int Init(XMLIO &myXML, xmlNodePtr &cameraNode);
#endif //BIAS_HAVE_XML2

    bool IsInited() const;
    
    /** @brief Set to 'true', if camera has a zoom lens.
        @author streckel 06/2004*/
    void SetZoomCamera(const bool zoomCam);

    /** @brief True, if camera has a zoom lens.
        @author streckel 06/2004*/
    bool IsZoomCamera() const;

    /** @brief Set the current camera zoom, the focal length will be 
        calculated.
        @author buck */
    void SetZoom(const double abszoom);

    /** @brief Set the current camera focal length, the zoom will be 
        calculated.
        @author buck */
    void SetFocalLength(const double f);

    /** @brief Set the zoom and focal length of a zoom camera to the 
        default settings.
        @author streckel 06/2004*/
    void SetUnzoomed();

    /** @brief Set the current camera zoom, the focal length will be 
        calculated.
        @author buck */
    double GetZoom() const;

    /** @brief Get the current camera focal length.
        @author buck */
    double GetFocalLength() const;

    /** @brief Set CCD cell-size in microns.
        @author buck */
    void SetCellSizeInMicr(const double CellSizeX, const double CellSizeY);

    /** @brief Set principal point
        @author streckel */
    void SetPrincipal(const double x, const double y);

    /** @brief get camera image width.
        @author buck */
    unsigned int GetWidth() const; 
    /** @brief get camera image height.
        @author buck */
    unsigned int GetHeight() const;

    /** @brief get principal point, x
        @author buck */
    double GetPrincipalX() const; 
    /** @brief get principal point, y
        @author buck */
    double GetPrincipalY() const; 

    inline int GetCamID() const {
      return CamID_;
    }
    
    inline double GetRadius() const{
      if (IsSpherical()) {
        return radius_;
      } else {
        return -1;
      }
    }

    
    inline void GetDefaultUndistortion(double& f,
                                       double& kc1, double& kc2,
                                       double& kc3, double& kc4) const 
    {
      f   = focallengthDef_;
      kc1 = kc1Def_;
      kc2 = kc2Def_;
      kc3 = kc3Def_;
      kc4 = kc4Def_;
    }
    
    inline void GetZoomUndistortion(std::vector<CPDiscreteParam>& vp) const {
      std::vector<CPDiscreteParam>::const_iterator it;
      vp.clear();
      vp.reserve(knownparams_vect_.size());
      for (it=knownparams_vect_.begin();it!=knownparams_vect_.end();it++) {
        vp.push_back(*it);
      }
    }

    inline double GetExposureTime() const {
      return exposureTime_;
    }

    inline double GetIris() const {
      return iris_;
    }

    inline double GetReciprocalFocus() const {
      return reciprocalFocus_;
    }   

    inline double GetAspectRatio() const {
      return aspectratio_;
    }

    inline  bool IsSpherical() const { return LensIsSpherical_;}

    inline  bool IsPerspective() const {return !LensIsSpherical_;}

    /** @brief Set active area to WidthxHeight+xoff+yoff.
        @author buck */
//     int SetActiveArea(const float w,const float h, 
//                       const float xoff, const float yoff);
 
    /** @brief Set subsampling, if required.
        @author buck */
    int SetSubsampling(const float subx, const float suby);

    inline std::vector<double> GetAngleCorrX() const {
      return AngleCorrX_;
    }
    inline std::vector<double> GetAngleCorrY() const {
      return AngleCorrY_;
    }
    inline std::vector<double> GetIlluCorrX() const {
      return IlluCorrX_;
    }
    inline std::vector<double> GetIlluCorrY() const {
      return IlluCorrY_;
    }
    
#ifdef BIAS_HAVE_XML2
    /** @brief Write all data to a camera data parameter file )
        @param Filename Name of the XML-camera-data-file, 
        document tag will be to be "Camera"
        @author streckel 06/2004 */
    int WriteCameraData(const std::string& Filename) const;
    /** @brief Add a camera sub-tree-structure to an allready created 
        XML-tree and write the data from (*this) into the tree - 
        called from the rig parameter class.
        @param myXML points to the XML-Tree
        @param cameraDataNode points to a "CameraData"-node to append the 
        "Camera"-node
        @author streckel 06/2004 */
    int AddCameraDataToRig(XMLIO &myXML, xmlNodePtr &cameraDataNode) const;

    /** @brief Read a camera data parameter file and store the data in (*this)
        @param Filename Name of the XML-camera-data-file, 
        document tag has to be "Camera"
        @author streckel 06/2004 */
    int ReadCameraData(const std::string& Filename);
    /** @brief Read a camera data form a precreated XML-tree and store 
        the data in (*this)  - called from the rig parameter class.
        @param myXML points to the XML-Tree
        @param rigNode points to a "Camera"-node to evaluate
        @author streckel 06/2004 */
    int ReadCameraData(XMLIO &myXML, xmlNodePtr &cameraNode);
#endif //BIAS_HAVE_XML2

    // Settings for specific cameras
    void SetData_EVID31_147094();
    void SetData_OMNITECH_ORIFL_190_3(); 

    void DebugOut() const;   

  private:	
    bool IsInited_;
      
    // Which camera if used?
    std::string CamModel_;
    int CamID_;
    std::string LensModel_;

    // do we have zoom or fixed optics? 
    bool zoomCamera_;

    // Lens can be Spherical or Perspective
    bool LensIsSpherical_;
    
    // width and height of the camera image
    unsigned int width_, height_;
    
    // array of camera calibration constants for zoom cameras
    std::vector<CPDiscreteParam> knownparams_vect_;
    
    // filled by inheriting camera model
    double aspectratio_;

    // size of the single CCD-sensor-cells in meter
    double cellSizeX_, cellSizeY_;

    // Principal Point Pixel Coordinates (one for all zoom settings)
    double principalX_, principalY_;

    // ratio of the focal length to the diameter of a camera lens system
    double iris_;
    
    // reciprocal of the focus distance
    double reciprocalFocus_;

    // exposure time setting
    double exposureTime_;

    // image radius if image is spherical
    double radius_;
        
    // only for perspective lenses
    // standard settings for non zoom lenses or
    // for zoom default state
    // focal length is stored in pixel
    double focallengthDef_;
    double kc1Def_,kc2Def_,kc3Def_,kc4Def_;

    // calibration settings for the current zoom setting of a zoom camera
    // zoom and focallengthZoom are dependend:
    // if zoom is set, focallengthZoom will be calculated and vice versa
    // focal length is stored in pixel
    double zoom_;
    double focallengthZoom_;
    double kc1Zoom_,kc2Zoom_,kc3Zoom_,kc4Zoom_;

    // if only a portion of the CCD is used
//     BIAS::Vector2<float> ActiveArea_; // width and height
//     BIAS::Vector2<float> ActiveAreaOffset_; // ofset from top-left corner
    
    // if camera or frame grabber applies a subsampling
//     BIAS::Vector2<float> Subsampled_;
    
    // illumination vignette correction
    // x is distance in pixel from principal point
    // y is illumination correction factor
    std::vector<double> IlluCorrX_;
    std::vector<double> IlluCorrY_;

    // Only for sphere images:
    // Calibration measurements for angle correction 
    std::vector<double> AngleCorrX_;
    std::vector<double> AngleCorrY_;

    double InterpolateFocalLength_(); // w/ given abszoom
    double InterpolateAbsZoom_(); // w/ given focal length
    void InterpolateDistortionFromIndex_(unsigned int i,double relnorm);

#ifdef BIAS_HAVE_XML2
    int ReadCameraDataV0_8(XMLIO &myXML, xmlNodePtr &cameraNode);
    int AddCameraDataToXMLTree(XMLIO &myXML, xmlNodePtr &rootNode) const;
#endif //BIAS_HAVE_XML2

  };

};

#endif
