/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "HomographyMapping.hh"
#include <Geometry/RMatrix.hh>
#include <Base/Image/ImageConvert.hh>

using namespace BIAS;

template class HomographyMapping<unsigned char, unsigned char>;
template class HomographyMapping<float, float>;
template class HomographyMapping<unsigned char, float>;

#if defined(BUILD_IMAGE_CHAR)
template class HomographyMapping<unsigned char, char>;
template class HomographyMapping<char, char>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class HomographyMapping<unsigned short,unsigned short>;
#endif

#if defined(BUILD_IMAGE_SHORT)
template class HomographyMapping<short, short>;
#endif

#if defined(BUILD_IMAGE_SHORT)&&defined(BUILD_IMAGE_USHORT)
template class HomographyMapping<unsigned short, short>;
#endif

#if defined(BUILD_IMAGE_INT)
template class HomographyMapping<int,int>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class HomographyMapping<unsigned short,float>;
#endif

#if defined(BUILD_IMAGE_USHORT) && defined(BUILD_IMAGE_INT)
template class HomographyMapping<unsigned short,int>;
#endif

#if defined(BUILD_IMAGE_DOUBLE)
template class HomographyMapping<double,double>;
#endif














#undef MASK


template <class InputStorageType, class OutputStorageType>
int HomographyMapping<InputStorageType, OutputStorageType>::
RotateToGrey(Image<InputStorageType>& src, 
             Image<OutputStorageType>& sink,
             float PhiX, float PhiY, float PhiZ, 
             const Matrix3x3<double> &K) {
  Matrix3x3<double> Kinv;
  Matrix3x3<double> KRKinv;
  Image<InputStorageType> source;

  double centerx,centery;
    
  if (!src.IsEmpty()){
      
    if(src.GetColorModel() != ImageBase::CM_Grey)
      ImageConvert::Convert(src, source, ImageBase::CM_Grey);
    else
      source=src;
      
    if(!sink.IsEmpty())
      sink.Release();
      
    sink.Init(source.GetWidth(),source.GetHeight(),source.GetChannelCount());
      
    // Prepare inverse camera mattix
    Kinv.set_identity();
    Kinv[0][0] = 1 / K[0][0];
    Kinv[1][1] = 1 / K[1][1];
    Kinv[0][2] = -K[0][2]/K[0][0];
    Kinv[1][2] = -K[1][2]/K[1][1];
      
    centerx = K[0][0];
    centery = K[1][1];
      
    RMatrix R;
    R.SetXYZ(PhiX,PhiY,PhiZ);
    KRKinv = K*R*Kinv;
      
    /*cout << "K=" << K << endl;
      cout << "R=" << R << endl;
      cout << "Kinv = " << Kinv << endl;
      cout << "KRKinv=" << KRKinv << endl;*/
    
    return MapImage(src, sink, KRKinv);
  } else{
    return -1;
  }
}

template <class InputStorageType, class OutputStorageType>
int HomographyMapping<InputStorageType, OutputStorageType>::
MapImage(const Image<InputStorageType>& source, 
         Image<OutputStorageType>& sink,
         const Matrix3x3<double>& H) {

  BIASERR("Homograpy mapping: Untested code.");

#ifdef BIAS_DEBUG
  if (source.GetChannelCount()!=1){
    BIASERR("only for one channel images");
  }
#endif
  if (!source.SamePixelAndChannelCount(sink))
    sink.Release();
  if (sink.IsEmpty())
    sink.Init(source.GetWidth(), source.GetHeight(), source.GetChannelCount());

  double xsource,ysource;
  Vector3<double> p_sink, p_source;
  p_sink[2] = 1;
  
  OutputStorageType** Data=sink.GetImageDataArray();
  // iterate over sink and calculate the corresponding pix in source
  for (unsigned int y=0; y< sink.GetHeight(); y++) {
    for (unsigned int x=0; x<sink.GetWidth(); x++) {
      // image point in homogenous coords
      p_sink[0] = x; 
      p_sink[1] = y;
      
      //matvectmul(Kinv, p_sink, (float *)Kinv_p_sink);
      //matvectmul(KRt, Kinv_p_sink, (float *)p_source);
      p_source = H*p_sink;
      
      xsource = p_source[0]/p_source[2];
      ysource = p_source[1]/p_source[2];
  
      if ( xsource <= (double)source.GetWidth()-1.0 && 
           ysource <= (double)source.GetHeight()-1.0
           && xsource >= 0.0 && ysource >= 0.0 ) {
        for (unsigned int j=0; j<source.GetChannelCount(); j++) {
          double value = source.BilinearInterpolation(xsource,ysource,j);
#ifdef MASK
          if ((InputStorageType)value == (InputStorageType)0){
            value = 1.0;
            // 	    cout << "found zero value in image" << endl;
          }
          // 	  if ((StorageType)value < (StorageType)2)
          // 	    cout << (int)(StorageType)value << endl;
#endif
          Data[y][x*sink.GetChannelCount()+j] = (OutputStorageType) value;
        }
      } else {
        for (unsigned int j=0; j<source.GetChannelCount(); j++)
          Data[y][x*sink.GetChannelCount()+j] = (OutputStorageType) 0;
      }
    }
  }
  return 0;
}


 


