#ifndef __RIG_PARAMETER_
#define __RIG_PARAMETER_

//BIAS
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Math/Matrix3x3.hh>
#include <Base/Math/Vector3.hh>
#include <Base/Image/ImageBase.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Geometry/RMatrix.hh>
#include <Geometry/PMatrix.hh>
#include <Base/Math/Matrix4x4.hh>

#include <iostream>
#include <fstream>
#include <sstream>
#include <stdio.h>
#include <string>

#include "CameraParam.hh"

#define RIG_PARAMETER_CURRENT_XML_VERSION 1.0

namespace BIAS {

  class RigParameter { 
  public:
    RigParameter() {};

    RigParameter(const RigParameter& RigParam) {
      Camera      = RigParam.Camera;
      Center      = RigParam.Center;
      Rotation    = RigParam.Rotation;
      Calibration = RigParam.Calibration;
      SensorCenter = RigParam.SensorCenter;
      SensorRotation = RigParam.SensorRotation;
    };

    ~RigParameter() {};

    inline unsigned int getRigSize() const { return Center.size(); };

    bool ReadFromCMP(const std::string &filename);

#ifdef BIAS_HAVE_XML2
    /** @brief Write all data camera data and camera positions in the rig
        to a rig parameter file in XML-notation
        @param Filename Name of the XML-rig-data-file, 
        document tag will be to be "Rig"
        @author streckel 06/2004 */
    int WriteRigData(const std::string &filename);
    /** @brief Read a rig parameter file and store the data in (*this)
        the Camera-, Center-, RMatrix-, and KMatrix-Vectors are 
        reset from the file.
        @param Filename Name of the XML-camera-data-file, 
        document tag has to be "Rig"
        @author streckel 06/2004 */
    int ReadRigData(const std::string &filename);
#endif //BIAS_HAVE_XML2

    void SetLadybugCenters(double value = 1.0);


    /** @brief compute mean rigid transformations of virtual camera into 
        rigcameras given a number of rigs
        @param Pvec contains n rigs of size Rigsize, first all cameras of rig1
        cameras must be in world coordinate system
        @param Rigsize size of each rig
        @param WeightOfCurrentParams if not zero, the current parameters
        are averaged with the new ones (weight 1.0)
        @author koeser 03/2004 */
    bool Compute(std::vector<BIAS::PMatrix>& Pvec,
                 const unsigned int &Rigsize,
                 const double& WeightOfCurrentParams = 0.0);

    std::vector< CameraParam >              Camera;
    std::vector< BIAS::Vector3<double> >    Center;
    std::vector< BIAS::RMatrix >            Rotation;
    std::vector< BIAS::KMatrix >            Calibration;
    std::vector< BIAS::RMatrix >            SensorRotation;
    std::vector< BIAS::Vector3<double > >   SensorCenter;

    void SetData_DummyRig();

  protected:
    int ParseView(const std::string &block);
    
    // read old rig datafile version
    int ReadRigDataV0_8(const std::string &filename);

    /** @brief adapts rotation representation in Axis/Angle such that averaging
        against RefAxis and RefAngle is possible without wrap-around problems 
        @author koeser 03/2004 */
    void AdaptAxisAngleRepresentation(BIAS::Vector3<double>& Axis,
                                      double& Angle,
                                      const BIAS::Vector3<double>& RefAxis,
                                      const double& RefAngle);
    
  };
  std::ostream& operator<<(std::ostream &os, BIAS::RigParameter &r);
};




#endif
