/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __SphericalUndistortion_hh__
#define __SphericalUndistortion_hh__

#include <Base/Image/Image.hh>
#include <Base/Math/Vector2.hh>
#include <Base/Math/Vector.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Debug/Debug.hh>

#include <MathAlgo/Interpolator.hh>
#include <Image/CameraParam.hh>

namespace BIAS{
  /** @class SphericalUndistortion
      @ingroup g_filter
      @brief transforms sphere images coming from a fisheye-lens according to 
      different projection models. 

      @author streckel
    
      Input is an image from a fisheye lens looking circular 
      with a black border. 5 types of projection are supported 
      (see: M.M.Fleck "Perspective Projection: The wrong imaging model")

      The algorithm expects a 3rd order calibration polynomial to get a 
      linear relation between viewing angle and distance to circle center.

      The supported projections are:
      PERSPECTIVE      : Perspective projection      - p=k*tan(theta)
      STEREOGRAPHIC    : Stereographic projection    - p=k*tan(theta/2)
      EQUIDISTANT      : Equidistant projection      - p=k*theta
      SINE_LAW         : sine-law projection         - p=k*sin(theta)
      EQUI_SOLID_ANGLE : equi-solid-angle projection - p=k*sin(theta/2)
      p     : projected distance from circle center
      theta : angle between projected point and optical axis
  */

#define DEBUG_SPHEREPROJ_AVG 1
#define DEBUG_SPHEREPROJ_CIRCLE 2

  class SphericalUndistortion : public BIAS::Debug{

  public:

    /** @brief Projection-type constants
        @author Birger Streckel */
    enum{PERSPECTIVE, STEREOGRAPHIC, EQUIDISTANT, SINE_LAW, EQUI_SOLID_ANGLE};
  
    /** @brief default constructor
        @author Birger Streckel */
    SphericalUndistortion();

    /** @brief copy constructor
        @author Birger Streckel */
    SphericalUndistortion(const SphericalUndistortion& sp);

    /** @brief common constructor - initialization with CameraParam-object!
        @author Birger Streckel */
    SphericalUndistortion(const CameraParam& cp);

    /** @brief filter initialization with calibration curve measurements
        @author Birger Streckel */
    SphericalUndistortion(const double& centerX, const double& centerY, 
                          const double& radius,
                          const std::vector<double> &AngleCorrX, 
                          const std::vector<double> &AngleCorrY,
                          const std::vector<double> &IlluCorrX, 
                          const std::vector<double> &IlluCorrY);

    ~SphericalUndistortion();
  
    /** @brief Initialization by setting all necessary parameters 
        (center, radius, Vectors of Angle and Vignette correction 
        measurement points) from the outside */
    int Init(const double& centerX, const double& centerY, 
             const double& radius,
             const std::vector<double> &AngleCorrX, 
             const std::vector<double> &AngleCorrY,
             const std::vector<double> &IlluCorrX, 
             const std::vector<double> &IlluCorrY);

    /** @brief Initialization with CameraParam-object!
        @author Birger Streckel */
   int Init(const CameraParam& cp);

    /** @brief Performs a projection of a spherical image given 
        a projection-type and a zoom factor.

        @param src source image 
        @param dst (output) uninitialized destination image
        @param SizeX, SizeY define the size of the output image
        @param FovX, FovY define the Field Of View angle (in deg) in x and y 
        direction that is used for the creation of the K-Matrix for the 
        projected image. If one of these is =0 then the resulting image will 
        be undistorted.
        @param CenterX, Centery is the pixel in the spherical uncalibrated 
        camera image, that lies in viewing direction. This point will be 
        mapped to the principal point in the projected image.
        @param zoomFactor expands the image by a given factor
        @param projectionType use the projection-type constants
        @author Birger Streckel */
    template <class StorageType>
    int ProjectImage(const Image<StorageType>& src, 
                     Image<StorageType>& dst,
                     const double& SizeX, const double& SizeY, 
                     const double& FovX, const double& FovY,
                     const double& CenterX, const double& CenterY, 
                     const int projectionType, const double& zoomFactor,
                     KMatrix& KMatDst);
    
    /** @brief Performs a projection of a spherical image coordinate
        to a projected and calibrated coordinate
        given a projection-type and a K-matrix.
        @author Birger Streckel */
    int ProjectCoord(const int projectionType, const KMatrix& kMatrix,
                     const HomgPoint2D& Source, HomgPoint2D &Dest);
  
    /** @brief Performs a projection of a projected and calibrated coordinate
        to a spherical image coordinate
        given a projection-type and a  K-matrix.
        @author Birger Streckel */
    int ProjectCoordBack(const int projectionType, const KMatrix& kMatrix,
                         const HomgPoint2D& Source, HomgPoint2D &Dest); 

    /** @brief Transforms a coordinate pair of image coordinates x, y in the 
        uncalibrated spherical image, to a calibrated pair of polar coordinates
        phi, theta.
        @author Birger Streckel */
    int TransfCoordImage2Sphere (const HomgPoint2D& Source,
                                 double &theta, double &phi);
    int TransfCoordImage2Sphere (const double& SourceX, const double& SourceY, 
                                 double &theta, double &phi);

    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in the uncalibrated 
        spherical image.
        @author Birger Streckel */
    int TransfCoordSphere2Image (const double& theta, const double& phi,
                                 HomgPoint2D &Source);
    int TransfCoordSphere2Image (const double& theta, const double& phi,
                                 double &SourceX, double &SourceY);
    
    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in the uncalibrated 
        spherical image.
        @author Birger Streckel */
    int TransfCoordRotateSphere2Image (double theta, double phi,
                                       const double& viewCenterX,
                                       const double& viewCenterY,
                                       HomgPoint2D &Source);
    int TransfCoordRotateSphere2Image (double theta, double phi,
                                       const double& viewCenterX, 
                                       const double& viewCenterY,
                                       double &SourceX, double &SourceY);
      
    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in an equidistant
        spherical image.
        @author Birger Streckel */
    int TransfCoordSphere2CalibImage (const double& theta, const double& phi, 
                                      HomgPoint2D &Source) const;
    int TransfCoordSphere2CalibImage (const double& theta, const double& phi,
                                      double &SourceX, double &SourceY) const;
     
   /** @brief Get the Cameras KMatrix corresponding
        @author streckel */
    KMatrix GetKMatrix();

    //int TransfRectImage2Sphere (double DestX, double DestY, int imgWidth,
		//		 double zoomFactor,double &theta, double &phi);
    /** @brief Returns the principal point of the camera image.
        @author Birger Streckel */
    int GetSphereCenter(double& centerX, double& centerY) const;
    int GetSphereCenter(double& centerX, double& centerY, double& radius)const;

    /** @brief Returns the radius in pixels of the camera image circle.
        @author Birger Streckel */
    double GetSphereRadius() const;
     
    /** @brief Returns the maximum fov-angle of the fisheye optics.
        @author Birger Streckel */
    double GetMaxCamAngle() const;
     
    /** @brief Draw circles of constant angle steps into the uncalibrated 
        spherical image using the angle calibration.
        @author Birger Streckel */
    template <class StorageType>
    int DrawCircles(Image<StorageType>& src, const double& DegreeStep);
    
    /** @brief Corrects the illumination vignette of an uncalibrated
        (source-)image.
        @author Birger Streckel */
    template <class StorageType>
    int CorrectVignette(Image<StorageType>& src);

    inline int IsInitialized() const { return isInitialized_; }

    /** @brief theta and phi are periodic in 2*M_PI, if they are out of a 
        suitable range, try to find best equivalent angle by adding 2*l*M_PI
        @return 0 on success, <0 indicates that at least one angle is invalid
        @author koeser 10/2004 */
    int EnforceNormalRange(double& theta, double& phi) const;

  
  private: 
    // Calibration measurements for angle correction
    std::vector<double> AngleCorrX_;
    std::vector<double> AngleCorrY_;

    // Calibration measurements and illumination correction
    std::vector<double> IlluCorrX_;
    std::vector<double> IlluCorrY_; 

    // Spline interpolators for the measurement interpolation
    Interpolator interpolatorAngle_;
    Interpolator interpolatorAngleBack_;
    Interpolator interpolatorIllu_;

    // Maximum image angle of the fisheye camera
    double MaxCamAngle_;

    // Center and Radius of Sphere in Image
    double CenterX_, CenterY_;
    double radius_;
    // true if center and radius is initialized
    bool isInitialized_;
    
    // helper function for sphere center and radius initialisation
    template <class StorageType>
    int CalcCentGrav_(const Image<StorageType> &src, 
                     double &centerX, double &centerY) const;

    // helper function for sphere center and radius initialisation
    template <class StorageType>
    int CalcCircularGradient_(const Image<StorageType>& src,
         double curCentX, double curCentY, double curRad, 
         double &CentGravX, double &CentGravY, double &curGradSum) const;

    // helper function for sphere center and radius initialisation
    template <class StorageType>
    int FitCircle(const Image<StorageType>& src, 
                  double &CenterX, double &CenterY, double &Radius);

  };

} // namespace 


#endif // __SphericalUndistortion_hh__
