#ifndef __VIDEOSTREAM_HH__
#define __VIDEOSTREAM_HH__


#include <string>
#include <fstream>

#include <Base/Image/ImageBase.hh>


namespace BIAS {

  /** To allow capturing of camera image to disk efficiently, (avoid
      multiple memcopy operations), this class defines an interface to
      file containing one header and multiple images. All data for
      write() must be aligned to 512-byte boundaries to allow
      unbuffered operations with open(,O_DIRECT). This can only
      guaranteed by circumventing all C++ stuff like new() and
      operator<<(). These video stream file are written by
      MIP::VideoSource_() and can be split into multiple images with
      this class.
      @author Jan-Friso Evers-Senne
  */
  /*
    File structure:
    - one header: 512 byte
    - N sets of size AlignedData_:
      - VSImageHeader (time stamp and raw uuid)
      - ImageData (size: width * height * *channelsount * bytesperpixel)
  */
  class VideoStream {
  public:
    
    // this header only occur once at the beginning of each stream file
    // size doesn't matter, but alignment matters.
    struct VSHeader {
      unsigned int Width,Height;                 // 8 byte
      unsigned char ChannelCount,BytesPerPixel;  // 4 byte 
      ImageBase::EStorageType StorageType;       // 4 byte
      ImageBase::EColorModel  ColorModel;        // 4 byte
      // size of image in byte = width *height*channelcount * bytesperpixel
      // this is the next multiple-of-512 size
      unsigned int AlignedSize;                  // 4 byte
      //                                      Sum: 24

      // Padding must fill-up this struct to _exactly_ 512 byte
      unsigned char Padding[488];
    };

    struct VSImageHeader {
      long int tsec,tusec;
      // uuid as string with like (use 40 instead of 37 char)
      // c08ded6e-8419-11d9-8e06-000d614f5bc5
      char uuid[40];
    };

  public :
    VideoStream();
    VideoStream(const std::string & file);
    ~VideoStream();
    
    int Open(const std::string & file);
    int Close();
    void PrintHeader();
    // go through video stream and export count frames starting at start 
    int SplitIntoFiles(const std::string &prefix, unsigned int start=0,
		       unsigned int count =0);
  protected:
    VSHeader Header_;
    std::ifstream Ifs_;
    unsigned int ImageSize_;
    unsigned int NumImages_;
    unsigned long int FileSize_;
    
  };

}// namespace 


#endif
