/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __Tracker_hh__
#define __Tracker_hh__

#include <vector>

#include <Base/Math/Vector.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Image/PyramidImage.hh>
#include <Utils/Param.hh>
#include <Filter/GradientSobel3x3.hh>
#include <Filter/GradientGauss.hh>
#include <Filter/GradientGaussAsymmetric.hh>
#include <Filter/Gauss.hh>
#include <Filter/Mean.hh>

#include "TrackerBase.hh"
#include "TrackerSimpleKLT.hh"
#include "TrackerGaussianWeightedKLT.hh"

#include <deque>

#define D_TRACKER_INIT       0x00000001
#define D_TRACKER_TRACK      0x00000002
#define D_TRACKER_PARAM      0x00000004
#define D_TRACKER_WRITE_IM   0x00000008
#define D_TRACKER_FILL_UP    0x00000010
#define D_TRACKER_IMAGES     0x00000020

namespace BIAS{

  template <class StorageType, class CalculationType>
  class DLLEXPORTAPI Tracker : public Debug
  {
  public:
    /// Default constructor without param object
    Tracker();

    Tracker(Param& para, const std::string& postfix="");

    Tracker(Param& para, const int grp, const std::string& postfix);

    ~Tracker();

    /** adopt to changes in parameter
        @author woelk 09/2004 */
    void ParameterChanged();

    /** 
        @author woelk 09/2004 */
    void ClearPoints();


    /** 
        @author streckel 11/2004 */
    void Reset();

    /** @brief inits data structures for subsequent calls to replaceallpoints
        @author koeser 11/2004 */
    int InitArrays();

    /** Adds all points further away than *MinBorderDist_ from roi border 
        to the list of points to be tracked.
        Track must have been called before this function.
        @author woelk 09/2004 */
    void AddPoints(const std::vector<HomgPoint2D>& points,
                   const std::vector<float>& quality);

    /** Adds first *MaxFeatures_ points further away than *MinBorderDist_ 
        from roi border to the list of points to be tracked.
        Track must have been called before this function.
        returns -1 if not neough input points were available
        @author woelk 09/2004 */
    int AddNumPoints(const std::vector<HomgPoint2D>& points,
                     const std::vector<float>& quality);

    /** fills up all points lost in tracking so far (i.e. all points where 
        Points_[k][2]=0.0 ).
        returns number of used points, negative on error.
        qual is the quality from th corner detector.
        Only points further away than *MinBorderDist_ from roi border are 
        added.
        @author woelk 09/2004 */ 
    int FillUpPoints(const std::vector<HomgPoint2D>& points,
                     const std::vector<float>& qual);


    /** Place every point from points which is not at infinity, in the internal
        data structure Points_. 
        @author woelk 09/2004 */
    int ReplacePoints(const std::vector<HomgPoint2D>& points,
                      const std::vector<float>& qual);

    /** Place every point from points even those at infinity, in the internal
        data structure Points_. 
        @author woelk 09/2004 */
    int ReplaceAllPoints(const std::vector<HomgPoint2D>& points,
                         const std::vector<float>& qual);

    /** Prepares the pyramid images:
        Filters im to pim[0], downsamples pim[0] and
        calculates the gradients gx and gy of pim in every pyramid level. 
        @author woelk 09/2004 */
    int PreparePyramide(const Image<StorageType>& im, 
                        PyramidImage<CalculationType>& pim,
                        PyramidImage<CalculationType>& gx, 
                        PyramidImage<CalculationType>& gy);

    /** tracks all points in Points_ 
        @author woelk 09/2004 */
    int Track(PyramidImage<CalculationType>& pim, 
              PyramidImage<CalculationType>& gradx,
              PyramidImage<CalculationType>& grady);

    /** tracks all points in Points_ 
        using prediction as previous knowledge
        @author woelk 09/2004 */
    int Track(PyramidImage<CalculationType>& pim, 
              PyramidImage<CalculationType>& gradx,
              PyramidImage<CalculationType>& grady,
              const std::vector<HomgPoint2D>& prediction);

    /** returns the point positions in the actual image
        points at infinity (w=0.0) are returned, if tracking failed
        @author woelk */
    inline void GetPoints(std::vector<HomgPoint2D>& p) const 
    { p = Points_[*PyIndex_]; }

    /** @brief returns the number of iterations used in the biggest 
        pyramid level
        @author woelk 09/2004 */
    inline void GetNumIter(std::vector<int>& numiter) const 
    { numiter = NumIter_[*PyIndex_]; }

    /** @brief returns the displacement in the last iteration on the biggest 
        pyramid level.
        @author woelk 09/2004 */
    inline void GetError(std::vector<double>& error) const
    { error = DisplInLastIter_[*PyIndex_]; }

    /** @brief returns the mean grey level difference in the tracking window
        on the biggest pyramid level.
        @author woelk 09/2004 */
    inline void GetResidui(std::vector<double>& residui) const
    { residui = Residui_[*PyIndex_]; }

    /** @brief returns the results from the base tracker class on the 
        biggest pyramid level.
        0 on success (error< maxerror) - the point was tracked
        -1 when no spatial structure is present
        -2 when point slid out of image
        -3 when maxiter is reached an error is above maxerror
        -4 when a point lies outside of valid region in images
        -5 when the mean absolute difference between the two poinst 
        -6 when input point is at infinity
        1 when error increased from last iteration
    @author woelk 09/2004 */
    inline void GetResults(std::vector<int>& results) const
    { results = res_[*PyIndex_]; }

    /** @brief write a track file in Birchfeld-style.

        The file is independent of internal data structures.
        p[i][k] is point number k in image i and 
        results[i][k] is result for point number k in image i.
        
        @author woelk 09/2004 */
    int WriteTrackFile(std::deque<vector<HomgPoint2D> >& p, 
                       std::deque<vector<int> >& results, 
                       const string& name);

    /////////////////////////////////////////////////////////////////////////
    
    /// returns the desrired num features t
    inline int GetNumFeatures()
    { return *MaxFeatures_; }
 
    /// returns the minimu distance to image border which is used in 
    /// AddPoints and FillUpPoints
    inline int GetMinBorderDistance()
    { return *MinBorderDist_; }
 

#ifdef BIAS_DEBUG
    void DumpResults(std::ostream& os=cout) {
      const int pysize=(int)LastImage_->size()-1;
      const int maxf=*MaxFeatures_;
      const int pyind=*PyIndex_;
      for (int i=0; i<maxf; i++){
        os << setw(5) << i << " : ";
        for (int p=pysize; p>=pyind; p--){
          os << setw(4) << res_[p][i];
        }
        os << endl;
      }
    }
#endif

  protected:
    /// the different trackers
    TrackerBase<CalculationType> *tb_;
    TrackerSimpleKLT<CalculationType> ts_;
    TrackerGaussianWeightedKLT<CalculationType> tgw_;

    /// the different gradient classes
    FilterNTo2N<CalculationType, CalculationType> *grad_;
    GradientSobel3x3<CalculationType, CalculationType> gsobel3x3_;
    GradientGauss<CalculationType, CalculationType> ggauss_;
    GradientGaussAsymmetric<CalculationType,CalculationType>ggaussasymmetric_;

    /// the different mean filter classes
    Convolution<StorageType, CalculationType> *lowpass_;
    Gauss<StorageType, CalculationType> gauss_;
    Binomial<StorageType, CalculationType> binomial_;
    Mean<StorageType, CalculationType> mean_;
    bool _ParamsAllocated;
    /// the parameter 
    Vector<int> *MaxIter_;
    Vector<KLT_TYPE> *MaxError_;
    Vector<int> *HalfWinSize_;
    int *PyIndex_;
    int *TrackerType_;
    int *GradientType_;
    int *MeanType_;
    /// minimal distance of every point from image border in biggest pyramid im
    int *MinBorderDist_; 
    /// maximum number of features which will be tracked
    int *MaxFeatures_;
    /// parameter from TrackerBase
    int *_HalfWinSize, *_MaxIterations;
    KLT_TYPE *_MaxError; 
    /// parameter from mean filters
    double *_GaussSigma, *_GaussRatio;
    int *_MeanSize;

    std::vector<HomgPoint2D> p2tracked_;
    /// result code for every point 
    std::vector<std::vector<int> > res_;
    /// the prediction for every pyramid image
    std::vector<std::vector<HomgPoint2D> > Points2_;
    /// the points
    std::vector<std::vector<HomgPoint2D> > Points_;
    /// number of iterations used
    std::vector<std::vector<int> > NumIter_;
    /// displacement inlast iteration
    std::vector<std::vector<double> > DisplInLastIter_;
    /// mean grey level difference in tracking window 
    std::vector<std::vector<double> > Residui_;

    // the pyramid images from last call to Track
    PyramidImage<CalculationType> *LastImage_;
    PyramidImage<CalculationType> *LastGradX_, *LastGradY_; 

    void AddParameter_(Param& para, const int grp, const std::string& postfix);

  };

}

#endif // __Tracker_hh__
