/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TrackerBase.hh"

using namespace BIAS;
using namespace std;


//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////

template class TrackerBase<unsigned char>; 
template class TrackerBase<float>; 

// fill in instances as required
#ifdef BUILD_IMAGE_INT
template class TrackerBase<int>; 
#endif
#ifdef BUILD_IMAGE_CHAR
template class TrackerBase<char>; 
#endif
#ifdef BUILD_IMAGE_SHORT
template class TrackerBase<short>; 
#endif
#ifdef BUILD_IMAGE_USHORT
template class TrackerBase<unsigned short>; 
#endif
#ifdef BUILD_IMAGE_UINT
template class TrackerBase<unsigned int>; 
#endif
#ifdef BUILD_IMAGE_DOUBLE
#endif

//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////


template <class StorageType>
TrackerBase<StorageType>::
TrackerBase() 
  : Debug()
{
  _HalfWinSize=_MaxIterations=NULL; 
  _MaxError=NULL;
  _maxx=_maxy=_minx=_miny=0;
  _ParamsAllocated=true;

  _HalfWinSize=new int;
  *_HalfWinSize=3;
  _MaxIterations=new int;
  *_MaxIterations=10;
  _MaxError=new double;
  *_MaxError=0.01;
  _MaxResiduum=new double;
  *_MaxResiduum=10.0;
}

template <class StorageType>
TrackerBase<StorageType>::
TrackerBase(Param& para, const std::string& postfix) 
  : Debug()
{
  _HalfWinSize=_MaxIterations=NULL; 
  _MaxError=NULL;
  _maxx=_maxy=_minx=_miny=0;
  int grp=para.GetFreeGroupID();
  _AddParam(para, grp, "-TrackerBase"+postfix);
  if (para.IsUsedGroupID(grp)) para.SetGroupName(grp, "TrackerBase");
  _ParamsAllocated=false;
}

template <class StorageType>
TrackerBase<StorageType>::
TrackerBase(Param& para, const int grp, const std::string& postfix) 
  : Debug()
{
  _HalfWinSize=_MaxIterations=NULL; 
  _MaxError=NULL;
  _maxx=_maxy=_minx=_miny=0;
  _AddParam(para, grp, "-TrackerBase"+postfix);
  _ParamsAllocated=false;
}

template <class StorageType>
TrackerBase<StorageType>::
~TrackerBase()
{
  if (_ParamsAllocated){
    delete _HalfWinSize;
    delete _MaxIterations;
    delete _MaxError;
    delete _MaxResiduum;
  }
}

template <class StorageType>
void TrackerBase<StorageType>::
Init(StorageType **ida1, StorageType **ida2, StorageType **gradx1, 
     StorageType **grady1, StorageType **gradx2, StorageType **grady2, 
     unsigned int width, unsigned int height, unsigned gradient_half_mask)
{
  _Init(ida1, ida2, gradx1, grady1, gradx2, grady2, width, height, 
        gradient_half_mask);
}

template <class StorageType>
void TrackerBase<StorageType>::
_Init(StorageType **ida1, StorageType **ida2, StorageType **gradx1, 
      StorageType **grady1, StorageType **gradx2, StorageType **grady2, 
      unsigned int width, unsigned int height, unsigned gradient_half_mask)
{
  _maxx=width-*_HalfWinSize-gradient_half_mask-1;
  _maxy=height-*_HalfWinSize-gradient_half_mask-1;
  _minx=_miny=gradient_half_mask+*_HalfWinSize+1;
  BIASCDOUT(D_TRACKERB_INIT, "tracking roi: ( "<<_minx<<", "<<_miny
            <<")  <-->  ( "<<_maxx<<", "<<_maxy<<")\n");

  _ida1=ida1;
  _ida2=ida2;
  _gradx1=gradx1;
  _gradx2=gradx2;
  _grady1=grady1;
  _grady2=grady2;
  BIASASSERT(*_HalfWinSize>0);
  BIASASSERT(_maxy>_miny && _maxx>_minx);
}

template <class StorageType>
void TrackerBase<StorageType>::
Init(Image<StorageType>& im1, Image<StorageType>& im2, 
     Image<StorageType>& gradx1, Image<StorageType>& grady1, 
     Image<StorageType>& gradx2, Image<StorageType>& grady2)
{
  unsigned width=im1.GetWidth(), height=im1.GetHeight();
  unsigned tlx, tly, brx, bry;
  gradx1.GetROI(tlx, tly, brx, bry);
  unsigned itlx, itly, ibrx, ibry;
  im1.GetROI(itlx, itly, ibrx, ibry);
  BIASASSERT(tlx>itlx);
  unsigned gradient_half_mask=tlx-itlx;
  BIASCDOUT(D_TRACKERB_INIT, "gradient_hws= "<<gradient_half_mask<<endl);
  BIASASSERT(im2.SamePixelAndChannelCount(im1));
  BIASASSERT(gradx1.SamePixelAndChannelCount(im1));
  BIASASSERT(grady1.SamePixelAndChannelCount(im1));
  BIASASSERT(gradx2.SamePixelAndChannelCount(im1));
  BIASASSERT(grady2.SamePixelAndChannelCount(im1));
  _Init(im1.GetImageDataArray(), im2.GetImageDataArray(),
        gradx1.GetImageDataArray(), grady1.GetImageDataArray(),
        gradx2.GetImageDataArray(), grady2.GetImageDataArray(),
        width, height, gradient_half_mask);
}


template <class StorageType>
int TrackerBase<StorageType>::
Track(HomgPoint2D& p1, HomgPoint2D& p2, HomgPoint2D& result, 
      KLT_TYPE& error, int &iter, KLT_TYPE& residuum)
{
  BIASERR("call Track only from derived classes");
  return -10;
}


template <class StorageType>
void TrackerBase<StorageType>::
Track(std::vector<HomgPoint2D>& p1, std::vector<HomgPoint2D>& p2tracked, 
      std::vector<HomgPoint2D>& result, std::vector<KLT_TYPE>& error, 
      std::vector<int>& numiter, std::vector<KLT_TYPE>& residuum, 
      std::vector<int>& res)
{
  BIASERR("call Track only from derived classes");
  BIASASSERT(false);
}


template <class StorageType>
int TrackerBase<StorageType>::
Track(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], KLT_TYPE& error, 
      int &iter, KLT_TYPE& residuum)
{
  BIASERR("call Track only from derived classes");
  return -10;
}

template <class StorageType>
int* TrackerBase<StorageType>::
AddTrackerType(Param& para, const std::string& name, const int grp)
{
  int *res;
    vector<string> tt;
  vector<int> tti;
  tt.push_back("TT_SimpleKLT");tti.push_back(TT_SimpleKLT);
  tt.push_back("TT_GaussianWeightedKLT");tti.push_back(TT_GaussianWeightedKLT);
  //tt.push_back(""); tti.push_back();
  //tt.push_back(""); tti.push_back();
  if (para.CheckParam(name)){ 
    res=para.GetParamEnum(name);
  } else { 
    res=para.AddParamEnum(name, "type of tracker to use", tt, TT_SimpleKLT,
                          &tti, 0, grp); 
  }
  return res;
}

template <class StorageType>
void TrackerBase<StorageType>::
_AddParam(Param& para, const int grp, const std::string& postfix)
{
  if (para.CheckParam("HalfWinSize"+postfix)){ 
    _HalfWinSize=para.GetParamInt("HalfWinSize"+postfix);
  } else { 
    _HalfWinSize=
      para.AddParamInt("HalfWinSize"+postfix, 
                       "half size of support window used in gradient"
                       " descent tracking", 
                       3, 1, INT_MAX, 0, grp);
  }

  if (para.CheckParam("MaxIterations"+postfix)){ 
    _MaxIterations=para.GetParamInt("MaxIterations"+postfix);
  } else { 
    _MaxIterations=
      para.AddParamInt("MaxIterations"+postfix, 
                       "maximum number of iterations", 
                       10, 1, INT_MAX, 0, grp);
  }

  if (para.CheckParam("MaxError"+postfix)){ 
    _MaxError=para.GetParamDouble("MaxError"+postfix);
  } else { 
    _MaxError=
      para.AddParamDouble("MaxError"+postfix, 
                          "iterations are aborted when displacement"
                          " drops below MaxError", 
                          0.01, 0.0, DBL_MAX, 0, grp);
  }

  if (para.CheckParam("MaxResiduum"+postfix)){ 
    _MaxResiduum=para.GetParamDouble("MaxResiduum"+postfix);
  } else { 
    _MaxResiduum=
      para.AddParamDouble("MaxResiduum"+postfix, 
                          "tracking failed when mean image difference is above"
                          " MaxResiduum after otherwise successfull tracking", 
                          10.0, 0.0, DBL_MAX, 0, grp);
  }

}
