/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __TrackerBase__hh__
#define __TrackerBase__hh__

#include <Base/Debug/Debug.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Image/Image.hh>
#include <Utils/Param.hh>

#include <vector>

#define KLT_TYPE double

#define D_TRACKERB_INIT        0x00000001
#define D_TRACKERB_KLT         0x00000002
#define D_TRACKERB_BILINEAR    0x00000004
#define D_TRACKERB_INPUT       0x00000008

#define TRACKER_SUCCESS      0 
#define TRACKER_NOSTRUCTURE -1 
#define TRACKER_OUTOFIMAGE  -2
#define TRACKER_TOOMANYITER -3
#define TRACKER_OUTOFROI    -4 
#define TRACKER_MEANABSDIF  -5
#define TRACKER_IGNOREPOINT -6 
#define TRACKER_ERRORINCREASED  1



namespace BIAS {

    enum ETrackerType {TT_Invalid, TT_SimpleKLT, TT_GaussianWeightedKLT};

  /** @class TrackerBase
      @ingroup g_geometry
      @brief defines the interface for the different tracking algorithms 

      Base class for the different tracking algorithms, defining the interfaces
      for the tracking functions. Also stores some basic options and datas, 
      i.e. the pointer to the image datas.

      @author woelk 09/2004 */
  template <class StorageType>
  class DLLEXPORTAPI TrackerBase : public Debug
  {
  public:
  public :
    TrackerBase();

    TrackerBase(Param& para, const std::string& postfix="");

    TrackerBase(Param& para, const int grp, const std::string& postfix);

    virtual ~TrackerBase();

    /** Track from im1 to im2, 
        gradx1/2: horizontal gradient imgs
        grady1/2: vertical gradient imgs
        halfwinsize: the support window size that should be used in the
                     algorithms
        The ROI must be set correctly in gradient images (representing the 
        gradient filter mask size).      
        The algorithms either terminates if the displacement resulting from
        one iteration drops below maxerror or the maximum number of iterations
        maxiterations is reached or if the displacement increased with 
        respect to the last iteration */
    virtual void Init(Image<StorageType>& im1, 
                      Image<StorageType>& im2, 
                      Image<StorageType>& gradx1, 
                      Image<StorageType>& grady1, 
                      Image<StorageType>& gradx2, 
                      Image<StorageType>& grady2);
    
    /** all homogenous points must be homogenized
        @return   see defines
                  0 on success (error< maxerror)
                 -1 when no spatial structure is present
                 -2 when point slid out of image
                 -3 when maxiter is reached and error is above maxerror
                 -4 when a point lies outside of valid region in images
                 -5 when the mean absolute difference between the two poinst 
                  1 when error increased from last iteration */
    virtual int Track(HomgPoint2D& p1, HomgPoint2D& p2, HomgPoint2D& result, 
                      KLT_TYPE& error, int &iter, KLT_TYPE& residuum);

    /** just the function prototype */
    virtual void Track(std::vector<HomgPoint2D>& p1, 
                       std::vector<HomgPoint2D>& p2, 
                       std::vector<HomgPoint2D>& p2tracked, 
                       std::vector<KLT_TYPE>& error, 
                       std::vector<int>& numiter,
                       std::vector<KLT_TYPE>& residuum, 
                       std::vector<int>& res);

    /////////////////////////////////////////////////////////////////////////
    //    basic functions, needed, when images are no longer present
    /////////////////////////////////////////////////////////////////////////

    /** most basic function, probably later needed by Scan2DTracking */
    virtual void Init(StorageType **ida1, StorageType **ida2, 
                      StorageType **gradx1, StorageType **grady1, 
                      StorageType **gradx2, StorageType **grady2, 
                      unsigned int width, unsigned int height,
                      unsigned gradient_half_mask);

    /** most basic function, probably later needed by Scan2DTracking */
    virtual int Track(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
                      KLT_TYPE& error, int &iter, KLT_TYPE& residuum);

    static 
    int *AddTrackerType(Param& para, const std::string& name="TrackerType", 
                        const int group_id=GRP_ALL);
    

    ////////////////////////////////////////////////////////////////////////

    inline void SetMaxError(const KLT_TYPE maxerr) { *_MaxError=maxerr; }

    inline void SetMaxResiduum(const KLT_TYPE maxres) { *_MaxResiduum=maxres; }

    inline void SetMaxIterations(const int maxiter){ *_MaxIterations=maxiter; }
    
    inline void SetHalfWinSize(const int hws) { *_HalfWinSize=hws; }
    

    inline int *GetMaxIterationsPointer() {return _MaxIterations;};
    inline int *GetHalfWinSizePointer() {return _HalfWinSize;};
    inline KLT_TYPE *GetMaxErrorPointer() {return _MaxError;};

    /////////////////////////////////////////////////////////////////////////
    //        TODO

    /** @brief Bakers inverse compositional approach 
        int BakerKLT(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
        KLT_TYPE& error, unsigned &iter); */

    /** guesses an affine transformation for each patch 
        | 1+p1 p3   p5 |
        | p2   1+p4 p6 |
        where d=(p5, p6)^T is the translation and p1-p4 stand for the affine 
        part 
    int AffineKLT(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
                  KLT_TYPE& error, unsigned &iter);
    */
    
    /** guesses an affine transformation and and compensation for illumination
        changes 
    int AffineIllumCompKLT(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2],
                           KLT_TYPE& error,unsigned &iter);
    */
    
    /** @brief Gradient decent using Fusiellos X84 outlier rejection rule 
    int X84KLT(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2],
               KLT_TYPE& error,unsigned &iter);
    */

  protected:
    bool _ParamsAllocated;
    // the image data arrays
    StorageType **_ida1, **_ida2;
    // gradients are alwys signed (or float) can not use StorageType
    StorageType **_gradx1, **_grady1, **_gradx2, **_grady2;
    // the image borders modulo filter and support windows
    int _maxx, _maxy, _minx, _miny;
    // iteration stops after _MaxIterations
    int *_HalfWinSize, *_MaxIterations;
    // iteration stops if displacement refinement is less than maxerror_
    KLT_TYPE *_MaxError, *_MaxResiduum; 

    void _Init(StorageType **ida1, StorageType **ida2, 
               StorageType **gradx1, StorageType **grady1, 
               StorageType **gradx2, StorageType **grady2, 
               unsigned int width, unsigned int height,
               unsigned gradient_half_mask);

    void _AddParam(Param& para, const int grp, const std::string& postfix);

  }; // class


} // namespace


#endif // __TrackerBase__hh__
