/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "IterativExtendedKalman3DPoint.hh"

#include "Base/Math/Vector2.hh"
#include "Base/Math/Vector3.hh"
#include "Base/Math/Matrix2x2.hh"
#include "Base/Math/Matrix.hh"

using namespace BIAS;
using namespace std;

void 
BIAS::IterativExtendedKalman3DPoint(const HomgPoint2D& x, 
                                    const PMatrixBase& P, 
                                    HomgPoint3D& X, Matrix3x3<double>& L,
                                    double MeasurementNoise, 
                                    unsigned int NumIterations, 
                                    double DiffusionFactor)
{
  Vector2<double> xp;    // xp contains the observed image point
  xp[0]=x[0]/x[2];       // the new measurement
  xp[1]=x[1]/x[2];

  Matrix2x2<double> R;     // contains the covariance matrix of the noise
  R.Fill(0.0);             // on the observed image points
  R[0][0]=R[1][1]=MeasurementNoise*MeasurementNoise;

  Vector3<double> XP;    // XP contains the state (3D point) of
  XP[0]=X[0]/X[3];       // the system
  XP[1]=X[1]/X[3];
  XP[2]=X[2]/X[3];
 
  L*=DiffusionFactor; // diffusion of system state

  double N;  // the homg. coo w of the 2d point calculated from P and X
  // Delta_h is the Jacabian Matrix of "normalized P" will say P/N
  Matrix<double> Delta_h(2,3); // Jacobian of measurement matrix
  Matrix<double> W(3,2); // Kalman gain/weight
  Matrix2x2<double> S; // denominator of Kalman gain update equation
  Vector2<double> H_X, Nu; 

  for (unsigned int n = 1; n <= NumIterations; n++){ 
    N = P[2][0] * XP[0]+ P[2][1] * XP[1]+ P[2][2] * XP[2] + P[2][3];
    for (int i = 0; i < 2; i++)
      for (int j = 0; j < 3; j++)
        Delta_h[i][j] = P[i][j]/N - 
          P[2][j]*(P[i][0]*XP[0]+P[i][1]*XP[1]+P[i][2]*XP[2]+P[i][3])/(N*N);
      
    // demoninator of Kalman weight update equation
    S = Delta_h * L * Delta_h.Transpose() + R;
 
    H_X[0] = (P[0][0]*XP[0] + P[0][1]*XP[1] + P[0][2]*XP[2] + P[0][3])/N;
    H_X[1] = (P[1][0]*XP[0] + P[1][1]*XP[1] + P[1][2]*XP[2] + P[1][3])/N;

    // Update equations
  
    // Kalman weight update equation
    W = L * Delta_h.Transpose() * S.Invert(); 

    Nu = xp - H_X; // messfehler
    XP = XP + (Vector3<double>)(W*Nu);

    L = L - W*S*W.Transpose();
  }

  X.Set(XP[0],XP[1],XP[2]);
} 
