/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Image/Camera.hh>
#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <Geometry/PMatrix.hh>
// #include <BIAS/Filter.hh>
#include <stdio.h>

#ifdef   WIN32
#  include "Base/Common/getopt_W32.h"
#else  //not WIN32
#  include <getopt.h>
#endif //WIN32

using namespace BIAS;
using namespace std;

void usage()
{
  cout << "Usage: biasconvert [options] infile outfile"<<endl;
  cout <<" Options: "<<endl
       <<" -g         : convert to grey values"<<endl
//        <<" -d <int>   : downsample by power of 2"<<endl
       <<" -h         : show this help"<<endl
       <<" -p <file>  : add P-matrix"<<endl
       <<" -v         : be verbose"<<endl
       <<" -m <float> : modulate (scale) all pixel values with <float> (def:1)"<<endl
       <<" -o <float> : offset (shift) all pixel values with <float> (def: 0)"<<endl
       <<" if neither -m nor -o is given float images are scaled and shifted automatically"<<endl
//        <<" while rescaleing, P-matrix are adopted."<<endl
       <<endl;
}




int DetermineFiletype(const string &fn)
{
  int EndOfPrefix;
  string suffix;
  
  EndOfPrefix = fn.rfind('.');
  
  suffix = fn.substr(EndOfPrefix+1,fn.length());
  if ((suffix=="jpg") || (suffix=="jpeg")) return ImageIO::FF_jpg; //JW 01/2003
  if (suffix == "pnm") return ImageIO::FF_ppm;
  if (suffix == "ppm") return ImageIO::FF_ppm;
  if (suffix == "pgm") return ImageIO::FF_pgm;
  if (suffix == "mip") return ImageIO::FF_mip;
  if (suffix == "png") return ImageIO::FF_png;
  if (suffix == "tif") return ImageIO::FM_tif;
  if ((suffix=="viff") || (suffix=="vif") || (suffix=="vff")) 
    return ImageIO::FF_viff; // JW 01/2003
  
  cout <<"Unknown filename extension: "<<suffix<<endl;
  return -1;
}



void AdoptPMatrix(PMatrix &P,float scalex,float scaley)
{
  PMatrix *newP=NULL;
  Vector3<double> C;
  KMatrix K;
  Matrix3x3<double> R;

  if (P[0][0] != 0) {
    // JW 01/2003
    //    C = P.GetC();
    //    K = P.GetK();
    //    R = P.GetR();
    P.GetC(C);
    P.GetK(K);
    P.GetR(R);


    K[0][0] *= scalex;
    K[1][1] *= scaley;
    K[0][2] *= scalex;
    K[1][2] *= scaley;
    // the skew
    K[0][1] *= scaley;
    newP = new PMatrix(K,R,C);
    P = *newP;
  }
}



int ReadPMatrix(PMatrixBase &pmat,const string &filename)
{
  int row,col;
  ifstream pmatfile(filename.c_str());
  if (!pmatfile){
    cerr << "unable to open " << filename<< endl; 
    perror("");
    return -1;
  }
  pmatfile >> row >> col;
  if (!pmatfile){
    cerr << "error reading " << filename << endl;
    return -1;
  }
  if (row!=3 || col!=4){
    cerr << "P Matrix should have size 3x4, not: " <<row<<"x"<<col << endl;
    return -1;
  }


  for (int r=0; r < row; r++)
    for (int c=0; c< col; c++)
      pmatfile >> pmat[r][c];
  return 0;
  
}



/** @class biasconvert
    @ingroup g_tools
    @brief ???
    @author evers */
int main (int argc, char *argv[])
{
  //    pic.SetDebugLevel(D_CAM_IO);
  bool togrey = false;
  bool verbose=false;
  string pmatfile="";
  string ifile,ofile;
  //long int sec,usec;
  //int downbypoweroftwo =0;
  //float scalex=1.0;
  //float scaley=1.0;
  float scale = 1.0;
  float shift = 0.0;
  bool scaleshift = false;
  bool clip = false;
  
  


  ImageBase ipic;
  Camera<unsigned char> ucocam;
  Camera<float> focam;
  
  ImageBase::EStorageType OutST=ImageBase::ST_invalid;
  
  ImageIO::TFileFormat ofiletype;
  
  Camera<unsigned char> opic;
  PMatrix P_;
  

  if (argc <3) {
    usage();
    //exit(1);
	return 0; // for Dart testing
  }

  // scan options
  char c=0;
  while (c>=0) {
    c = getopt(argc,argv,"ghd:p:vm:o:c");
    switch (c) {
    case 'h':
    case ':':
    case '?':
      usage();
      exit(1);
      break;
    case 'g':
      togrey = true;
      break;
      //     case 'd':
      //       downbypoweroftwo = atoi(optarg);
      //       scalex = 1.0/(float)(1<< downbypoweroftwo);
      //       scaley = 1.0/(float)(1<< downbypoweroftwo);
      //      break;
    case 'p':
      pmatfile=optarg;
      break;
    case 'v':
      verbose = true;
      break;
    case 'c':
      clip = true;
      break;
    case 'o':
      shift = float(atof(optarg));
      scaleshift = true;
      break;
    case 'm':
      scale = float(atof(optarg));
      scaleshift = true;
      break;
    }
  }

  // after getopt() all recognized options are before non-options
  ifile = argv[argc-2];
  ofile = argv[argc-1];
  //   cout <<"mipconvert from "<<ifile<<" to "<<ofile<<endl;
  
  ofiletype = (ImageIO::TFileFormat)DetermineFiletype(ofile);
  if (ofiletype ==-1) exit(1);
  
  if (ofiletype == ImageIO::FF_pgm) togrey = true;
  
  if (verbose) {
    cout <<"load "<<ifile<<", "<<flush;
  }
  
  if (ImageIO::Load(ifile,ipic)<0)
    exit(1);

  if (ipic.GetStorageType() == ImageBase::ST_float) {
    if (verbose) cout <<"storagetype is float"<<endl;
    Camera<float> fcam(ipic);
    fcam.ParseMetaData();
    P_ = fcam.GetP();
    // all meta datas are preserved now (felix)
    // fcam.GetTime(sec,usec);
    
    if (ofiletype != ImageIO::FF_mip) {
      if (scaleshift){ 
       
        if (verbose) {
          float min,max;
          fcam.GetMinMaxPixelValue(min,max);
          cout << "min = " << min << ", max = " << max << endl;
        }
        
        float mean =0.0; 
        float *fcamptr=fcam.GetImageData();
        float num_elem=0;
        //float max=(255-shift)/scale;
        //float min=-shift/scale;
        for (unsigned int i = 0; i < fcam.GetWidth()*fcam.GetHeight(); i++){
          if (*fcamptr != -1.0){
            mean += *fcamptr;
            num_elem++;
          }
          fcamptr++;
        }

        mean /= num_elem;
        fcamptr=fcam.GetImageData();
        float variance=0.0;
        float scaledmean=0.0;
        for (unsigned int i = 0; i < fcam.GetWidth()*fcam.GetHeight();i++){
          if (*fcamptr != -1.0)
            variance += (*fcamptr-mean)*(*fcamptr-mean);
        }           
        
        fcam.ScaleShift(scale,shift);
        fcamptr=fcam.GetImageData();
        if (clip)           
          for (unsigned int i = 0; i < fcam.GetWidth()*fcam.GetHeight();i++){
            *fcamptr = (*fcamptr < 255) ? *fcamptr : 255 ;
            *fcamptr = (*fcamptr > 0) ? *fcamptr : 0 ;
            scaledmean += *fcamptr;
            fcamptr++;
          }
        
        if (verbose){
          variance/= num_elem;
          cout << "mean = " << mean  << ", scaledmean=" 
               << scaledmean/num_elem
               << ", variance = " 
               << variance <<  ", fill rate = " 
               <<  num_elem/fcam.GetWidth()/fcam.GetHeight() << endl;
        } 
      } else {
        if (verbose) {
          float min,max;
          fcam.GetMinMaxPixelValue(min,max);
          cout << "min = " << min << ", max = " << max << endl;
        }
        fcam.ScaleShiftBetween(0,255);
      }
      //ucocam = fcam;
      //OutST = ImageBase::ST_float;
      OutST = ImageBase::ST_unsignedchar;
      ImageConvert::ConvertST(fcam, ucocam, OutST );
      // copy meta data to ucocam
      *ucocam.GetMetaData()=*fcam.GetMetaData();
      ucocam.ParseMetaData();
    } else {
      focam = fcam;
      OutST = ImageBase::ST_float; 
    }
  } else {
    if (ipic.GetStorageType() == ImageBase::ST_unsignedchar)  {
#ifdef BIAS_DEBUG
      if (verbose) {
        if (Debug::GetGlobalDebugLevel()<D_CONVERT)
          Debug::SetGlobalDebugLevel(D_CONVERT);
      }
#endif
      if (verbose) cout <<"storagetype is unsigned char"<<endl;
      Camera<unsigned char> uccam(ipic);
      OutST = ImageBase::ST_unsignedchar;       
      uccam.ParseMetaData();
      P_ = uccam.GetP();
      // uccam.GetTime(sec,usec);
      // all meta datas are preserved now (felix)
      if (verbose) cout <<"P: "<<P_<<endl;
           
      // meta datas are lost here
      if (togrey) ImageConvert::ToGrey(uccam, ucocam);
      else ImageConvert::ToRGB(uccam, ucocam);
      
      // copy meta data to ucocam
      *ucocam.GetMetaData()=*uccam.GetMetaData();
      ucocam.ParseMetaData();
    }
  }

  
  if (pmatfile != "") {
    if (verbose) cout <<"read "<<pmatfile<<", ";
    if (ReadPMatrix(P_,pmatfile) < 0)
      exit(1);
    ucocam.SetP(P_);
  }
  
  //   if (downbypoweroftwo >0) {
  //     if (verbose) cout <<"rescale, "<<flush;
  //     Filter f;
  //     Camera<unsigned char> tmppic;
  //     f.DownsampleByPowerOfTwo(opic,tmppic,downbypoweroftwo);
  //     opic = tmppic;
  //     AdoptPMatrix(P_, scalex,scaley);
  //   }

  
  if (OutST == ImageBase::ST_unsignedchar) {
    
    // all meta datas are preserved now (felix)
    //ucocam.SetTime(sec,usec);
    
    ucocam.UpdateMetaData();

    if (verbose) cout <<"write unsigned char "<<ofile<<", "<<flush;
    if (ofiletype == ImageIO::FF_mip) {
      
      ImageIO::Write(ofile, ucocam);
    } else {
      //     if (ofiletype !=ImageIO::FF_ppm && ofiletype !=ImageIO::FF_pgm) {
      int endofprefix = ofile.rfind('.');
      ofile = ofile.substr(0,endofprefix);
      //       }
      ImageIO::ExportImage(ofile,ucocam,ofiletype);
    }
  }
  else  
    if (OutST == ImageBase::ST_float) {
      // all meta datas are preserved now (felix)
      //focam.SetP(P_);
      //focam.SetTime(sec,usec);
      focam.UpdateMetaData();
      ImageIO::Write(ofile, focam);
    }
      
    
  if (verbose) cout <<"done"<<endl;
    
  
  return 0;
    
}












