/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <stdio.h>
#include <map>
#include <string>
#include "Geometry/PMatrix.hh"
#include "Image/Camera.hh"

using namespace BIAS;
using namespace std;

void usage()
{
  cout << "Usage: biasmerge [options] files"<<endl;
  cout <<" Options: "<< endl
       <<" -s         : search next smaller(string compare) PMatrix file"<<endl
       <<" -b         : assume bbc camera file format"<<endl
       << endl;
  cout <<"This program opens the image files given by the command line"
       <<", e.g. image037.pgm, looks at the command line for a file "
       <<" image037.pgm.mat and attaches the pmatrix to the pgm header."<<endl
       <<"Returns: 0 on success, 1 if errors occured."<<endl<<endl;
}

int DetermineFiletype(const string &fn, ImageIO::TFileFormat &thetype)
{
  int EndOfPrefix;
  string suffix;
  
  EndOfPrefix = fn.rfind('.');
  
  suffix = fn.substr(EndOfPrefix+1,fn.length());
  if (suffix == "pnm") thetype =  ImageIO::FF_ppm;
  else if (suffix == "ppm") thetype =  ImageIO::FF_ppm;
  else if (suffix == "pgm") thetype =  ImageIO::FF_pgm;
  else if (suffix == "mip") thetype =  ImageIO::FF_mip;
  else return -1;

  return 0;
}


/** @class biasmerge
    @ingroup g_tools
    @brief merge a projection matrix and an image, write header data to image
    @author koeser */
int main (int argc, char *argv[])
{
  bool Nearest = false, BBC = false;
  int i = 1;
  bool param = true;
  while ((i<argc) && (param)) {
    param = false;
    if (!strcmp("-s",argv[i])) {
      std::cout << "Using nearest P from command line also."<<endl;
      Nearest = true;
      i++;
      param = true;
    } 
    if (!strcmp("-b",argv[i])) {
      std::cout << "Trying to load BBC camera data."<<endl;
      BBC = true;
      i++;
      param = true;
    }
  }
  
  if (i>=argc) { 
    usage();
    return -1;
  }
  
  ImageIO::TFileFormat FileType;
  bool ErrorFlag = false;
  int IOResult = 0;
  BIAS::ImageBase CurImage;

  PMatrix P;
  map<string, PMatrix> mapP;
  for (int j=i; j<argc; j++) {
    if (strstr(argv[j], ".mat")!=NULL) {
      cout<<"arg "<<j<<" is .mat";
      PMatrix P;
      P.SetZero();
      if (!BBC) {
        P.Load(argv[j]);
      } else {
        P.LoadBBC(argv[j]);
      }
      cout << "P is "<<P<<endl;
      mapP[argv[j]] = P;
    }
  }
  if (mapP.size()==0) {
    BIASERR("No matrices found !");
    return -1;
  }




  for (;i<argc;i++) {
    if (strstr(argv[i], ".mat")!=NULL) continue;
    cout<< "Trying "<<argv[i]<<endl;
    if (DetermineFiletype(argv[i], FileType)!=0) {
      BIASERR("Skipping current file because of unknown/unsuported format: "
              <<argv[i]<<endl<<"PMatrix can be written to mip,pgm,ppm,pnm");
      ErrorFlag = true;
      continue;
    }

    IOResult = ImageIO::Load(argv[i], CurImage);
    if (IOResult!=0) {
      ErrorFlag = true;
      BIASERR("Error reading "<< argv[i]);
      continue;
    }

    stringstream ss;
    ss << argv[i] <<".mat";
    if (mapP.find(ss.str())!=mapP.end()) {
      cout<<"Found a mat file for "<<argv[i]<<endl;
      P = mapP[ss.str()];
    } else {
      if (!Nearest) {
        cout <<"Found NO mat file for "<<argv[i]<<" !!!!! SKIPPING"<<endl;
        continue;
      }
      map<string,PMatrix>::iterator it = mapP.lower_bound(ss.str());
      if (it != mapP.begin()) it--;
      else cout<<"First element, cant use smaller, must use larger !";
      P = it->second; 
      cout <<"Found next mat file for "<<argv[i]<<": "<<it->first<<endl;
    }
    // just a dummy class to do the metadata job for us
    BIAS::Camera<unsigned char> c;
    c.SetP(P);
    if (P.IsMetric()) {
      c.SetK(P.GetK());
      c.SetR(RMatrix(P.GetR()));
      c.SetC(P.GetC());
    }
    c.UpdateMetaData();
    CurImage.SetMetaData(*c.GetMetaData());


    if (FileType==ImageIO::FF_mip) {
      IOResult = ImageIO::Write(argv[i], CurImage, false, false );
    } else {
      IOResult = ImageIO::ExportImage(argv[i], CurImage, FileType, false, 
                                      0, false);
    }
    if (IOResult!=0) {
      ErrorFlag = true;
      BIASERR("Error writing "<< argv[i]);
      continue;
    }
    
  }
  if (ErrorFlag) cout << "There were errors." <<endl;
  if (ErrorFlag) return 1;
  return 0;
    
}












