/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _BIASTHREEDOUT_HH_
#define _BIASTHREEDOUT_HH_

#include "bias_config.h"
#include "Base/Debug/Debug.hh"
#include "Geometry/PMatrix.hh"
#include "Base/Geometry/HomgPoint3D.hh"
#include "Base/Math/Vector3.hh"
#include "Geometry/Quadric3D.hh"
#include "Geometry/CovMatrix3x3.hh"
#include "Base/Geometry/HomgLine3D.hh"
#include "Base/Geometry/HomgPlane3D.hh"
#include "Base/Image/Image.hh"


#include <list>
#include <vector>
#include <map>

/// use one sigma as the standard confidence region for ellipsoid surfaces
/// if you want to supply a probability, see defines in Quadric3D.hh
#define DEF_P_SCALE 1.0

#define D_3DOUT_PDECOMP 0x00001000

namespace BIAS {

  /** @class RGBA
      @brief stores rgb and alpha values [0;255] */
  class RGBA {
  public:
    unsigned char R;
    unsigned char G;
    unsigned char B;
    unsigned char A;
    RGBA(){};
    RGBA(unsigned char r, unsigned char g, unsigned char b, unsigned char a)
      {R=r; G=g; B=b; A=a;};
    inline unsigned char& operator[](const unsigned char& offset) {
     return ((unsigned char*)this)[offset];
    };
    inline const unsigned char& operator[](const unsigned char& offset) const {
     return ((unsigned char*)this)[offset];
    };
  };

  /** @class PMatrix3DData
      @brief stores important data for drawing a camera */
  class PMatrix3DData {
  public:
    Vector3<double>  C;        ///< camera center in wcs
    Vector3<double>  UL;       ///< upper left corner of image plane in wcs
    Vector3<double>  UR;       ///< upper right corner of image plane in wcs
    Vector3<double>  LL;       ///< lower left corner of image plane in wcs
    Vector3<double>  LR;       ///< lower right corner of image plane in wcs
    Vector3<double>  PP;       ///< principle point in wcs
    Vector3<double>  Up;       ///< Up-Vector (V)
    RGBA Color;                ///< color and transparency
  };

  /** @class Ellipsoid3DData
      @brief stores important data for drawing an ellipsoid */
  class Ellipsoid3DData {
  public:
    BIAS::Vector3<double> RotAxis; ///< rot to get 1st mainaxis from (1,0,0)
    double dAngle;                 ///< rotation in radians around RotAxes 
    double s0, s1, s2;             ///< lengths of main axes
    BIAS::Vector3<double> C;       ///< center of ellipsoid
    RGBA Color;
  };

 /** @class Ellipsoid3DData
     @brief stores data for drawing a point */
  class Point3DData {
  public:
    BIAS::Vector3<double> Point;
    RGBA Color;
  };

  /** @class Image3DData
      @brief  stores datas for a textured plane */
  class Image3DData {
  public:
    Vector3<double>  UL;       ///< upper left corner of image plane in wcs
    Vector3<double>  UR;       ///< upper right corner of image plane in wcs
    Vector3<double>  LL;       ///< lower left corner of image plane in wcs
    Vector3<double>  LR;       ///< lower right corner of image plane in wcs
    Image<unsigned char> ImageData;      ///< data of texture image

    Image3DData(){};

    Image3DData(const Image3DData& d){
      UL=d.UL; UR=d.UR; LL=d.LL; LR=d.LR; ImageData=d.ImageData;     
    }

    };

  /** @class Line3DData
      @brief stores data for drawing a line */
 class Line3DData {
  public:
    BIAS::Vector3<double> Start;
    BIAS::Vector3<double> End;
    RGBA Color;
  };

  /** @enum CameraDrawingStyle
      @brief different styles when drawing cameras, e.g. in GL or VRML */
  enum CameraDrawingStyle { PyramidMesh, PyramidSolid, Arrow };

  /** @enum PointDrawingStyle
      @brief different styles when drawing points, e.g. in GL or VRML */
  enum PointDrawingStyle { Point, PointSprite, Sphere, Box};

  /** @enum PointDrawingStyle
      @brief different styles when drawing lines, e.g. in GL or VRML */
  enum LineDrawingStyle { Solid, Dashed };

  /** @struct ThreeDOutParameters
      @brief configuration struct for drawing styles of various 3d objects */
  class ThreeDOutParameters {
  public:
    ThreeDOutParameters();
    PointDrawingStyle PointStyle;
    CameraDrawingStyle CameraStyle;
    LineDrawingStyle LineStyle;
    bool DrawUncertaintyEllipsoids;
    double PointSize;
    unsigned char LineWidth; // in pixel
  };

  /** @class ThreeDOut
      @brief unified output of 3D entities via GL or VRML */
  class ThreeDOut : public Debug 
  {
  public:
                   
    ThreeDOut(const BIAS::ThreeDOutParameters& params=
              BIAS::ThreeDOutParameters());
    ~ThreeDOut();

    void SetParams(const ThreeDOutParameters& params);

    /** @name Functions to add geometric objects for output
        @brief save the return value to update or delete the objects later 
        @{ */

    /** @return index value to change the point later
        @param v point coordinates */
    unsigned int AddPoint(const BIAS::Vector3<double> &v,
                          const RGBA &Color=RGBA(255,255,255,0));

    /** @return index value to change the point later
        @param v point coordinates */
    unsigned int AddPoint(const BIAS::HomgPoint3D &v, 
                          const RGBA &Color=RGBA(255,255,255,0));

    /** @return index value to change the line later
        @param Start Start point of line
        @param End End point of line */
    unsigned int AddLine(const BIAS::Vector3<double> &Start,
                         const BIAS::Vector3<double> &End,
                         const RGBA &Color=RGBA(255,255,255,0));

    /** @return index value to change the line later
        @param Start Start point of line
        @param End End point of line */
    unsigned int AddLine(const BIAS::HomgPoint3D &Start,
                         const BIAS::HomgPoint3D &End, 
                         const RGBA &Color=RGBA(255,255,255,0));
    
    /** @param dConfidenceRegionScale see Quadric3D.hh for the relationship
        between enlarging the standard confidence region and the probability
        that a measurement will be inside that region, 1.0 means sigma */
    unsigned int AddEllipsoid(BIAS::CovMatrix3x3 &cov, 
                              BIAS::Vector3<double> &C,
                              const RGBA &Color=RGBA(255,255,255,128),
                              const double& dConfidenceRegionScale = 
                              DEF_P_SCALE);

    /** @param dConfidenceRegionScale see Quadric3D.hh for the relationship
        between enlarging the standard confidence region and the probability
        that a measurement will be inside that region, 1.0 means sigma */
    unsigned int AddEllipsoid(BIAS::CovMatrix3x3 &cov, 
                              BIAS::HomgPoint3D &C,
                              const RGBA &Color=RGBA(255,255,255,128),
                              const double& dConfidenceRegionScale = 
                              DEF_P_SCALE);

    /** decompose P and add to internal data structures,
        @author woelk 01/2004 */
    unsigned int AddPMatrix(BIAS::PMatrix &P,
                            const unsigned int& width,
                            const unsigned int& height,
                            const RGBA &Color = RGBA(255,255,255,0),
                            const double &dScale = DEF_P_SCALE);


    /** @brief add a small image patch spanned by four 3d points
        small means really small (only a few pixels), the patch is approximated
        in a very complicated way in vrml to allow for inline data
        @param channels number of channels: 1=grey, 3=rgb
        @author koeser 09/2004 */
    unsigned int AddImage(const Vector3<double>& UL,
                          const Vector3<double>& UR,
                          const Vector3<double>& LL,
                          unsigned int Width, unsigned int Height,
                          unsigned char* pData,
                          unsigned int channels = 1);

    /** @} 
        @name Functions to update previously added objects for output
        @{ */

    /** @param PointIndex value previously returned by AddPoint */
    void UpdatePoint(const unsigned int& PointIndex,
                     const BIAS::Vector3<double> &v,
                     const RGBA &Color = RGBA(255,255,255,0));

    /** @param PointIndex value previously returned by AddPoint */
    void UpdatePoint(const unsigned int& PointIndex,
                     const BIAS::HomgPoint3D &v,
                     const RGBA &Color = RGBA(255,255,255,0));

    /** @param PIndex value previously returned by AddPMatrix */
    void UpdatePMatrix(const unsigned int& PIndex, BIAS::PMatrix &P, 
                       const unsigned int& width,
                       const unsigned int& height,
                       const RGBA &Color = RGBA(255,255,255,0),
                       const double& dScale = DEF_P_SCALE);

    /** @param PIndex value previously returned by AddPMatrix */
    void UpdatePMatrixColor(const unsigned int& PIndex, 
                            const RGBA &Color);

    /** @param LineIndex value previously returned by AddLine
        @param Start Start point of line
        @param End End point of line */
    void UpdateLine(const unsigned int& LineIndex,
                    const BIAS::Vector3<double> &Start,
                    const BIAS::Vector3<double> &End,
                    const RGBA &Color = RGBA(255,255,255,0));

    /** @param LineIndex value previously returned by AddLine
        @param Start Start point of line
        @param End End point of line */
    void UpdateLine(const unsigned int& LineIndex,
                    const BIAS::HomgPoint3D &Start,
                    const BIAS::HomgPoint3D &End, 
                    const RGBA &Color = RGBA(255,255,255,0));
    
    /** @param dConfidenceRegionScale see Quadric3D.hh for the relationship
        between enlarging the standard confidence region and the probability
        that a measurement will be inside that region, 1.0 means sigma */
    void UpdateEllipsoid(const unsigned int& EIndex,
                         BIAS::CovMatrix3x3 &cov, 
                         BIAS::Vector3<double> &C,
                         const RGBA &Color = RGBA(255,255,255,128),
                         const double& dConfidenceRegionScale = DEF_P_SCALE);

    /** @param dConfidenceRegionScale see Quadric3D.hh for the relationship
        between enlarging the standard confidence region and the probability
        that a measurement will be inside that region, 1.0 means sigma */
    void UpdateEllipsoid(const unsigned int& EIndex,
                         BIAS::CovMatrix3x3 &cov, 
                         BIAS::HomgPoint3D &C,
                         const RGBA &Color = RGBA(255,255,255,128),
                         const double& dConfidenceRegionScale = DEF_P_SCALE);
    /** @} 
        @name Functions to remove previously added objects for output
        @brief PMatrices cannot be removed, however, you can set alpha to 255
        if you dont wanna see them.
        @{ */

    /** @brief completely removes the point from the internal structures,
        there is no way to get it back */
    void RemovePoint(const unsigned int& PointIndex);

    /** @brief completely removes all points from the internal structures,
        there is no way to get them back */
    void RemoveAllPoints();

    /** @brief completely removes the line from the internal structures,
        there is no way to get it back */
    void RemoveLine(const unsigned int& LineIndex);
   
    /** @brief completely removes the ellipsoid from the internal structures,
        there is no way to get it back */
    void RemoveEllipsoid(const unsigned int& EIndex);

    /** @brief completely removes the ellipsoid from the internal structures,
        there is no way to get it back */
    void RemoveAllEllipsoids();

    /** @brief completely removes the image from the internal structures,
        there is no way to get it back */
    void RemoveImage(const unsigned int& EIndex);

    /** @brief completely removes all images from the internal structures,
        there is no way to get it back */
    void RemoveAllImages();
    
    /** @brief removes the P-matrix index,
        there is no way to get it back */
    void RemovePMatrix(int index);

    /** @brief completely removes all images from the internal structures,
        there is no way to get it back */
    void RemoveAllPMatrices();
    
    /** @} 
        @name Functions for openGL output (if BIAS_HAVE_OPENGL)
        @{ */

#ifdef BIAS_HAVE_OPENGL
     
    /// @brief draw all points with the same color:
    void OpenGLOutPoints(const RGBA &Color);

    /// @brief draw all points with their own color:
    void OpenGLOutPoints();

    /// @brief draw all pmatrices with the same color:
    void OpenGLOutPMatrices(const RGBA &Color);

    /// @brief draw all pmatrices with their own color:
    void OpenGLOutPMatrices();

    /// @brief draw all lines with the same color:
    void OpenGLOutLines(const RGBA &Color);

    /// @brief draw all pmatrices with their own color
    /// @todo linedrawingstyle=dashed not implmented
    void OpenGLOutLines();

    /// @brief draw all ellipsoids with the same color:
    void OpenGLOutEllipsoids(const RGBA &Color);

    /// @brief draw all ellipsoids with their own color:
    void OpenGLOutEllipsoids();

    /// @brief draw all image patches
    void OpenGLOutImages();

    /// @brief draw everything (in the same color)
    void OpenGLOut(const RGBA &Color);

    /// @brief draw everything
    void OpenGLOut();
#endif

    Vector3<double> GetCenterOfPoints();


    /** @} 
        @name Functions for VRML output
        @{ */

    /** @brief flush all 3d objects to a vrml file with name sFilename, this 
        is the function most users would call */
    int VRMLOut(const std::string &sFilename);

    /** @brief flush all 3d objects to a vrml file with name sFilename, use
        Color as the color for all elements */
    int VRMLOut(const std::string &sFilename, const RGBA &Color);

    /** @brief only write (previously stored) points into an open vrml file,
        override their individual colors with Color */
    int VRMLOutPoints(std::ostream& vrml, const RGBA &Color);

    /** @brief only write (previously stored) points into an open vrml file */
    int VRMLOutPoints(std::ostream& vrml);

    /** @brief only write (previously stored) cameras into an open vrml file,
        override their individual colors with Color */
    int VRMLOutPMatrices(std::ostream& vrml, const RGBA &Color);

    /** @brief only write (previously stored) cameras into an open vrml file */
    int VRMLOutPMatrices(std::ostream& vrml);

    /** @brief only write (previously stored) lines into an open vrml file,
        override their individual colors with Color */
    int VRMLOutLines(std::ostream& vrml, const RGBA &Color);

    /** @brief only write (previously stored) points into an open vrml file */
    int VRMLOutLines(std::ostream& vrml);

    /** @brief only write (previously stored) ellipsoids to an open vrml file,
        override their individual colors with Color */
    int VRMLOutEllipsoids(std::ostream& vrml, const RGBA &Color);

    /** @brief only write (previously stored) ellipsoids into an open vrml 
        file */
    int VRMLOutEllipsoids(std::ostream& vrml);

    /** @brief only write (previously stored) images into an open vrml file */
    int VRMLOutImages(ostream& VRMLFile);

    /** @brief write the VRML 2.0 header into an open file */
    int VRMLOutWriteHeader(std::ostream& vrml);

    /** @} 
        @name Debug functions
        @{ */
    /** @brief list all internal 3d objects on cout */
    void Dump();
    /** @}  */
  protected:
#ifdef BIAS_HAVE_OPENGL
    void OpenGLOutPMatrices_(const RGBA &Color, bool UseOneColor);
#endif
    
    // helper for VRMLOutPoints()
    int VRMLOutPointsAsPointSet_(std::ostream &VRMLFile);
    int VRMLOutPointsAsBoxes_(std::ostream &VRMLFile);
    int VRMLOutPointsAsSpheres_(std::ostream &VRMLFile);
	
    /// holds the params set by constructor or SetParams
    ThreeDOutParameters Params_;
      
    /// the cameras are saved in a vector for efficiency reasons
    std::map<unsigned int, PMatrix3DData > mapPMatrices_;
    /// counter for the next index, gaps (due to delete) are not reused by now
    unsigned int NextPIndex_;

    /// map of all image structures
    std::map<unsigned int, Image3DData> mapImages_;
    /// counter for the next index, gaps (due to delete) are not reused by now
    unsigned int NextImageIndex_;

    /// map of all point structures
    std::map<unsigned int, Point3DData> mapPoints_;
    /// counter for the next index, gaps (due to delete) are not reused by now
    unsigned int NextPointIndex_;
    
    /// map of all line structures
    std::map<unsigned int, Line3DData> mapLines_;
    /// counter for the next line index, gaps (due to delete) arent reused
    unsigned int NextLineIndex_;

    /// map of all ellipsoid structures
    std::map<unsigned int, Ellipsoid3DData> mapEllipsoids_;
    /// counter for the next ellipsoid index, gaps are not reused by now
    unsigned int NextEllipsoidIndex_;
  }; // end class
} // end namespace


#endif





